﻿using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Branch.Entities;
using CRM.Domain.Branch.Events;
using CRM.Domain.Branch.Models;
using CRM.Domain.Branch.Snapshots;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Branch
{
    public class BranchAggregate : SnapshotAggregateRoot<BranchAggregate, BranchId, BranchSnapshot>,
        IEmit<BranchCreatedEvent>,
        IEmit<BranchNameModifiedEvent>,
        IEmit<BranchAddressModifiedEvent>,
        IEmit<BranchDeletedEvent>,
        IEmit<BranchCustomerIdModifiedEvent>,
        IEmit<BranchContactDetailsModifiedEvent>
    {
        private bool _created;

        public BranchAggregate(BranchId id) : base(id, SnapshotEveryFewVersionsStrategy.With(5))
        {
        }

        public string BranchName { get; private set; } = "";
        public AddressModel Address { get; private set; } = new AddressModel();
        public bool Deleted { get; private set; }
        public string CustomerId { get; private set; }
        public ICollection<ContactDetail> ContactDetails { get; private set; } = new List<ContactDetail>();

        protected override Task<BranchSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(
                () => new BranchSnapshot(BranchName, Address, CustomerId, ContactDetails, Deleted, _created),
                cancellationToken);
        }

        protected override Task LoadSnapshotAsync(BranchSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                _created = snapshot.Created;
                Deleted = snapshot.Deleted;
                BranchName = snapshot.BranchName;
                Address = snapshot.Address;
                CustomerId = snapshot.CustomerId;
                ContactDetails = snapshot.ContactDetails;
            }, cancellationToken);
        }

        #region IExecutionResult

        public IExecutionResult CreateBranch(string name,
            AddressModel address,
            string customerId,
            ICollection<ContactDetailModel> contactDetails)
        {
            if (_created) return ExecutionResult.Failed();

            var _event = new BranchCreatedEvent(name, address, customerId, contactDetails);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveBranch()
        {
            if (!_created || Deleted)
                return ExecutionResult.Failed($"Branch with id {Id.Value} does not exist");

            var _event = new BranchDeletedEvent();

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyBranchAddress(AddressModel address)
        {
            if (!_created || Deleted)
                return ExecutionResult.Failed($"Branch with id {Id.Value} does not exist");

            var _event = new BranchAddressModifiedEvent(address);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyBranchContactDetails(ICollection<ContactDetailModel> contactDetails)
        {
            if (!_created || Deleted)
                return ExecutionResult.Failed($"Branch with id {Id.Value} does not exist");

            var _event = new BranchContactDetailsModifiedEvent(contactDetails);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyBranchName(string name)
        {
            if (!_created || Deleted)
                return ExecutionResult.Failed($"Branch with id {Id.Value} does not exist");

            var _event = new BranchNameModifiedEvent(name);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyBranchCustomerId(string customerId)
        {
            if (!_created || Deleted)
                return ExecutionResult.Failed($"Branch with id {Id.Value} does not exist");

            var _event = new BranchCustomerIdModifiedEvent(customerId);

            Emit(_event);
            return ExecutionResult.Success();
        }

        #endregion IExecutionResult

        #region Apply

        public void Apply(BranchCreatedEvent aggregateEvent)
        {
            _created = true;
            BranchName = aggregateEvent.BranchName;
            Address = aggregateEvent.Address;
            CustomerId = aggregateEvent.CustomerId;
            ContactDetails = aggregateEvent.ContactDetails.Select(x =>
                new ContactDetail
                (ContactDetailId.With(x.Id),
                    x.ContactDetailName,
                    x.ContactDetailValue,
                    x.IsPrivate,
                    x.ContactType)).ToList();
        }

        public void Apply(BranchContactDetailsModifiedEvent aggregateEvent)
        {
            ContactDetails = aggregateEvent.ContactDetails.Select(x =>
                new ContactDetail
                (ContactDetailId.With(x.Id),
                    x.ContactDetailName,
                    x.ContactDetailValue,
                    x.IsPrivate,
                    x.ContactType)).ToList();
        }

        public void Apply(BranchNameModifiedEvent aggregateEvent)
        {
            BranchName = aggregateEvent.BranchName;
        }

        public void Apply(BranchAddressModifiedEvent aggregateEvent)
        {
            Address = aggregateEvent.Address;
        }

        public void Apply(BranchDeletedEvent aggregateEvent)
        {
            Deleted = true;
        }

        public void Apply(BranchCustomerIdModifiedEvent aggregateEvent)
        {
            CustomerId = aggregateEvent.CompanyId;
        }

        #endregion Apply
    }
}