﻿using System.Collections.Generic;
using System.Linq;
using CRM.Domain.CallScript.Entities;
using CRM.Domain.CallScript.Events;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;

namespace CRM.Domain.CallScript
{
    public class CallScriptAggregate : AggregateRoot<CallScriptAggregate, CallScriptId>,
        IEmit<CallScriptCreatedEvent>,
        IEmit<CallScriptModifiedEvent>,
        IEmit<CommentCreatedEvent>,
        IEmit<CommentModifiedEvent>,
        IEmit<CommentRemovedEvent>,
        IEmit<TagAddedEvent>,
        IEmit<CallScriptRemovedEvent>,
        IEmit<CallScriptDossierIdModifiedEvent>
    {
        public CallScriptAggregate(CallScriptId id) : base(id)
        {
        }

        private bool _created { get; set; }

        public bool Deleted { get; private set; }
        public string? CallScriptBody { get; private set; }
        public string? DossierId { get; private set; }
        public ICollection<string> Tags { get; private set; } = new List<string>();

        public ICollection<CallScriptComment> CommentList { get; }
            = new List<CallScriptComment>();

        #region IExecutionResult

        public IExecutionResult CreateCallScript(string callScriptBody,
            string dossierId,
            List<string> tags)
        {
            if (_created) return ExecutionResult.Failed("This callscript already exists!");
            var _event = new CallScriptCreatedEvent(callScriptBody, dossierId, tags);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyCallScript(string callScriptBody)
        {
            if (!_created) return ExecutionResult.Failed();
            var _event = new CallScriptModifiedEvent(callScriptBody);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveCallScript()
        {
            var _event = new CallScriptRemovedEvent();
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult UpsertComment(string commentBody,
            string commentQuote,
            string userId,
            CallScriptCommentId? callScriptCommentId,
            string callscriptId)
        {
            if (!_created) return ExecutionResult.Failed();
            //var existingComment = CommentList.SingleOrDefault(x => x.Id == callScriptCommentId);
            //if (existingComment == null)
            //{
            var _event =
                new CommentCreatedEvent(CallScriptCommentId.New, commentBody, commentQuote, userId, callscriptId);
            Emit(_event);
            return ExecutionResult.Success();
            //}
            //else
            //{
            //    var _event = new CommentModifiedEvent(commentBody, commentQuote, callScriptCommentId);
            //    Emit(_event);
            //    return ExecutionResult.Success();
            //}
        }

        public IExecutionResult RemoveComment(CallScriptCommentId callScriptCommentId)
        {
            if (!_created) return ExecutionResult.Failed();
            if (!CommentList.Any(x => x.Id == callScriptCommentId)) return ExecutionResult.Failed();
            var _event = new CommentRemovedEvent(callScriptCommentId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddTag(string tag)
        {
            if (!_created) return ExecutionResult.Failed();
            if (Tags.Any(x => x == tag)) return ExecutionResult.Failed();
            var _event = new TagAddedEvent(tag);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDossierIdCallScript(string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Callscript with id {Id.Value} does not exist");
            var e = new CallScriptDossierIdModifiedEvent(dossierId);
            Emit(e);
            return ExecutionResult.Success();
        }

        #endregion IExecutionResult

        #region Apply

        public void Apply(CallScriptCreatedEvent aggregateEvent)
        {
            CallScriptBody = aggregateEvent.CallScriptBody;
            DossierId = aggregateEvent.DossierId;
            Tags = aggregateEvent.Tags;
            _created = true;
        }

        public void Apply(CallScriptModifiedEvent aggregateEvent)
        {
            CallScriptBody = aggregateEvent.CallScriptBody;
        }

        public void Apply(CommentCreatedEvent aggregateEvent)
        {
            CommentList.Add(new CallScriptComment(aggregateEvent.Id, aggregateEvent.CommentBody,
                aggregateEvent.CommentQuote, aggregateEvent.UserId));
        }

        public void Apply(CommentModifiedEvent aggregateEvent)
        {
            var comment = CommentList.Single(x => x.Id == aggregateEvent.CallscriptCommentId);
            comment.CommentBody = aggregateEvent.CommentBody;
            comment.CommentQuote = aggregateEvent.CommentQuote;
        }

        public void Apply(CommentRemovedEvent aggregateEvent)
        {
            var comment = CommentList.Single(x => x.Id == aggregateEvent.CallscriptCommentId);
            CommentList.Remove(comment);
        }

        public void Apply(TagAddedEvent aggregateEvent)
        {
            Tags.Add(aggregateEvent.Tag);
        }

        public void Apply(CallScriptRemovedEvent aggregateEvent)
        {
            Deleted = true;
        }
        public void Apply(CallScriptDossierIdModifiedEvent aggregateEvent)
        {
            DossierId = aggregateEvent.DossierId;
        }

        #endregion Apply

    }
}