﻿using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Customer.Events;
using CRM.Domain.Customer.Snapshots;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Customer
{
    public class CustomerAggregate : SnapshotAggregateRoot<CustomerAggregate, CustomerId, CustomerSnapshot>,
        IEmit<CustomerCreatedEvent>,
        IEmit<CustomerMainCustomerBranchModifiedEvent>,
        IEmit<CustomerWebPortalLicenseAmountModifiedEvent>,
        IEmit<CustomerNameModifiedEvent>,
        IEmit<CustomerDeletedEvent>,
        IEmit<CustomerSubscriptionsModifiedEvent>
    {
        private bool _created;

        public CustomerAggregate(CustomerId id) : base(id, SnapshotEveryFewVersionsStrategy.With(5))
        {
        }

        public bool Deleted { get; private set; }
        public string CustomerName { get; private set; } = "";
        public string Tenant { get; private set; } = "";
        public int? CCMCLientId { get; private set; }
        public int WebPortalLicenseAmount { get; private set; }
        public string CustomerBranchId { get; private set; } = "";
        public ICollection<string> Subscriptions { get; private set; } = new List<string>();

        protected override Task<CustomerSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() =>
                new CustomerSnapshot(_created, Deleted, CustomerName, Tenant, CCMCLientId, WebPortalLicenseAmount,
                    CustomerBranchId, Subscriptions));
        }

        protected override Task LoadSnapshotAsync(CustomerSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                _created = snapshot.Created;
                Deleted = snapshot.Deleted;
                CustomerName = snapshot.CustomerName;
                Tenant = snapshot.Tenant;
                CCMCLientId = snapshot.CCMCLientId;
                WebPortalLicenseAmount = snapshot.WebPortalLicenseAmount;
                CustomerBranchId = snapshot.CustomerBranchId;
                Subscriptions = snapshot.Subscriptions;
            }, cancellationToken);
        }


        #region IExecutionResult

        public IExecutionResult CreateCustomer(string customerName,
            string tenant,
            int? cCMCLientId,
            int webPortalLicenseAmount)
        {
            if (_created) return ExecutionResult.Failed();
            var _event = new CustomerCreatedEvent(customerName, tenant, cCMCLientId, webPortalLicenseAmount);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyMainCustomerBranch(string customerBranchId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Customer with id {Id.Value} does not exist");
            var _event = new CustomerMainCustomerBranchModifiedEvent(customerBranchId);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyWebPortalLicenseAmount(int webPortalLicenseAmount)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Customer with id {Id.Value} does not exist");
            var _event = new CustomerWebPortalLicenseAmountModifiedEvent(webPortalLicenseAmount);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyCustomerName(string customerName)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Customer with id {Id.Value} does not exist");
            var _event = new CustomerNameModifiedEvent(customerName);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteCustomer()
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Customer with id {Id.Value} does not exist");
            var _event = new CustomerDeletedEvent();

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifySubscriptions(ICollection<string> subscriptions)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Customer with id {Id.Value} does not exist");
            var _event = new CustomerSubscriptionsModifiedEvent(subscriptions);
            Emit(_event);
            return ExecutionResult.Success();
        }

        #endregion

        #region Apply

        public void Apply(CustomerCreatedEvent aggregateEvent)
        {
            CustomerName = aggregateEvent.CustomerName;
            Tenant = aggregateEvent.Tenant;
            CCMCLientId = aggregateEvent.CCMCLientId;
            WebPortalLicenseAmount = aggregateEvent.WebPortalLicenseAmount;
            _created = true;
        }

        public void Apply(CustomerMainCustomerBranchModifiedEvent aggregateEvent)
        {
            CustomerBranchId = aggregateEvent.CustomerBranchId;
        }

        public void Apply(CustomerWebPortalLicenseAmountModifiedEvent aggregateEvent)
        {
            WebPortalLicenseAmount = aggregateEvent.WebPortalLicenseAmount;
        }

        public void Apply(CustomerNameModifiedEvent aggregateEvent)
        {
            CustomerName = aggregateEvent.CustomerName;
        }

        public void Apply(CustomerDeletedEvent aggregateEvent)
        {
            Deleted = true;
        }

        public void Apply(CustomerSubscriptionsModifiedEvent aggregateEvent)
        {
            Subscriptions = aggregateEvent.Subscriptions;
        }

        #endregion
    }
}