﻿using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Dossier.Events;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Dossier
{
    public class DossierAggregate : SnapshotAggregateRoot<DossierAggregate, DossierId, DossierSnapshot>,
        IEmit<DossierNameModifiedEvent>,
        IEmit<DossierDeletedEvent>,
        IEmit<DossierCreatedEvent>,
        IEmit<DossierBranchAddedEvent>,
        IEmit<DossierBranchRemovedEvent>,
        IEmit<DossierCustomerIdModifiedEvent>
    {
        private bool _created;


        public DossierAggregate(DossierId id) : base(id, SnapshotEveryFewVersionsStrategy.Default)
        {
        }

        public bool Deleted { get; private set; }
        public string DossierName { get; private set; } = "";
        public int? CCMId { get; private set; }
        public string CustomerId { get; private set; } = "";
        public ICollection<string> BranchIds { get; private set; } = new List<string>();


        public IExecutionResult ModifyDossierName(string name)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Dossier with id {Id.Value} does not exist");
            var e = new DossierNameModifiedEvent(name);
            Emit(e);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteDossier()
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Dossier with id {Id.Value} does not exist");
            var e = new DossierDeletedEvent();
            Emit(e);
            return ExecutionResult.Success();
        }

        public IExecutionResult CreateDossier(
            string name,
            string customerId,
            int? ccmId = null)
        {
            if (_created) return ExecutionResult.Failed($"Dossier with id {Id.Value} already exist");
            var e = new DossierCreatedEvent(name, customerId, ccmId);
            Emit(e);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddBranchId(string id)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Dossier with id {Id.Value} does not exist");
            if (BranchIds.Contains(id))
                return ExecutionResult.Failed(
                    $"Branch with id {id} already assigned to Dossier {DossierName} : {Id.Value}");
            var e = new DossierBranchAddedEvent(id);
            Emit(e);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveBranchId(string id)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Dossier with id {Id.Value} does not exist");
            if (!BranchIds.Contains(id))
                return ExecutionResult.Failed(
                    $"Branch with id {id} has not been assigned to Dossier {DossierName} : {Id.Value}");
            var e = new DossierBranchRemovedEvent(id);
            Emit(e);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDossierCustomerId(string customerId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Dossier with id {Id.Value} does not exist");
            var e = new DossierCustomerIdModifiedEvent(customerId);
            Emit(e);
            return ExecutionResult.Success();

        }

        #region Snapshotting

        protected override Task<DossierSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new DossierSnapshot(Deleted, DossierName, CCMId, CustomerId), cancellationToken);
        }

        protected override Task LoadSnapshotAsync(DossierSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                Deleted = snapshot.Deleted;
                DossierName = snapshot.DossierName;
                CCMId = snapshot.CCMId;
                CustomerId = snapshot.ClientId;
                BranchIds = snapshot.BranchIds;
                _created = true;
            }, cancellationToken);
        }

        #endregion

        #region Apply

        public void Apply(DossierNameModifiedEvent aggregateEvent)
        {
            DossierName = aggregateEvent.DossierName;
        }

        public void Apply(DossierDeletedEvent aggregateEvent)
        {
            Deleted = true;
        }

        public void Apply(DossierCreatedEvent aggregateEvent)
        {
            _created = true;
            DossierName = aggregateEvent.DossierName;
            CustomerId = aggregateEvent.CustomerId;
            CCMId = aggregateEvent.CcmId;
        }

        public void Apply(DossierBranchAddedEvent aggregateEvent)
        {
            BranchIds.Add(aggregateEvent.BranchId);
        }

        public void Apply(DossierBranchRemovedEvent aggregateEvent)
        {
            BranchIds.Remove(aggregateEvent.BranchId);
        }

        public void Apply(DossierCustomerIdModifiedEvent aggregateEvent)
        {
            CustomerId = aggregateEvent.CustomerId;
        }

        #endregion

    }
}