﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Duty.Entities;
using CRM.Domain.Duty.Events;
using CRM.Domain.Duty.Models;
using CRM.Domain.Duty.Snapshots;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Exceptions;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Duty
{
    public class DutyAggregate : SnapshotAggregateRoot<DutyAggregate, DutyId, DutySnapshot>,
        IEmit<DutyAddedEvent>,
        IEmit<DutyEmployeeAddedEvent>,
        IEmit<DutyEmployeeRemovedEvent>,
        IEmit<DutyRemovedEvent>,
        IEmit<DutyCategoryModifiedEvent>,
        IEmit<DutyEndModifiedEvent>,
        IEmit<DutyStartModifiedEvent>,
        IEmit<DutyActionDescriptionModifiedEvent>,
        IEmit<DutyUpdatedEvent>,
        IEmit<DutyEmployeeUpdatedEvent>,
        IEmit<DutyModifiedDossierIdEvent>,
        IEmit<DutyAllEmployeesRemovedEvent>

    {
        private bool _isCreated;

        public DutyAggregate(DutyId id) : base(id, SnapshotEveryFewVersionsStrategy.With(5))
        {
        }

        public string DossierId { get; set; } = "";
        public string Category { get; private set; } = "";
        public DateTime StartDate { get; private set; }
        public DateTime EndDate { get; private set; }
        public string StartAt { get; set; }
        public string EndAt { get; set; }
        public ICollection<DutyContact> Contacts { get; private set; } = new List<DutyContact>();
        public bool Deleted { get; private set; }
        public string ActionDescription { get; set; } = "";

        protected override Task<DutySnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new DutySnapshot(Category, StartDate, EndDate, Contacts
                .Select(x => new DutyEmployeeModel(x.Id.Value, x.Tags, x.OrderId, x.ContactId,
                    x.ContactPoints.Select(y =>
                        new ContactPointModel(y.Id.Value, y.ContactDetailId, y.Description)).ToList()))
                .ToList(), _isCreated, Deleted, DossierId, StartAt, EndAt), cancellationToken);
        }

        protected override Task LoadSnapshotAsync(DutySnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                Category = snapshot.Category;
                StartDate = snapshot.StartDate;
                EndDate = snapshot.EndDate;
                Contacts = snapshot.Contacts
                    .Select(x => new DutyContact(DutyContactId.With(x.Id),
                        x.EmployeeId,
                        x.ContactPoints.Select(y => new DutyContactPoint(DutyContactPointId.With(y.Id),
                            y.Description,
                            y.ContactDetailId)).ToList(),
                        x.Tags, x.OrderId)).ToList();
                Deleted = snapshot.Deleted;
                _isCreated = snapshot.IsCreated;
                DossierId = snapshot.DossierId;
                EndAt = snapshot.EndAt;
                StartAt = snapshot.StartAt;
            }, cancellationToken);
        }

        public IExecutionResult Update(
            DateTime startDate,
            DateTime endDate,
            string startAt,
            string endAt,
            string category,
            string actionDescription,
            string dossierId,
            ICollection<DutyEmployeeModel> dutyEmployees)
        {
            var _event = new DutyUpdatedEvent(dossierId, category, startDate, endDate, actionDescription, endAt,
                startAt);
            Emit(_event);

            var updateDutyEmployees = new DutyEmployeeUpdatedEvent(dutyEmployees);
            Emit(updateDutyEmployees);
            return ExecutionResult.Success();
        }

        #region Apply

        public void Apply(DutyAddedEvent aggregateEvent)
        {
            Category = aggregateEvent.Category;
            StartDate = aggregateEvent.StartDate;
            EndDate = aggregateEvent.EndDate;
            DossierId = aggregateEvent.DossierId;

            _isCreated = true;
        }

        public void Apply(DutyEmployeeAddedEvent aggregateEvent)
        {
            Contacts.Add(new DutyContact(aggregateEvent.Id, aggregateEvent.EmployeeId, aggregateEvent.ContactPoints
                    .Select(
                        x => new DutyContactPoint(DutyContactPointId.With(x.Id), x.Description,
                            x.ContactDetailId)).ToList()
                , aggregateEvent.Tags, aggregateEvent.OrderId));
        }

        public void Apply(DutyEmployeeRemovedEvent aggregateEvent)
        {
            var dutyEmployeeRemoved = Contacts.Single(x => x.Id == aggregateEvent.Id);
            Contacts.Remove(dutyEmployeeRemoved);
        }

        public void Apply(DutyRemovedEvent aggregateEvent)
        {
            Deleted = true;
        }

        public void Apply(DutyCategoryModifiedEvent aggregateEvent)
        {
            Category = aggregateEvent.Category;
        }

        public void Apply(DutyEndModifiedEvent aggregateEvent)
        {
            EndDate = aggregateEvent.EndDate;
        }

        public void Apply(DutyStartModifiedEvent aggregateEvent)
        {
            StartDate = aggregateEvent.StartDate;
        }

        public void Apply(DutyActionDescriptionModifiedEvent aggregateEvent)
        {
            ActionDescription = aggregateEvent.ActionDescription;
        }

        public void Apply(DutyUpdatedEvent aggregateEvent)
        {
            Category = aggregateEvent.Category;
            StartDate = aggregateEvent.StartDate;
            EndDate = aggregateEvent.EndDate;
            DossierId = aggregateEvent.DossierId;
        }
        public void Apply(DutyEmployeeUpdatedEvent aggregateEvent)
        {
            Contacts.Clear();

            foreach (var dutyEmployee in aggregateEvent.DutyEmployees)
            {
                Contacts.Add(new DutyContact(new DutyContactId(dutyEmployee.Id), dutyEmployee.EmployeeId, dutyEmployee.ContactPoints
                        .Select(
                            x => new DutyContactPoint(DutyContactPointId.With(x.Id), x.Description,
                                x.ContactDetailId)).ToList()
                    , dutyEmployee.Tags, dutyEmployee.OrderId));
            }
           
        }
        public void Apply(DutyModifiedDossierIdEvent aggregateEvent)
        {
            DossierId = aggregateEvent.DossierId;
        }

        public void Apply(DutyAllEmployeesRemovedEvent aggregateEvent)
        {
            Contacts.Clear();
        }

        #endregion Apply

        #region IExecutionResult

        public IExecutionResult CreateDutyWithEmployees(string category,
            DateTime startTime,
            DateTime endTime,
            string actionDescription,
            string dossierId,
            ICollection<DutyEmployeeModel> employees,
            string startAt,
            string endAt)
        {
            CreateDuty(category, startTime, endTime, actionDescription, dossierId, startAt, endAt);
            if (employees == null || employees.Count <= 0) return ExecutionResult.Success();
            return employees.Any(emp =>
                UpsertEmployee(new DutyContactId(emp.Id), emp.EmployeeId, emp.ContactPoints, emp.Tags, emp.OrderId,
                    emp.MarkedForRemoval) == ExecutionResult.Failed())
                ? ExecutionResult.Failed()
                : ExecutionResult.Success();
        }

        public IExecutionResult CreateDuty(string category,
            DateTime startTime,
            DateTime endTime,
            string actionDescription,
            string dossierId,
            string startAt,
            string endAt)
        {
            if (_isCreated) return ExecutionResult.Failed();

            var _event = new DutyAddedEvent(dossierId, category, startTime, endTime, actionDescription, endAt, startAt);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteDuty()
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) throw DomainError.With("already deleted!");

            var _event = new DutyRemovedEvent();

            Emit(_event);
            return ExecutionResult.Success();
        }

        /// <summary>
        ///     Gebruiken bij aanmaak duty
        /// </summary>
        /// <param name="employeeId"></param>
        /// <param name="dutyContactPoints"></param>
        /// <param name="tags"></param>
        /// <param name="orderId"></param>
        /// <returns></returns>
        public IExecutionResult AddDutyEmployee(string employeeId,
            ICollection<ContactPointModel> dutyContactPoints,
            ICollection<string> tags,
            int orderId)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var _event = new DutyEmployeeAddedEvent(DutyContactId.New, employeeId, dutyContactPoints, tags, orderId);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyEmployees(ICollection<DutyEmployeeModel> employees)
        {
            foreach (var emp in employees)
                if (UpsertEmployee(new DutyContactId(emp.Id), emp.EmployeeId, emp.ContactPoints, emp.Tags, emp.OrderId,
                    emp.MarkedForRemoval) == ExecutionResult.Failed())
                    return ExecutionResult.Failed();
            return ExecutionResult.Success();
        }

        /// <summary>
        ///     Gebruiken bij aanmaken, verwijderen en aanpassen van Employees bij duties.
        /// </summary>
        /// <param name="id"></param>
        /// <param name="employeeId"></param>
        /// <param name="dutyContactPoints"></param>
        /// <param name="tags"></param>
        /// <param name="orderId"></param>
        /// <param name="remove"></param>
        /// <returns></returns>
        public IExecutionResult UpsertEmployee(DutyContactId id,
            string employeeId,
            ICollection<ContactPointModel> dutyContactPoints,
            ICollection<string> tags,
            int orderId,
            bool remove)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var employee = Contacts.SingleOrDefault(x => x.Id.Equals(id));
            if (employee != null)
            {
                if (remove) return RemoveDutyEmployee(id);
                RemoveDutyEmployee(id);
            }

            //Check of de employeeId al in de duty zit. Indien zo: Failed
            employee = Contacts.FirstOrDefault(x => x.ContactId.Equals(employeeId));
            if (employee != null) return ExecutionResult.Failed();
            var _event = new DutyEmployeeAddedEvent(id, employeeId, dutyContactPoints, tags, orderId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveDutyEmployee(DutyContactId id)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var employeeToRemove = Contacts.SingleOrDefault(x => x.Id.Equals(id));

            if (Contacts.All(x => x.Id != id))
                return ExecutionResult.Failed();

            var _event = new DutyEmployeeRemovedEvent(id);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult UpdateDutyEmployee(ICollection<DutyEmployeeModel> employees)
        {
            var _event = new DutyEmployeeUpdatedEvent(employees);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyCategory(string category)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var _event = new DutyCategoryModifiedEvent(category);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyStartTime(DateTime startDate,
            string startTime)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var _event = new DutyStartModifiedEvent(startDate, startTime);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyEndTime(DateTime endDate,
            string endTime)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var _event = new DutyEndModifiedEvent(endDate, endTime);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyActionDescription(string actionDescription)
        {
            if (!_isCreated) return ExecutionResult.Failed();
            if (Deleted) return ExecutionResult.Failed();

            var _event = new DutyActionDescriptionModifiedEvent(actionDescription);

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDutyDossierId(string dossierId)
        {
            if (!_isCreated || Deleted) return ExecutionResult.Failed($"Duty with id {Id.Value} does not exist");
            
            var _event = new DutyModifiedDossierIdEvent(dossierId);
            Emit(_event);
            return ExecutionResult.Success();

        }

        public IExecutionResult RemoveDutyAllEmployees()
        {
            if (!_isCreated || Deleted) return ExecutionResult.Failed($"Duty with id {Id.Value} does not exist");

            var _event = new DutyAllEmployeesRemovedEvent();
            Emit(_event);
            return ExecutionResult.Success();
        }
        #endregion IExecutionResult


    }
}