﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.DutyCategoryAndTemplate.Entities;
using CRM.Domain.DutyCategoryAndTemplate.Events.GlobalDutyTemplate;
using CRM.Domain.DutyCategoryAndTemplate.Events.GlobalDutyTemplate.TimeSlots;
using CRM.Domain.DutyCategoryAndTemplate.Models;
using CRM.Domain.DutyCategoryAndTemplate.Snapshots;
using CRM.Domain.Shared.ValueObjects;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.DutyCategoryAndTemplate.Aggregates
{
    public class GlobalDutyTemplateAggregate: SnapshotAggregateRoot<GlobalDutyTemplateAggregate, GlobalDutyTemplateId, GlobalDutyTemplateSnapshot>,
        IEmit<GlobalDutyTemplateCreatedEvent>,
        IEmit<GlobalDutyTemplateDeletedEvent>,
        IEmit<GlobalDutyTemplateTitleModifiedEvent>,
        IEmit<GlobalDutyTemplateNoteModifiedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotAddedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotDeletedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotEndTimeModifiedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotNameModifiedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotNoteModifiedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotStartTimeModifiedEvent>,
        IEmit<GlobalDutyTemplateTimeSlotWeekdaysModifiedEvent>,
        IEmit<GlobalDutyTemplateModifiedUsageEvent>,
        IEmit<GlobalDutyTemplateSaveEvent>,
        IEmit<GlobalDutyTemplateCancelEvent>
    {
        private bool _created;
        private bool _deleted;
        public string Title { get; private set; } = "";
        public string Note { get; private set; } = "";
        public int Usage { get; private set; } = 0;
        public ICollection<TimeSlotTemplate> TimeSlotTemplates { get; private set; } = new List<TimeSlotTemplate>();
        public GlobalDutyTemplateModel Changes { get; private set; }

        public GlobalDutyTemplateAggregate(GlobalDutyTemplateId templateId) : base(templateId, SnapshotEveryFewVersionsStrategy.With(10))
        {

        }

        protected override Task<GlobalDutyTemplateSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new GlobalDutyTemplateSnapshot(_created, _deleted, Title, Note, Usage,
                Changes == null ? null
                : new GlobalDutyTemplateChangesReadModel(Changes), 
                TimeSlotTemplates == null 
                    ? new List<TimeSlotTemplateModel>() 
                    : TimeSlotTemplates.Select(x 
                    => new TimeSlotTemplateModel(x)).ToList()), cancellationToken);
        }

        protected override Task LoadSnapshotAsync(GlobalDutyTemplateSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                _created = snapshot.Created;
                _deleted = snapshot.Deleted;
                Title = snapshot.Title;
                Note = snapshot.Note;
                Usage = snapshot.Usage;
                TimeSlotTemplates = snapshot.TimeSlotTemplates == null 
                                        ? new List<TimeSlotTemplate>() 
                                        : snapshot.TimeSlotTemplates.Select(x => new TimeSlotTemplate(TimeSlotTemplateId.With(x.Id), x))
                    .ToList();
                Changes = snapshot.Changes == null ? null: new GlobalDutyTemplateModel(snapshot.Changes);
            });
        }

        #region IExcecutionResult

        #region GlobalDutyTemplate
        public IExecutionResult CreateGlobalDutyTemplate()
        {
            if (_created) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateCreatedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteGlobalDutyTemplate()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateDeletedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyTitleGlobalDutyTemplate(string title)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTitleModifiedEvent(title));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyNoteGlobalDutyTemplate(string note)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateNoteModifiedEvent(note));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyUsageGlobalDutyTemplate(bool increase)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateModifiedUsageEvent(increase));
            return ExecutionResult.Success();
        }

        public IExecutionResult SaveGlobalDutyTemplate()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateSaveEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult CancelGlobalDutyTemplate()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateCancelEvent());
            return ExecutionResult.Success();

        }

        #endregion
        #region TimeSlots

        public IExecutionResult AddTimeSlotToGlobalDutyTemplate(TimeSlotTemplateId id)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotAddedEvent(id));
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteTimeSlotFromGlobalDutyTemplate(TimeSlotTemplateId id)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotDeletedEvent(id));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyStartTimeTimeSlotGlobalDutyTemplate(TimeSlotTemplateId id,
            TimeSpan startTime)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotStartTimeModifiedEvent(id,startTime));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyEndTimeTimeSlotGlobalDutyTemplate(TimeSlotTemplateId id,
            TimeSpan endTime)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotEndTimeModifiedEvent(id, endTime));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyTimeSlotNameGlobalDutyTemplate(TimeSlotTemplateId id,
            string name)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotNameModifiedEvent(id, name));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyNoteTimeSlotGlobalDutyTemplate(TimeSlotTemplateId id,
            string note)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotNoteModifiedEvent(id, note));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyWeekdaysTimeSlotGlobalDutyTemplate(TimeSlotTemplateId id,
            ICollection<WeekDay> weekDays)
        {
            if (!CheckChanges(id)) return ExecutionResult.Failed();
            Emit(new GlobalDutyTemplateTimeSlotWeekdaysModifiedEvent(id, weekDays));
            return ExecutionResult.Success();
        }
        #endregion

        #endregion

        #region Apply

        #region GlobalDutyTemplate
        public void Apply(GlobalDutyTemplateCreatedEvent aggregateEvent)
        {
            _created = true;
            Changes = new GlobalDutyTemplateModel();
        }

        public void Apply(GlobalDutyTemplateDeletedEvent aggregateEvent)
        {
            _deleted = true;
        }

        public void Apply(GlobalDutyTemplateTitleModifiedEvent aggregateEvent)
        {
            if (Changes == null)
                Changes = new GlobalDutyTemplateModel(Title, Note, TimeSlotTemplates);

            Changes.Title = aggregateEvent.Title;
        }

        public void Apply(GlobalDutyTemplateNoteModifiedEvent aggregateEvent)
        {
            if (Changes == null)
                Changes = new GlobalDutyTemplateModel(Title, Note, TimeSlotTemplates);

            Changes.Note = aggregateEvent.Note;
        }

        public void Apply(GlobalDutyTemplateModifiedUsageEvent aggregateEvent)
        {
            var currentUsage = 0 + Usage;
            currentUsage = aggregateEvent.Increase ? currentUsage+= 1 : currentUsage-= 1;
            if (currentUsage < 0)
            {
                Usage = 0;
                return;
            }
            Usage = currentUsage;
        }


        public void Apply(GlobalDutyTemplateSaveEvent aggregateEvent)
        {
            if (Changes == null) return;
            Title = Changes.Title;
            Note = Changes.Note;
            TimeSlotTemplates = Changes.TimeSlotTemplates;
            Changes = null;

        }

        public void Apply(GlobalDutyTemplateCancelEvent aggregateEvent)
        {
            if (Changes == null) return;

            Changes = null;
        }


        #endregion
        #region TimeSlots
        public void Apply(GlobalDutyTemplateTimeSlotAddedEvent aggregateEvent)
        {
            if (Changes == null)
                Changes = new GlobalDutyTemplateModel(Title, Note, TimeSlotTemplates);

            if (Changes.TimeSlotTemplates == null)
                Changes.TimeSlotTemplates = new List<TimeSlotTemplate>();

            Changes.TimeSlotTemplates.Add(new TimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate));
        }

        public void Apply(GlobalDutyTemplateTimeSlotDeletedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            Changes.TimeSlotTemplates.Remove(timeSlotTemplate);

        }

        public void Apply(GlobalDutyTemplateTimeSlotEndTimeModifiedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            timeSlotTemplate.EndTime = aggregateEvent.EndTime;
        }

        public void Apply(GlobalDutyTemplateTimeSlotNameModifiedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            timeSlotTemplate.TimeSlotName = aggregateEvent.Name;
        }

        public void Apply(GlobalDutyTemplateTimeSlotNoteModifiedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            timeSlotTemplate.Note = aggregateEvent.Note;
        }

        public void Apply(GlobalDutyTemplateTimeSlotStartTimeModifiedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            timeSlotTemplate.StartTime = aggregateEvent.StartTime;
        }

        public void Apply(GlobalDutyTemplateTimeSlotWeekdaysModifiedEvent aggregateEvent)
        {
            var timeSlotTemplate = CheckChangesAndGetTimeSlotTemplate(aggregateEvent.IdTimeSlotTemplate);
            timeSlotTemplate.WeekDays = aggregateEvent.Weekdays;
        }



        #endregion
        #endregion

        private TimeSlotTemplate CheckChangesAndGetTimeSlotTemplate(
            TimeSlotTemplateId id)
        {
            if (Changes == null)
                Changes = new GlobalDutyTemplateModel(Title, Note, TimeSlotTemplates);
            var timeSlot = Changes.TimeSlotTemplates
                .Single(x => x.Id == id);
            return timeSlot;
        }

        private bool CheckChanges(TimeSlotTemplateId id)
        {
            if (!_created || _deleted) return false;

            if (Changes == null)
                Changes = new GlobalDutyTemplateModel(Title, Note, TimeSlotTemplates);

            if (Changes.TimeSlotTemplates.All(x => x.Id != id)) return false;

            return true;

        }


    }
}
