﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Employee.Entities;
using CRM.Domain.Employee.Events;
using CRM.Domain.Employee.Models;
using CRM.Domain.Employee.Snapshots;
using CRM.Domain.Shared.Enums;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Employee

{
    public class EmployeeAggregate :
        SnapshotAggregateRoot<EmployeeAggregate, EmployeeId, EmployeeSnapshot>,
        IEmit<ContactDetailAddedEvent>,
        IEmit<ContactDetailModifiedEvent>,
        IEmit<ContactDetailTypeModifiedEvent>,
        IEmit<ContactDetailRemovedEvent>,
        IEmit<EmployeeCreatedEvent>,
        IEmit<EmployeeModifiedEvent>,
        IEmit<TagAddedEvent>,
        IEmit<TagRemovedEvent>,
        IEmit<EmployeeDeletedEvent>,
        IEmit<EmployeeCanBeOnDutyModifiedEvent>,
        IEmit<EmployeeCustomerModifiedEvent>

    {
        private bool _created;
        private bool _deleted;

        public EmployeeAggregate(EmployeeId id) : base(id, SnapshotEveryFewVersionsStrategy.With(5))
        {
        }

        public Address EmployeeAddress { get; private set; }
        public bool SendMailWithTicket { get; private set; }

        public ICollection<EmployeeContactDetail> ContactDetails { get; private set; } =
            new List<EmployeeContactDetail>();

        public ICollection<string> Tags { get; private set; } = new List<string>();
        public ICollection<string> DefaultTags { get; private set; } = new List<string>();

        public string CCEmail { get; private set; }
        public string CustomerId { get; private set; }
        public string EmployeeComment { get; private set; }
        public string FirstName { get; private set; }
        public string LastName { get; private set; }
        public bool SetTicketAsRead { get; private set; }
        public bool CanBeOnDuty { get; private set; }


        #region IExecutionResult

        public IExecutionResult CreateEmployee(string customerId,
            string firstName,
            string lastName,
            string cCEmail = null,
            string employeeComment = null,
            bool sendMailWithTicket = false,
            bool setTicketAsRead = false,
            ICollection<EmployeeContactDetailModel> contactDetails = null,
            ICollection<string> tags = null,
            ICollection<string> defaultTags = null,
            bool canBeOnDuty = false,
            Address employeeAddress = null)
        {
            if (_created) return ExecutionResult.Failed();
            var _event = new EmployeeCreatedEvent(firstName, lastName, cCEmail, employeeComment, sendMailWithTicket,
                customerId, setTicketAsRead, contactDetails, tags, defaultTags, canBeOnDuty, employeeAddress);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteEmployee()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            Emit(new EmployeeDeletedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyEmployee(string firstName,
            string lastName,
            string cCEmail,
            string employeeComment,
            bool sendMailWithTicket,
            bool setTicketAsRead,
            string customerId,
            ICollection<EmployeeContactDetailModel> contactDetails,
            ICollection<string> tags, 
            ICollection<string> defaultTags,
            bool canBeOnDuty,
            Address employeeAddress)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            var _event = new EmployeeModifiedEvent(firstName, lastName, cCEmail, employeeComment, sendMailWithTicket,
                customerId, setTicketAsRead, contactDetails, tags, defaultTags, canBeOnDuty, employeeAddress);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddContactDetailToEmployee(ContactDetailType contactDetailType,
            bool isPrivate,
            bool isInternal,
            bool isDefault,
            string contactDetailName,
            string contactDetailValue,
            EmployeeContactDetailId id)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (ContactDetails.Any(x => x.Id.Equals(id))) return ExecutionResult.Failed();
            var _event = new ContactDetailAddedEvent(isPrivate, isInternal, isDefault, contactDetailType, contactDetailName,
                contactDetailValue, id);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ChangeContactDetail(EmployeeContactDetailId id,
            bool isPrivate,
            bool isInternal,
            bool isDefault,
            ContactDetailType contactDetailType,
            string contactDetailName,
            string contactDetailValue)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (!ContactDetails.Any(x => x.Id.Equals(id))) return ExecutionResult.Failed();
            var _event = new ContactDetailModifiedEvent(id, isPrivate, isInternal, isDefault, contactDetailType, contactDetailName,
                contactDetailValue);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyContactDetailType(EmployeeContactDetailId id,
            ContactDetailType contactDetailType)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (!ContactDetails.Any(x => x.Id.Equals(id))) return ExecutionResult.Failed();
            var _event = new ContactDetailTypeModifiedEvent(id, contactDetailType);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddTag(string tag)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (Tags.Any(x => x.Equals(tag))) return ExecutionResult.Success();
            var _event = new TagAddedEvent(tag);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveContactDetail(EmployeeContactDetailId id)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (!ContactDetails.Any(x => x.Id.Equals(id))) return ExecutionResult.Failed();
            var _event = new ContactDetailRemovedEvent(id);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveTag(string tag)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (!Tags.Any(x => x.Equals(tag))) return ExecutionResult.Failed();
            var _event = new TagRemovedEvent(tag);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyEmployeeCustomer(string customerId)
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            var _event = new EmployeeCustomerModifiedEvent(customerId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        #endregion IExecutionResult

        #region Apply

        public void Apply(EmployeeCreatedEvent aggregateEvent)
        {
            _created = true;
            FirstName = aggregateEvent.FirstName;
            LastName = aggregateEvent.LastName;
            CCEmail = aggregateEvent.CCEmail;
            EmployeeComment = aggregateEvent.EmployeeComment;
            SendMailWithTicket = aggregateEvent.SendMailWithTicket;
            CustomerId = aggregateEvent.CustomerId;
            SetTicketAsRead = aggregateEvent.SetTicketAsRead;
            ContactDetails = aggregateEvent.ContactDetails.Select(x =>
                new EmployeeContactDetail(EmployeeContactDetailId.With(x.Id), x.ContactDetailName, x.ContactDetailValue,
                    x.IsPrivate, x.IsInternal, x.IsDefault, x.ContactType)).ToList();
            Tags = aggregateEvent.Tags;
            DefaultTags = aggregateEvent.DefaultTags;
            CanBeOnDuty = aggregateEvent.CanBeOnDuty;
            EmployeeAddress = aggregateEvent.EmployeeAddress;
        }

        protected override Task<EmployeeSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new EmployeeSnapshot
            {
                CCEmail = CCEmail,
                CustomerId = CustomerId,
                ContactDetails = ContactDetails,
                Created = _created,
                Deleted = _deleted,
                EmployeeAddress = EmployeeAddress,
                EmployeeComment = EmployeeComment,
                FirstName = FirstName,
                LastName = LastName,
                SendMailWithTicket = SendMailWithTicket,
                Tags = Tags,
                DefaultTags = DefaultTags,
                SetTicketAsRead = SetTicketAsRead,
                CanBeOnDuty = CanBeOnDuty
            }, cancellationToken);
        }

        protected override Task LoadSnapshotAsync(EmployeeSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                CCEmail = snapshot.CCEmail;
                CustomerId = snapshot.CustomerId;
                ContactDetails = snapshot.ContactDetails;
                _created = snapshot.Created;
                _deleted = snapshot.Deleted;
                EmployeeAddress = snapshot.EmployeeAddress;
                EmployeeComment = snapshot.EmployeeComment;
                FirstName = snapshot.FirstName;
                LastName = snapshot.LastName;
                SendMailWithTicket = snapshot.SendMailWithTicket;
                SetTicketAsRead = snapshot.SetTicketAsRead;
                Tags = snapshot.Tags;
                DefaultTags = snapshot.DefaultTags;
                CanBeOnDuty = snapshot.CanBeOnDuty;
            }, cancellationToken);
        }

        public void Apply(EmployeeModifiedEvent aggregateEvent)
        {
            SendMailWithTicket = aggregateEvent.SendMailWithTicket;
            CCEmail = aggregateEvent.CCEmail;
            CustomerId = aggregateEvent.CustomerId;
            EmployeeComment = aggregateEvent.EmployeeComment;
            FirstName = aggregateEvent.FirstName;
            LastName = aggregateEvent.LastName;
            SetTicketAsRead = aggregateEvent.SetTicketAsRead;
            EmployeeAddress = aggregateEvent.EmployeeAddress;
            ContactDetails = new Collection<EmployeeContactDetail>();

            aggregateEvent.ContactDetails.ToList().ForEach(x =>
            {
                var contactDetail = new EmployeeContactDetail(new EmployeeContactDetailId(x.Id));
                contactDetail.ContactDetailName = x.ContactDetailName;
                contactDetail.ContactDetailValue = x.ContactDetailValue;
                contactDetail.ContactType = x.ContactType;
                contactDetail.IsPrivate = x.IsPrivate;
                ContactDetails.Add(contactDetail);
            });
        }

        public void Apply(ContactDetailAddedEvent aggregateEvent)
        {
            var contactDetail = new EmployeeContactDetail(aggregateEvent.Id);
            contactDetail.ContactDetailName = aggregateEvent.ContactDetailName;
            contactDetail.ContactDetailValue = aggregateEvent.ContactDetailValue;
            contactDetail.ContactType = aggregateEvent.ContactDetailType;
            contactDetail.IsPrivate = aggregateEvent.IsPrivate;
            contactDetail.IsInternal = aggregateEvent.IsInternal;
            ContactDetails.Add(contactDetail);
        }

        public void Apply(TagRemovedEvent aggregateEvent)
        {
            Tags.Remove(aggregateEvent.Tag);
        }

        public void Apply(TagAddedEvent aggregateEvent)
        {
            Tags.Add(aggregateEvent.Tag);
        }

        public void Apply(ContactDetailRemovedEvent aggregateEvent)
        {
            ContactDetails.Remove(ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id)));
        }

        public void Apply(ContactDetailModifiedEvent aggregateEvent)
        {
            //todo: Hier ook IsDefault toevoegen
            var contactDetail = ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id));
            contactDetail.ContactDetailName = aggregateEvent.ContactDetailName;
            contactDetail.ContactDetailValue = aggregateEvent.ContactDetailValue;
            contactDetail.ContactType = aggregateEvent.ContactDetailType;
            contactDetail.IsPrivate = aggregateEvent.IsPrivate;
            contactDetail.IsInternal = aggregateEvent.IsInternal;
            contactDetail.IsDefault = aggregateEvent.IsDefault;
        }

        public void Apply(ContactDetailTypeModifiedEvent aggregateEvent)
        {
            var contactDetail = ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id));
            contactDetail.ContactType = aggregateEvent.ContactDetailType;
        }


        public void Apply(EmployeeDeletedEvent aggregateEvent)
        {
            _created = true;
        }

        public void Apply(EmployeeCanBeOnDutyModifiedEvent aggregateEvent)
        {
            CanBeOnDuty = aggregateEvent.CanBeOnDuty;
        }

        #endregion Apply

        public void Apply(EmployeeCustomerModifiedEvent aggregateEvent)
        {
            CustomerId = aggregateEvent.CustomerId;
        }
    }
}