﻿using System;
using System.Collections.Generic;
using System.Linq;
using CRM.Domain.Employee.Events;
using CRM.Domain.Employee.Models;
using EventFlow.Aggregates;
using EventFlow.MongoDB.ReadStores;
using EventFlow.ReadStores;
using MongoDB.Bson.Serialization.Attributes;

namespace CRM.Domain.Employee.ReadModels
{
    public class EmployeeReadModel : IMongoDbReadModel,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, EmployeeCreatedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, ContactDetailAddedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, ContactDetailModifiedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, ContactDetailRemovedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, EmployeeDeletedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, EmployeeModifiedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, TagAddedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, TagRemovedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, EmployeeCanBeOnDutyModifiedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, ContactDetailTypeModifiedEvent>,
        IAmReadModelFor<EmployeeAggregate, EmployeeId, EmployeeCustomerModifiedEvent>
    {
        public string Id { get; set; }
        public long? Version { get; set; }

        public bool Deleted { get; private set; }
        public Address EmployeeAddress { get; private set; }
        public bool SendMailWithTicket { get; private set; }

        public ICollection<EmployeeContactDetailModel> ContactDetails { get; private set; } =
            new List<EmployeeContactDetailModel>();

        public ICollection<string> Tags { get; private set; } = new List<string>();
        public ICollection<string> DefaultTags { get; private set; } = new List<string>();
        public string CcEmail { get; private set; }
        public string CustomerId { get; private set; }
        public string EmployeeComment { get; private set; }

        private string _firstName;
        public string FirstName
        {
            get => _firstName.Trim();
            private set => _firstName = value.Trim();
        }

        private string _lastName;
        public string LastName
        {
            get => _lastName.Trim(); 
            private set => _lastName = value.Trim(); 
        }

        public bool SetTicketAsRead { get; private set; }
        public bool CanBeOnDuty { get; set; }

        [BsonIgnore] public string SearchName => GetSearchName();

        [BsonIgnore] public string FullName => GetFullName();

        private string GetFullName()
        {
            return $"{FirstName} {LastName.ToUpper()}";
        }

        private string GetSearchName()
        {
            return $"{FirstName} {LastName} {FirstName}";
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, ContactDetailAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            var contactDetail = new EmployeeContactDetailModel(aggregateEvent.Id.Value,
                aggregateEvent.ContactDetailName, aggregateEvent.ContactDetailValue, aggregateEvent.IsPrivate,
                aggregateEvent.IsInternal, aggregateEvent.IsDefault,
                aggregateEvent.ContactDetailType)
            {
                ContactDetailName = aggregateEvent.ContactDetailName,
                ContactDetailValue = aggregateEvent.ContactDetailValue,
                ContactType = aggregateEvent.ContactDetailType,
                IsPrivate = aggregateEvent.IsPrivate,
                IsInternal = aggregateEvent.IsInternal,
                IsDefault = aggregateEvent.IsDefault
            };
            ContactDetails.Add(contactDetail);
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, ContactDetailModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            var contactDetail = ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id.ToString()));
            contactDetail.ContactDetailName = aggregateEvent.ContactDetailName;
            contactDetail.ContactDetailValue = aggregateEvent.ContactDetailValue;
            contactDetail.ContactType = aggregateEvent.ContactDetailType;
            contactDetail.IsPrivate = aggregateEvent.IsPrivate;
            contactDetail.IsDefault = aggregateEvent.IsDefault;
            contactDetail.IsInternal = aggregateEvent.IsInternal;
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, ContactDetailRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            ContactDetails.Remove(ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id.ToString())));
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, EmployeeCanBeOnDutyModifiedEvent> domainEvent)
        {
            CanBeOnDuty = domainEvent.AggregateEvent.CanBeOnDuty;
        }


        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, EmployeeCreatedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            FirstName = aggregateEvent.FirstName;
            LastName = aggregateEvent.LastName;
            CcEmail = aggregateEvent.CCEmail;
            EmployeeComment = aggregateEvent.EmployeeComment;
            SendMailWithTicket = aggregateEvent.SendMailWithTicket;
            CustomerId = aggregateEvent.CustomerId;
            SetTicketAsRead = aggregateEvent.SetTicketAsRead;
            Id = domainEvent.AggregateIdentity.Value;
            ContactDetails = aggregateEvent.ContactDetails;
            Tags = aggregateEvent.Tags;
            DefaultTags = aggregateEvent.DefaultTags;
            EmployeeAddress = aggregateEvent.EmployeeAddress;
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, EmployeeDeletedEvent> domainEvent)
        {
            Deleted = true;
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, EmployeeModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            SendMailWithTicket = aggregateEvent.SendMailWithTicket;
            CcEmail = aggregateEvent.CCEmail;
            CustomerId = aggregateEvent.CustomerId;
            EmployeeComment = aggregateEvent.EmployeeComment;
            FirstName = aggregateEvent.FirstName;
            LastName = aggregateEvent.LastName;
            SetTicketAsRead = aggregateEvent.SetTicketAsRead;
            EmployeeAddress = aggregateEvent.EmployeeAddress;
            ContactDetails = aggregateEvent.ContactDetails;
            CanBeOnDuty = aggregateEvent.CanBeOnDuty;
            Tags = aggregateEvent.Tags;
            DefaultTags = aggregateEvent.DefaultTags;
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, TagAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            Tags.Add(aggregateEvent.Tag);
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, TagRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            Tags.Remove(aggregateEvent.Tag);
        }


        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, ContactDetailTypeModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            var contactDetail = ContactDetails.Single(x => x.Id.Equals(aggregateEvent.Id.ToString()));
            contactDetail.ContactType = aggregateEvent.ContactDetailType;
        }

        public void Apply(IReadModelContext context,
            IDomainEvent<EmployeeAggregate, EmployeeId, EmployeeCustomerModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            CustomerId = aggregateEvent.CustomerId;
        }
    }
}