﻿using System;
using CRM.Domain.InteractionStatistic.Enums;
using CRM.Domain.InteractionStatistic.Events;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;

namespace CRM.Domain.InteractionStatistic
{
    public class InteractionStatisticAggregate : AggregateRoot<InteractionStatisticAggregate, InteractionStatisticId>,
        IEmit<InteractionStatisticCreatedEvent>,
        IEmit<InteractionStatisticDeletedEvent>,
        IEmit<InteractionStatisticModifiedDossierIdEvent>,
        IEmit<InteractionStatisticModifiedTicketIdEvent>
    {
        public InteractionStatisticAggregate(InteractionStatisticId id) : base(id)
        {
        }

        private bool _created { get; set; }
        public bool Deleted { get; private set; }

        public string InteractionIdKey { get; private set; }
        public string DossierId { get; private set; }
        public int DNIS { get; private set; }
        public DateTime? Established { get; private set; }
        public DateTime Connected { get; private set; }
        public TimeSpan Duration { get; private set; }
        public DateTime Disconnected { get; private set; }
        public InteractionType InteractionTypeEnum { get; private set; }
        public int? OriginalCcmMessageId { get; private set; }
        public string TicketId { get; private set; }

        #region Apply
        public void Apply(InteractionStatisticCreatedEvent aggregateEvent)
        {
            InteractionIdKey = aggregateEvent.InteractionIdKey;
            DossierId = aggregateEvent.DossierId;
            DNIS = aggregateEvent.DNIS;
            Established = aggregateEvent.Established;
            Connected = aggregateEvent.Connected;
            Duration = aggregateEvent.Duration;
            Disconnected = aggregateEvent.Disconnected;
            InteractionTypeEnum = aggregateEvent.InteractionTypeEnum;
            OriginalCcmMessageId = aggregateEvent.OriginalCcmMessageId;
            _created = true;
        }

        public void Apply(InteractionStatisticDeletedEvent aggregateEvent)
        {
            Deleted = true;
        }
        public void Apply(InteractionStatisticModifiedDossierIdEvent aggregateEvent)
        {
            DossierId = aggregateEvent.DossierId;
        }

        public void Apply(InteractionStatisticModifiedTicketIdEvent aggregateEvent)
        {
            TicketId = aggregateEvent.TicketId;
        }
        #endregion

        #region IExecutionResult
        public IExecutionResult DeleteInteractionStat()
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Statistic with id {Id.Value} does not exist");

            var _event = new InteractionStatisticDeletedEvent();
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult CreateInteractionStat(string interactionIdKey,
            string dossierId,
            int dNIS,
            DateTime? established,
            DateTime connected,
            TimeSpan duration,
            DateTime disconnected,
            InteractionType interactionTypeEnum,
            int? originalCcmMessageId,
            string ticketId)
        {
            if (_created) return ExecutionResult.Failed();
            var _event = new InteractionStatisticCreatedEvent(interactionIdKey, dossierId, dNIS, established, connected,
                duration, disconnected, interactionTypeEnum, originalCcmMessageId, ticketId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyInteractionStatDossierId(string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Statistic with id {Id.Value} does not exist");

            var _event = new InteractionStatisticModifiedDossierIdEvent(dossierId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyInteracationStatTicketId(string ticketId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Statistic with id {Id.Value} does not exist");

            var _event = new InteractionStatisticModifiedTicketIdEvent(ticketId);
            Emit(_event);
            return ExecutionResult.Success();
        }
        #endregion

    }
}