﻿using CRM.Domain.NeoDuty.Entities;
using CRM.Domain.NeoDuty.Events;
using CRM.Domain.NeoDuty.Events.Timeslot;
using CRM.Domain.NeoDuty.Events.Timeslot.Contact;
using CRM.Domain.NeoDuty.Models;
using CRM.Domain.NeoDuty.Snapshots;
using CRM.Domain.Shared.ValueObjects;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace CRM.Domain.NeoDuty
{
    public class NeoDutyAggregate : SnapshotAggregateRoot<NeoDutyAggregate, NeoDutyId, NeoDutySnapshot>,
        IEmit<NeoDutyCreatedEvent>,
        IEmit<NeoDutySavedEvent>,
        IEmit<NeoDutyCanceledEvent>,
        IEmit<NeoDutyNoteModifiedEvent>,
        IEmit<NeoDutyStartModifiedEvent>,
        IEmit<NeoDutyEndModifiedEvent>,
        IEmit<NeoDutyStartEndModifiedEvent>,
        IEmit<NeoTimeSlotAddedEvent>,
        IEmit<NeoTimeSlotRemovedEvent>,
        IEmit<NeoWeekdaysModifiedEvent>,
        IEmit<NeoTimeSlotTitleModifiedEvent>,
        IEmit<NeoTimeSlotStartTimeModifiedEvent>,
        IEmit<NeoTimeSlotEndTimeModifiedEvent>,
        IEmit<NeoTimeSlotNoteModifiedEvent>,
        IEmit<NeoContactAddedEvent>,
        IEmit<NeoContactAddedToDutyEvent>,
        IEmit<NeoContactRemovedEvent>,
        IEmit<NeoContactOrderIdModifiedEvent>,
        IEmit<NeoContactPointAddedEvent>,
        IEmit<NeoContactPointRemovedEvent>,
        IEmit<NeoTagAddedEvent>,
        IEmit<NeoTagRemovedEvent>,
        IEmit<NeoDutyDeletedEvent>
    {
        public NeoDutyAggregate(NeoDutyId id) : base(id, SnapshotEveryFewVersionsStrategy.With(20)) { }

        private bool _created;
        private bool _deleted;
        private bool _justEmited = false;
        public string DutyCategoryId { get; private set; } = string.Empty;
        public string Note { get; private set; } = string.Empty;
        public DateTime Start { get; private set; } = DateTime.Now;
        public DateTime End { get; private set; } = DateTime.Now;
        public bool SavedOnce { get; private set; }
        public bool AutomaticallyDestroy { get; private set; }
        public ICollection<TimeSlot> TimeSlots { get; private set; } = new List<TimeSlot>();
        public NeoDutyModel Changes { get; private set; } = new NeoDutyModel();


        protected override Task<NeoDutySnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new NeoDutySnapshot(_created, _deleted, DutyCategoryId, Note, Start, End, SavedOnce, AutomaticallyDestroy,
                Changes == null
                    ? null
                    : new NeoDutyChangesReadModel(Changes, DutyCategoryId, SavedOnce),
                TimeSlots == null
                    ? new List<TimeSlotModel>()
                    : TimeSlots.Select(x =>
                        new TimeSlotModel(x)).ToList()), cancellationToken);
        }

        protected override Task LoadSnapshotAsync(NeoDutySnapshot snapshot, ISnapshotMetadata metadata, CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                _created = snapshot.Created;
                _deleted = snapshot.Deleted;
                DutyCategoryId = snapshot.DutyCategoryId;
                Note = snapshot.Note;
                Start = snapshot.Start;
                End = snapshot.End;
                SavedOnce = snapshot.SavedOnce;
                AutomaticallyDestroy = snapshot.AutomaticallyDestroy;
                Changes = snapshot.Changes == null
                    ? null
                    : new NeoDutyModel(snapshot.Changes);
                TimeSlots = snapshot.TimeSlots == null
                    ? new List<TimeSlot>()
                    : snapshot.TimeSlots.Select(x =>
                        new TimeSlot(TimeSlotId.With(x.Id), x)).ToList();
            });
        }
        #region IExecutionResult
        #region NeoDuty
        public IExecutionResult CreateNeoDuty(string dutyCategoryId, string note, List<AddTimeslotModel> timeSlots)
        {
            if (_created) return ExecutionResult.Failed();
            Emit(new NeoDutyCreatedEvent(dutyCategoryId, note, timeSlots));
            return ExecutionResult.Success();
        }

        public IExecutionResult SaveNeoDuty()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (Changes == null) return ExecutionResult.Success();
            Emit(new NeoDutySavedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult CancelNeoDuty()
        {
            if (!_created || _deleted) return ExecutionResult.Failed();
            if (Changes == null) return ExecutionResult.Success();
            Emit(new NeoDutyCanceledEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteNeoDuty()
        {
            if (_deleted || !_created) return ExecutionResult.Failed();
            Emit(new NeoDutyDeletedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyNote(string note)
        {
            if (!CheckChanges()) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoDutyNoteModifiedEvent(note));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyStart (DateTime start)
        {
            if (!CheckChanges()) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoDutyStartModifiedEvent(start));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyEnd(DateTime end)
        {
            if (!CheckChanges()) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoDutyEndModifiedEvent(end));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyStartEnd(DateTime start, DateTime end)
        {
            if (!CheckChanges()) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoDutyStartEndModifiedEvent(end, start));
            return ExecutionResult.Success();
        }


        #endregion

        #region TimeSlot
        public IExecutionResult AddTimeSlot(TimeSlotId timeSlotId)
        {
            if (!CheckChanges()) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTimeSlotAddedEvent(timeSlotId));
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveTimeSlot(TimeSlotId timeSlotId)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTimeSlotRemovedEvent(timeSlotId));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyWeekDays(TimeSlotId timeSlotId, ICollection<WeekDay> weekDays)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoWeekdaysModifiedEvent(timeSlotId, weekDays));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyStartTimeTimeSlot(TimeSlotId timeSlotId, TimeSpan startTime)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTimeSlotStartTimeModifiedEvent(timeSlotId, startTime));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyEndTimeTimeSlot(TimeSlotId timeSlotId, TimeSpan endTime)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTimeSlotEndTimeModifiedEvent(timeSlotId, endTime));
            return ExecutionResult.Success();
        }


        public IExecutionResult ModifyTitleTimeSlot(TimeSlotId timeSlotId, string title)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTimeSlotTitleModifiedEvent(timeSlotId, title));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyNoteTimeSlot(TimeSlotId timeSlotId, string note)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited= true;
            Emit(new NeoTimeSlotNoteModifiedEvent(timeSlotId, note));
            return ExecutionResult.Success();
        }


        #endregion

        #region DutyContact
        public IExecutionResult AddContactToTimeSlot(TimeSlotId timeSlotId, DutyContactId dutyContactId, string employeeId, List<string> defaultTags, List<DutyContactPointModel> defaultContactPointIds)
        {
            if (!CheckChanges(timeSlotId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoContactAddedEvent(timeSlotId, dutyContactId, employeeId, defaultTags, defaultContactPointIds));
            return ExecutionResult.Success();

        }

        public IExecutionResult AddContactToDuty(string employeeId, List<string> defaultTags, List<AddContactToNeoDutyModel> contactList)
        {
            foreach (var contact in contactList)
            {
                if (!CheckChanges(contact.TimeSlotId)) return ExecutionResult.Failed();
            }
            _justEmited = true;
            Emit(new NeoContactAddedToDutyEvent(employeeId, defaultTags, contactList));
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveContactFromDuty(string employeeId, string timeSlotId = null)
        {
            if (timeSlotId == null)
            {
                if (!CheckChanges()) return ExecutionResult.Failed();

            }
            else
            {
                if (!CheckChanges(TimeSlotId.With(timeSlotId), employeeId)) return ExecutionResult.Failed();
            }
            _justEmited = true;
            Emit(new NeoContactRemovedEvent(employeeId, timeSlotId));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyContactOrder(TimeSlotId timeSlotId, string employeeId, int newOrderId)
        {
            if (!CheckChanges(timeSlotId, employeeId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoContactOrderIdModifiedEvent(timeSlotId, employeeId, newOrderId));
            return ExecutionResult.Success();

        }

        public IExecutionResult AddContactPoint(TimeSlotId timeSlotId, DutyContactId dutyContactId, DutyContactPointId dutyContactPointId, string contactDetailId)
        {
            if (!CheckChangesDetails(timeSlotId, dutyContactId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoContactPointAddedEvent(timeSlotId, dutyContactId, dutyContactPointId, contactDetailId));
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveContactPoint(TimeSlotId timeSlotId, DutyContactId dutyContactId, DutyContactPointId dutyContactPointId)
        {
            if (!CheckChangesDetails(timeSlotId, dutyContactId, dutyContactPointId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoContactPointRemovedEvent(timeSlotId, dutyContactId, dutyContactPointId));
            return ExecutionResult.Success();
        }

        public IExecutionResult AddTag(TimeSlotId timeSlotId, DutyContactId dutyContactId, string tag)
        {
            if (!CheckChangesDetails(timeSlotId, dutyContactId)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTagAddedEvent(timeSlotId, dutyContactId, tag));
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveTag(TimeSlotId timeSlotId, DutyContactId dutyContactId, string tag)
        {
            if (!CheckChangesDetails(timeSlotId, dutyContactId, tag:tag)) return ExecutionResult.Failed();
            _justEmited = true;
            Emit(new NeoTagRemovedEvent(timeSlotId, dutyContactId, tag));
            return ExecutionResult.Success();
        }
        #endregion
        #endregion
        #region apply
        #region NeoDuty
        public void Apply(NeoDutyCreatedEvent aggregateEvent)
        {
            _created = true;
            SavedOnce = false;
            DutyCategoryId = aggregateEvent.DutyCategoryId;
            Changes = new NeoDutyModel();
            Changes.Note = aggregateEvent.Note;
            SetDefaultStartAndEnd();
            Changes.AutomaticallyDestroy = false;
            Changes.TimeSlots = new List<TimeSlot>(aggregateEvent.TimeSlots.Select(x
                => new TimeSlot(TimeSlotId.With(x.Id), x)));
        }

        private void SetDefaultStartAndEnd()
        {
            var monDay = (int) DayOfWeek.Monday;
            var todaysWeekDay = (int) DateTime.Now.DayOfWeek;
            int daysToAdd;
            if (todaysWeekDay >= monDay)
            {
                daysToAdd = 8 - todaysWeekDay;
            }
            else
            {
                daysToAdd = 1;
            }

            Changes.Start = DateTime.Now.AddDays(daysToAdd).Date;
            Changes.End = Changes.Start.AddDays(7);
        }

        public void Apply(NeoDutySavedEvent aggregateEvent)
        {
            if (Changes == null) return;

            SavedOnce = true;
            Note = Changes.Note;
            Start = Changes.Start;
            End = Changes.End;
            AutomaticallyDestroy = Changes.AutomaticallyDestroy;
            TimeSlots = Changes.TimeSlots;

            Changes = null;
        }
        public void Apply(NeoDutyCanceledEvent aggregateEvent)
        {
            if (! SavedOnce)
            {
                _deleted = true;
                return;
            }
            if (Changes == null) return;

            Changes = null;
        }

        public void Apply(NeoDutyDeletedEvent aggregateEvent)
        {
            _deleted = true;
        }


        public void Apply(NeoDutyNoteModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            Changes.Note= aggregateEvent.Note;
            _justEmited = false;

        }

        public void Apply(NeoDutyStartModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            Changes.Start = aggregateEvent.Start;
            _justEmited = false;

        }

        public void Apply(NeoDutyEndModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            Changes.End = aggregateEvent.End;
            _justEmited = false;

        }

        public void Apply(NeoDutyStartEndModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            Changes.Start = aggregateEvent.Start;
            Changes.End = aggregateEvent.End;
            _justEmited = false;

        }


        #endregion

        #region timeSlot

        public void Apply(NeoTimeSlotAddedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = new TimeSlot(aggregateEvent.TimeslotId);
            Changes.TimeSlots.Add(ts);
            _justEmited = false;
        }

        public void Apply(NeoTimeSlotRemovedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            Changes.TimeSlots.Remove(ts);
            _justEmited = false;
        }

        public void Apply(NeoWeekdaysModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            ts.WeekDays = aggregateEvent.WeekDays;
            _justEmited = false;
        }



        public void Apply(NeoTimeSlotEndTimeModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            ts.EndTime = aggregateEvent.EndTime;
            _justEmited = false;

        }

        public void Apply(NeoTimeSlotStartTimeModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            ts.StartTime = aggregateEvent.StartTime;
            _justEmited = false;

        }


        public void Apply(NeoTimeSlotTitleModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            ts.TimeSlotName = aggregateEvent.Title;
            _justEmited = false;
        }

        public void Apply(NeoTimeSlotNoteModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            ts.Note = aggregateEvent.Note;
            _justEmited = false;
        }
        #endregion

        #region DutyContact
        public void Apply(NeoContactAddedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeSlotId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            AddOneEmployeeToTimeSlot(aggregateEvent.TimeSlotId, aggregateEvent.ContactId, aggregateEvent.EmployeeId, aggregateEvent.DefaultContactPoints, aggregateEvent.DefaultTags);
            _justEmited = false;

        }

        public void Apply(NeoContactAddedToDutyEvent aggregateEvent)
        {
            foreach (var contact in aggregateEvent.ContactList)
            {
                if (!_justEmited)
                {
                    if (!CheckChanges(contact.TimeSlotId))
                    {
                        throw new InvalidOperationException("Something's wrong with your changes-object.");
                    }
                }
                AddOneEmployeeToTimeSlot(contact.TimeSlotId, contact.ContactId, aggregateEvent.EmployeeId, contact.DefaultContactPoints, aggregateEvent.DefaultTags);
            }
            _justEmited = false;

        }

        public void Apply(NeoContactRemovedEvent aggregateEvent)
        {
            if (string.IsNullOrEmpty(aggregateEvent.TimeSlotId))
            {
                if (!_justEmited)
                {
                    if (!CheckChanges())
                    {
                        throw new InvalidOperationException("Something's wrong with your changes-object.");
                    }
                }

                foreach (var ts in Changes.TimeSlots)
                {
                    if (!CheckChanges(ts.Id, aggregateEvent.EmployeeId))
                    {
                        continue;
                    }
                    RemoveOneEmployeeFromTimeSlot(ts.Id, aggregateEvent.EmployeeId);
                }
                _justEmited= false;
                return;
            }

            if (!_justEmited)
            {
                if (!CheckChanges(TimeSlotId.With(aggregateEvent.TimeSlotId)))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }

            RemoveOneEmployeeFromTimeSlot(TimeSlotId.With(aggregateEvent.TimeSlotId), aggregateEvent.EmployeeId);

            _justEmited = false;
        }

        public void Apply(NeoContactOrderIdModifiedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChanges(aggregateEvent.TimeslotId, aggregateEvent.EmployeeId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            var contact = ts.DutyContacts.Single(x => x.EmployeeId == aggregateEvent.EmployeeId);
            ReOrderContactsInTimeslot(ts, contact, aggregateEvent.NewOrderId);

            _justEmited = false;

        }

        public void Apply(NeoTagRemovedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChangesDetails(aggregateEvent.TimeslotId, aggregateEvent.ContactId, tag: aggregateEvent.Tag))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId);
            contact.Tags.Remove(aggregateEvent.Tag);
            _justEmited = false;
        }

        public void Apply(NeoTagAddedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChangesDetails(aggregateEvent.TimeslotId, aggregateEvent.ContactId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId);
            contact.Tags.Add(aggregateEvent.Tag);
            _justEmited = false;
        }

        public void Apply(NeoContactPointRemovedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChangesDetails(aggregateEvent.TimeslotId, aggregateEvent.ContactId, dutyContactPointId: aggregateEvent.ContactPointId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId);
            var cp = contact.ContactPoints.Single(x => x.Id == aggregateEvent.ContactPointId);
            contact.ContactPoints.Remove(cp);
            _justEmited = false;
        }

        public void Apply(NeoContactPointAddedEvent aggregateEvent)
        {
            if (!_justEmited)
            {
                if (!CheckChangesDetails(aggregateEvent.TimeslotId, aggregateEvent.ContactId))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId);
            var cp = new DutyContactPoint(aggregateEvent.ContactPointId, "", aggregateEvent.ContactDetailId);
            contact.ContactPoints.Add(cp);
            _justEmited = false;
        }



        #endregion
        #endregion

        #region helpertjes
        private bool CheckChanges(TimeSlotId timeSlotId = null, string employeeId = null)
        {
            if (!_created || _deleted) return false;
            if (Changes == null)
                Changes = new NeoDutyModel()
                {
                    AutomaticallyDestroy = AutomaticallyDestroy,
                    End = End,
                    Note = Note,
                    Start = Start,
                    TimeSlots = TimeSlots == null 
                        ? new List<TimeSlot>()
                        : TimeSlots.Select(x => new TimeSlot(x)).ToList()
                };
            if (timeSlotId == null) return true;

            var rtn = false;

            var timeSlot = Changes.TimeSlots.SingleOrDefault(x =>x.Id == timeSlotId);

            if (timeSlot == null) return false;

            if (employeeId == null) return true;

            var dutyContact = timeSlot.DutyContacts.SingleOrDefault(x => x.EmployeeId == employeeId);

            if (dutyContact == null) return false;

            return true;

        }

        private bool CheckChangesDetails(TimeSlotId timeSlotId, DutyContactId dutyContactId, DutyContactPointId dutyContactPointId = null, string tag = null)
        {
            if (!_created || _deleted) return false;
            if (Changes == null)
                Changes = new NeoDutyModel()
                {
                    AutomaticallyDestroy = AutomaticallyDestroy,
                    End = End,
                    Note = Note,
                    Start = Start,
                    TimeSlots = TimeSlots == null
                        ? new List<TimeSlot>()
                        : TimeSlots.Select(x => new TimeSlot(x)).ToList()
                };

            var rtn = false;

            var timeSlot = Changes.TimeSlots.SingleOrDefault(x => x.Id == timeSlotId);

            if (timeSlot == null) return false;

            var dutyContact = timeSlot.DutyContacts.SingleOrDefault(x => x.Id == dutyContactId);

            if (dutyContact == null) return false;

            if (dutyContactPointId == null && tag == null) return true;

            if (dutyContactPointId != null)
            {
                var dutyContactPoint = dutyContact.ContactPoints.SingleOrDefault(x => x.Id == dutyContactPointId);

                return dutyContactPoint != null;
            }
            else
            {
                var myTag = dutyContact.Tags.SingleOrDefault(x => x == tag);

                return myTag != null;
            }
        }

        private void AddOneEmployeeToTimeSlot(TimeSlotId timeSlotId, DutyContactId contactId, string employeeId, List<DutyContactPointModel> defaultContactPoints, List<string> defaultTags)
        {
            var ts = Changes.TimeSlots.Single(x => x.Id == timeSlotId);

            if (ts.DutyContacts != null)
            {
                var alReadyExists = ts.DutyContacts.SingleOrDefault(x => x.EmployeeId == employeeId);

                if (alReadyExists != null) return;

            }
            var orderId = ts.DutyContacts == null ? 1: ts.DutyContacts.Count + 1;
            DutyContact dc = new DutyContact(contactId, employeeId, orderId);
            dc.ContactPoints = new List<DutyContactPoint>();
            dc.Tags = new List<string>();


            foreach (var cp in defaultContactPoints)
            {
                dc.ContactPoints.Add(new DutyContactPoint(DutyContactPointId.With(cp.Id), cp));
            }
            foreach (var tag in defaultTags)
            {
                dc.Tags.Add(tag);
            }
            if (ts.DutyContacts == null)
                ts.DutyContacts = new List<DutyContact>();

            ts.DutyContacts.Add(dc);
        }

        private void RemoveOneEmployeeFromTimeSlot(TimeSlotId timeSlotId, string employeeId)
        {
            var ts = Changes.TimeSlots.Single(x => x.Id == timeSlotId);

            var contact = ts.DutyContacts.Single(x => x.EmployeeId == employeeId);

            ReOrderContactsInTimeslot(ts, contact);
            ts.DutyContacts.Remove(contact);
        }

        private static void ReOrderContactsInTimeslot(TimeSlot ts, DutyContact dc, int newOrderId = -1)
        {
            if (ts.DutyContacts.Count == 1) return;

            if (newOrderId == -1 && dc.OrderId == ts.DutyContacts.Count) return;

            if (newOrderId == -1)
            {
                ts.DutyContacts.Where(x => x.OrderId > dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId--);
                return;
            }

            if (dc.OrderId == newOrderId) return;

            if (dc.OrderId < newOrderId)
            {
                ts.DutyContacts.Where(x => x.OrderId <= newOrderId && x.OrderId > dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId--);
            }
            else
            {
                ts.DutyContacts.Where(x => x.OrderId >= newOrderId && x.OrderId < dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId++);
            }
            dc.OrderId = newOrderId;
        }









        #endregion
    }
}
