﻿using CRM.Domain.NeoDuty.Abstract;
using CRM.Domain.NeoDuty.Enumerations;
using CRM.Domain.NeoDuty.Events;
using CRM.Domain.NeoDuty.Events.Timeslot;
using CRM.Domain.NeoDuty.Events.Timeslot.Contact;
using CRM.Domain.NeoDuty.Models;
using EventFlow.Aggregates;
using EventFlow.ReadStores;
using MongoDB.Bson.Serialization.Attributes;
using System;
using System.Collections.Generic;
using System.Linq;

namespace CRM.Domain.NeoDuty.ReadModels
{
    public class NeoDutyReadModel : ReadModel,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyCreatedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutySavedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyCanceledEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyDeletedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyNoteModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyStartModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyEndModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoDutyStartEndModifiedEvent>,

        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotAddedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotRemovedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoWeekdaysModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotTitleModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotNoteModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotStartTimeModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTimeSlotEndTimeModifiedEvent>,

        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactAddedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactAddedToDutyEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactRemovedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactOrderIdModifiedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactPointAddedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoContactPointRemovedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTagAddedEvent>,
        IAmReadModelFor<NeoDutyAggregate, NeoDutyId, NeoTagRemovedEvent>

    {
        public string DutyCategoryId { get; private set; } = string.Empty;
        public string Note { get; private set; } = string.Empty;
        public DateTime Start { get; private set; } = DateTime.Now;
        public DateTime End { get; private set; } = DateTime.Now;
        public bool SavedOnce { get; private set; }
        public bool AutomaticallyDestroy { get; private set; }
        public ICollection<TimeSlotModel> TimeSlots { get; private set; } = new List<TimeSlotModel>();
        public NeoDutyChangesReadModel Changes { get; private set; } = new NeoDutyChangesReadModel();

        public string LastModifiedId { get; private set; }
        public string CreatedId { get; private set; }
        public string LastModifiedName { get; private set; }
        public string CreatedName { get; set; }
        public DateTime LastModifiedOn { get; private set; }
        public DateTime CreatedOn { get; private set; }

        public bool Deleted { get; private set; }

        //Extra props, niet voor DB
        [BsonIgnore] public NeoDutyStatus Status => CheckStatus();
        [BsonIgnore] public bool IsActiveNow => CheckIfActive();
        [BsonIgnore] public List<string> EmployeeIds => GetEmployeeIds();

        private List<string> GetEmployeeIds()
        {
            if (TimeSlots == null || TimeSlots.Count == 0) return null;

            var rtn = TimeSlots
                .SelectMany(x => x.DutyContacts)
                .Select(n => n.EmployeeId).ToList();
            var rtnDit = rtn.Distinct();
            return rtnDit.ToList();
        }

        private bool CheckIfActive()
        {
            return true;
        }

        private NeoDutyStatus CheckStatus()
        {
            return NeoDutyStatus.Active;
        }

        #region NeoDuty
        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyCreatedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            Id = domainEvent.AggregateIdentity.Value;
            SavedOnce = false;
            DutyCategoryId = aggregateEvent.DutyCategoryId;
            Changes = new NeoDutyChangesReadModel();
            Changes.SavedOnce = false;
            Changes.DutyCategoryId = aggregateEvent.DutyCategoryId;
            Changes.Note = aggregateEvent.Note;
            SetDefaultStartAndEnd();            
            Changes.AutomaticallyDestroy = true;
            Changes.TimeSlots = new List<TimeSlotModel>(aggregateEvent.TimeSlots.Select(x
                => new TimeSlotModel(x)));

            CreatedOn = DateTime.UtcNow;
            LastModifiedOn = DateTime.UtcNow;

            try
            {
                CreatedId = domainEvent.Metadata["userId"];
                LastModifiedId = domainEvent.Metadata["userId"];
                CreatedName = domainEvent.Metadata["userName"];
                LastModifiedName = domainEvent.Metadata["userName"];
            }
            catch (Exception e)
            {
                CreatedId = "TestId";
                LastModifiedId = "TestId";
                CreatedName = "TestName";
                LastModifiedName = "TestName";
            }

        }

        private void SetDefaultStartAndEnd()
        {
            var monDay = (int) DayOfWeek.Monday;
            var todaysWeekDay = (int) DateTime.Now.DayOfWeek;
            int daysToAdd;
            if (todaysWeekDay >= monDay)
            {
                daysToAdd = 8 - todaysWeekDay;
            }
            else
            {
                daysToAdd = 1;
            }

            Changes.Start = DateTime.Now.AddDays(daysToAdd).Date;
            Changes.End = Changes.Start.AddDays(7);
        }


        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutySavedEvent> domainEvent)
        {
            if (Changes == null) return;

            SavedOnce = true;
            Note = Changes.Note;
            Start = Changes.Start;
            End = Changes.End;
            AutomaticallyDestroy = Changes.AutomaticallyDestroy;
            TimeSlots = Changes.TimeSlots;

            Changes = null;

            LastModifiedOn = DateTime.UtcNow;
            try
            {
                LastModifiedId = domainEvent.Metadata["userId"];
                LastModifiedName = domainEvent.Metadata["userName"];
            }
            catch (Exception e)
            {
                LastModifiedId = "TestId";
                LastModifiedName = "TestName";
            }

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyCanceledEvent> domainEvent)
        {
            if (!SavedOnce)
            {
                Deleted = true;
                context.MarkForDeletion();
                return;
            }
            if (Changes == null) return;
            Changes = null;

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyDeletedEvent> domainEvent)
        {
            if (Deleted) return;
            Deleted = true;
            context.MarkForDeletion();

        }


        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyNoteModifiedEvent> domainEvent)
        {
            if (!CheckChanges()) return;
            var aggregateEvent = domainEvent.AggregateEvent;
            Changes.Note = aggregateEvent.Note;
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyEndModifiedEvent> domainEvent)
        {
            if (!CheckChanges()) return;
            var aggregateEvent = domainEvent.AggregateEvent;
            Changes.End = aggregateEvent.End;
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyStartModifiedEvent> domainEvent)
        {
            if (!CheckChanges()) return;
            var aggregateEvent = domainEvent.AggregateEvent;
            Changes.Start = aggregateEvent.Start;
        }
        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoDutyStartEndModifiedEvent> domainEvent)
        {
            if (!CheckChanges()) return;
            var aggregateEvent = domainEvent.AggregateEvent;
            Changes.Start = aggregateEvent.Start;
            Changes.End = aggregateEvent.End;

        }


        #endregion

        #region Timeslot
        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges()) return;
            var ts = new TimeSlotModel() { Id = aggregateEvent.TimeslotId.Value };
            Changes.TimeSlots.Add(ts);
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            Changes.TimeSlots.Remove(ts);
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoWeekdaysModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            ts.WeekDays = aggregateEvent.WeekDays;
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotStartTimeModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            ts.StartTime = aggregateEvent.StartTime;
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotEndTimeModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            ts.EndTime = aggregateEvent.EndTime;
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotNoteModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            ts.Note = aggregateEvent.Note;


        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTimeSlotTitleModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeslotId.Value)) return;
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            ts.TimeSlotName = aggregateEvent.Title;

        }

        #endregion

        #region DutyContact
        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChanges(aggregateEvent.TimeSlotId.Value)) return;

            AddOneEmployeeToTimeSlot(aggregateEvent.TimeSlotId.Value, aggregateEvent.ContactId.Value, aggregateEvent.EmployeeId, aggregateEvent.DefaultContactPoints, aggregateEvent.DefaultTags);

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactAddedToDutyEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            foreach (var contact in aggregateEvent.ContactList)
            {
                if (!CheckChanges(contact.TimeSlotId.Value))
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }
                AddOneEmployeeToTimeSlot(contact.TimeSlotId.Value, contact.ContactId.Value, aggregateEvent.EmployeeId, contact.DefaultContactPoints, aggregateEvent.DefaultTags);
            }

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (string.IsNullOrEmpty(aggregateEvent.TimeSlotId))
            {
                if (!CheckChanges())
                {
                    throw new InvalidOperationException("Something's wrong with your changes-object.");
                }

                foreach (var ts in Changes.TimeSlots)
                {
                    if (!CheckChanges(ts.Id, aggregateEvent.EmployeeId)) continue;

                    RemoveOneEmployeeFromTimeSlot(ts.Id, aggregateEvent.EmployeeId);
                }
                return;
            }

            if (!CheckChanges(aggregateEvent.TimeSlotId))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }

            RemoveOneEmployeeFromTimeSlot(aggregateEvent.TimeSlotId, aggregateEvent.EmployeeId);
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactOrderIdModifiedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;

            if (!CheckChanges(aggregateEvent.TimeslotId.Value, aggregateEvent.EmployeeId))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            var contact = ts.DutyContacts.Single(x => x.EmployeeId == aggregateEvent.EmployeeId);
            ReOrderContactsInTimeslot(ts, contact, aggregateEvent.NewOrderId);
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactPointAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChangesDetails(aggregateEvent.TimeslotId.Value, aggregateEvent.ContactId.Value))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId.Value);
            var cp = new DutyContactPointModel(aggregateEvent.ContactPointId.Value, aggregateEvent.ContactDetailId, "");
            contact.ContactPoints.Add(cp);

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoContactPointRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChangesDetails(aggregateEvent.TimeslotId.Value, aggregateEvent.ContactId.Value, dutyContactPointId: aggregateEvent.ContactPointId.Value))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId.Value);
            var cp = contact.ContactPoints.Single(x => x.Id == aggregateEvent.ContactPointId.Value);
            contact.ContactPoints.Remove(cp);

        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTagAddedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChangesDetails(aggregateEvent.TimeslotId.Value, aggregateEvent.ContactId.Value))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId.Value);
            contact.Tags.Add(aggregateEvent.Tag);
        }

        public void Apply(IReadModelContext context, IDomainEvent<NeoDutyAggregate, NeoDutyId, NeoTagRemovedEvent> domainEvent)
        {
            var aggregateEvent = domainEvent.AggregateEvent;
            if (!CheckChangesDetails(aggregateEvent.TimeslotId.Value, aggregateEvent.ContactId.Value, tag:aggregateEvent.Tag))
            {
                throw new InvalidOperationException("Something's wrong with your changes-object.");
            }
            var ts = Changes.TimeSlots.Single(x => x.Id == aggregateEvent.TimeslotId.Value);
            var contact = ts.DutyContacts.Single(x => x.Id == aggregateEvent.ContactId.Value);
            contact.Tags.Remove(aggregateEvent.Tag);   
        }


        #endregion

        #region helpertjes
        private bool CheckChanges(string timeSlotId = null, string employeeId = null)
        {
            if (Deleted) return false;
            if (Changes == null)
                Changes = new NeoDutyChangesReadModel()
                {
                    AutomaticallyDestroy = AutomaticallyDestroy,
                    End = End,
                    Note = Note,
                    Start = Start,
                    TimeSlots = TimeSlots == null
                        ? new List<TimeSlotModel>()
                        : TimeSlots.Select(x => new TimeSlotModel(x)).ToList()
                };
            if (timeSlotId == null) return true;

            var rtn = false;

            var timeSlot = Changes.TimeSlots.SingleOrDefault(x => x.Id == timeSlotId);

            if (timeSlot == null) return false;

            if (employeeId == null) return true;

            var dutyContact = timeSlot.DutyContacts.SingleOrDefault(x => x.EmployeeId == employeeId);

            if (dutyContact == null) return false;

            return true;
        }

        private bool CheckChangesDetails(string timeSlotId, string dutyContactId, string dutyContactPointId = null, string tag = null)
        {
            if (Deleted) return false;
            if (Changes == null)
                Changes = new NeoDutyChangesReadModel()
                {
                    AutomaticallyDestroy = AutomaticallyDestroy,
                    End = End,
                    Note = Note,
                    Start = Start,
                    TimeSlots = TimeSlots == null
                        ? new List<TimeSlotModel>()
                        : TimeSlots.Select(x => new TimeSlotModel(x)).ToList()
                };

            var rtn = false;

            var timeSlot = Changes.TimeSlots.SingleOrDefault(x => x.Id == timeSlotId);

            if (timeSlot == null) return false;


            var dutyContact = timeSlot.DutyContacts.SingleOrDefault(x => x.Id == dutyContactId);

            if (dutyContact == null) return false;

            if (dutyContactPointId == null && tag == null) return true;

            if (dutyContactPointId != null)
            {
                var dutyContactPoint = dutyContact.ContactPoints.SingleOrDefault(x => x.Id == dutyContactPointId);

                return dutyContactPoint != null;
            }
            else
            {
                var myTag = dutyContact.Tags.SingleOrDefault(x => x == tag);

                return myTag != null;
            }
        }



        private void AddOneEmployeeToTimeSlot(string timeSlotId, string contactId, string employeeId, List<DutyContactPointModel> defaultContactPoints, List<string> defaultTags)
        {
            var ts = Changes.TimeSlots.Single(x => x.Id == timeSlotId);

            if (ts.DutyContacts != null)
            {
                var alReadyExists = ts.DutyContacts.SingleOrDefault(x => x.EmployeeId == employeeId);

                if (alReadyExists != null) return;
            }


            var orderId = ts.DutyContacts == null ? 1 : ts.DutyContacts.Count + 1;
            DutyContactModel dc = new DutyContactModel(contactId, employeeId, new List<DutyContactPointModel>(), new List<string>(), orderId);


            foreach (var cp in defaultContactPoints)
            {
                dc.ContactPoints.Add(cp);
            }
            foreach (var tag in defaultTags)
            {
                dc.Tags.Add(tag);
            }
            if (ts.DutyContacts == null)
                ts.DutyContacts = new List<DutyContactModel>();

            ts.DutyContacts.Add(dc);
        }

        private void RemoveOneEmployeeFromTimeSlot(string timeSlotId, string employeeId)
        {
            var ts = Changes.TimeSlots.Single(x => x.Id == timeSlotId);

            var contact = ts.DutyContacts.Single(x => x.EmployeeId == employeeId);

            ReOrderContactsInTimeslot(ts, contact);
            ts.DutyContacts.Remove(contact);
        }

        private static void ReOrderContactsInTimeslot(TimeSlotModel ts, DutyContactModel dc, int newOrderId = -1)
        {
            if (ts.DutyContacts.Count == 1) return;

            if (newOrderId == -1 && dc.OrderId == ts.DutyContacts.Count) return;

            if (newOrderId == -1)
            {
                ts.DutyContacts.Where(x => x.OrderId > dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId--);
                return;
            }

            if (dc.OrderId == newOrderId) return;

            if (dc.OrderId < newOrderId)
            {
                ts.DutyContacts.Where(x => x.OrderId <= newOrderId && x.OrderId > dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId--);
            }
            else
            {
                ts.DutyContacts.Where(x => x.OrderId >= newOrderId && x.OrderId < dc.OrderId)
                    .ToList()
                    .ForEach(x => x.OrderId++);
            }
            dc.OrderId = newOrderId;
        }

    }


    #endregion
}
