﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Ticket.Entities;
using CRM.Domain.Ticket.Events;
using CRM.Domain.Ticket.Snapshots;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;
using EventFlow.ReadStores;
using EventFlow.Snapshots;
using EventFlow.Snapshots.Strategies;

namespace CRM.Domain.Ticket
{
    public class TicketAggregate :
        SnapshotAggregateRoot<TicketAggregate, TicketId, TicketSnapshot>,
        IEmit<CommentAddedEvent>,
        IEmit<CommentModifiedEvent>,
        IEmit<CommentRemovedEvent>,
        IEmit<TagAddedEvent>,
        IEmit<TagRemovedEvent>,
        IEmit<TicketCreatedEvent>,
        IEmit<TicketDeletedEvent>,
        IEmit<TicketModifiedEvent>,
        IEmit<TicketReadEvent>,
        IEmit<TicketUnreadEvent>,
        IEmit<TicketModifiedDossierIdEvent>, IReadModel
    {
        private bool _created;

        public TicketAggregate(TicketId id) : base(id, SnapshotEveryFewVersionsStrategy.Default)
        {
        }

        public bool Deleted { get; private set; }
        public ICollection<string> TicketStatuses { get; private set; } = new List<string>();
        public ICollection<string> Tags { get; private set; } = new List<string>();
        public ICollection<TicketComment> Comments { get; private set; } = new List<TicketComment>();
        public int? OriginalCCMMessageId { get; private set; }
        public string DossierId { get; private set; } = "";
        public string TicketText { get; private set; } = "";
        public DateTime CreatedInCRM { get; private set; }
        public DateTime LastModifiedInCRM { get; private set; }
        public string CreatedCRMUser { get; private set; } = "";
        public string LastModifiedCRMUser { get; private set; } = "";

        protected override Task<TicketSnapshot> CreateSnapshotAsync(CancellationToken cancellationToken)
        {
            return Task.Run(() => new TicketSnapshot
            {
                DossierId = DossierId,
                Deleted = Deleted,
                Tags = Tags,
                Comments = Comments,
                CreatedCRMUser = CreatedCRMUser,
                CreatedInCRM = CreatedInCRM,
                LastModifiedCRMUser = LastModifiedCRMUser,
                LastModifiedInCRM = LastModifiedInCRM,
                OriginalCCMMessageId = OriginalCCMMessageId,
                TicketStatuses = TicketStatuses,
                TicketText = TicketText
            }, cancellationToken);
        }

        protected override Task LoadSnapshotAsync(TicketSnapshot snapshot,
            ISnapshotMetadata metadata,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                DossierId = snapshot.DossierId;
                Deleted = snapshot.Deleted;
                Tags = snapshot.Tags;
                Comments = Comments;
                CreatedCRMUser = snapshot.CreatedCRMUser;
                CreatedInCRM = snapshot.CreatedInCRM;
                LastModifiedCRMUser = snapshot.LastModifiedCRMUser;
                LastModifiedInCRM = snapshot.LastModifiedInCRM;
                OriginalCCMMessageId = snapshot.OriginalCCMMessageId;
                TicketStatuses = snapshot.TicketStatuses;
                TicketText = snapshot.TicketText;
                _created = true;
            }, cancellationToken);
        }

        #region IExecutionResult

        public IExecutionResult CreateTicket(string dossierId,
            string ticketText,
            int? originalCCMMessageId,
            DateTime createdInCRM,
            DateTime lastModifiedInCRM,
            string createdCRMUser,
            string lastModifiedCRMUser,
            ICollection<string> tags)
        {
            if (string.IsNullOrEmpty(ticketText) || string.IsNullOrEmpty(dossierId))
                return ExecutionResult.Failed();
            var _event = new TicketCreatedEvent(dossierId, ticketText, originalCCMMessageId, createdInCRM,
                lastModifiedInCRM, createdCRMUser, lastModifiedCRMUser, tags);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyTicket(string ticketText,
            DateTime lastModifiedInCRM,
            string lastModifiedCRMUser)
        {
            if (!_created) return ExecutionResult.Failed();
            var _event = new TicketModifiedEvent(ticketText, lastModifiedInCRM, lastModifiedCRMUser);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ReadTicket(string userId)
        {
            if (!_created) return ExecutionResult.Failed();
            var _event = new TicketReadEvent(userId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult UnReadTicket(string userId)
        {
            if (!_created) return ExecutionResult.Failed();
            var _event = new TicketUnreadEvent(userId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddTag(string tagId)
        {
            if (!_created) return ExecutionResult.Failed();
            if (Tags.Any(x => x.Equals(tagId))) return ExecutionResult.Success();
            var _event = new TagAddedEvent(tagId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveTag(string tagId)
        {
            if (!_created) return ExecutionResult.Failed();
            var _event = new TagRemovedEvent(tagId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult UpSertComment(string commentBody,
            string commentQuote,
            string userId,
            TicketCommentId ticketCommentId = null)
        {
            IAggregateEvent<TicketAggregate, TicketId> _event;
            if (!_created) return ExecutionResult.Failed();
            if (Comments.Any(x => x.Id.Equals(ticketCommentId)))
            {
                _event = new CommentModifiedEvent(commentBody, commentQuote, ticketCommentId);
            }
            else
            {
                if (string.IsNullOrEmpty(userId)) return ExecutionResult.Failed();
                _event = new CommentAddedEvent(TicketCommentId.New, commentBody, commentQuote, userId);
            }

            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveComment(TicketCommentId commentId)
        {
            if (!_created) return ExecutionResult.Failed();
            if (!Comments.Any(x => x.Id.Equals(commentId))) return ExecutionResult.Failed();
            var _event = new CommentRemovedEvent(commentId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteTicket()
        {
            if (!_created) return ExecutionResult.Failed();
            Emit(new TicketDeletedEvent());
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyTicketDossierId(string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"Duty with id {Id.Value} does not exist");
            var _event = new TicketModifiedDossierIdEvent(dossierId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        #endregion

        #region Apply

        public void Apply(TicketCreatedEvent aggregateEvent)
        {
            DossierId = aggregateEvent.DossierId;
            TicketText = aggregateEvent.TicketText;
            OriginalCCMMessageId = aggregateEvent.OriginalCCMMessageId;
            CreatedCRMUser = aggregateEvent.CreatedCRMUser;
            CreatedInCRM = aggregateEvent.CreatedInCRM;
            LastModifiedCRMUser = aggregateEvent.LastModifiedCRMUser;
            LastModifiedInCRM = aggregateEvent.LastModifiedInCRM;
            _created = true;
        }

        public void Apply(TicketModifiedEvent aggregateEvent)
        {
            LastModifiedCRMUser = aggregateEvent.LastModifiedCRMUser;
            LastModifiedInCRM = aggregateEvent.LastModifiedInCRM;
            TicketText = aggregateEvent.TicketText;
        }


        public void Apply(TagAddedEvent aggregateEvent)
        {
            Tags.Add(aggregateEvent.TagId);
        }

        public void Apply(TagRemovedEvent aggregateEvent)
        {
            Tags.Remove(aggregateEvent.TagId);
        }

        public void Apply(CommentAddedEvent aggregateEvent)
        {
            Comments.Add(new TicketComment(aggregateEvent.Id, aggregateEvent.CommentBody, aggregateEvent.CommentQuote,
                aggregateEvent.UserId));
        }

        public void Apply(CommentRemovedEvent aggregateEvent)
        {
            Comments.Remove(Comments.Single(x => x.Id.Equals(aggregateEvent.TicketCommentId)));
        }

        public void Apply(TicketDeletedEvent aggregateEvent)
        {
            Deleted = true;
        }

        public void Apply(CommentModifiedEvent aggregateEvent)
        {
            var comment = Comments.Single(x => x.Id.Equals(aggregateEvent.CommentId));
            comment.CommentBody = aggregateEvent.CommentBody;
            comment.CommentQuote = aggregateEvent.CommentQuote;
        }

        public void Apply(TicketUnreadEvent aggregateEvent)
        {
            TicketStatuses.Remove(aggregateEvent.UserId);
        }

        public void Apply(TicketReadEvent aggregateEvent)
        {
            TicketStatuses.Add(aggregateEvent.UserId);
        }

        public void Apply(TicketModifiedDossierIdEvent aggregateEvent)
        {
            DossierId = aggregateEvent.DossierId;
        }

        #endregion

    }
}