﻿using System.Collections.Generic;
using System.Linq;
using CRM.Domain.User.Entities;
using CRM.Domain.User.Events;
using CRM.Domain.User.Models;
using EventFlow.Aggregates;
using EventFlow.Aggregates.ExecutionResults;

namespace CRM.Domain.User
{
    public class UserAggregate : AggregateRoot<UserAggregate, UserId>,
        IEmit<UserCreatedEvent>,
        IEmit<UserNameModifiedEvent>,
        IEmit<UserConnectedToEmployeeEvent>,
        IEmit<UserDeletedEvent>,
        IEmit<ModuleRightDossierAddedEvent>,
        IEmit<ModuleRightCustomerAddedEvent>,
        IEmit<ModuleRightCustomerGroupAddedEvent>,
        IEmit<ModuleRightDossierRemovedEvent>,
        IEmit<ModuleRightCustomerRemovedEvent>,
        IEmit<ModuleRightCustomerGroupRemovedEvent>,
        IEmit<ModuleRightDossierModifiedEvent>,
        IEmit<ModuleRightCustomerModifiedEvent>,
        IEmit<ModuleRightCustomerGroupModifiedEvent>,
        IEmit<UserCompanyAdminAssignedEvent>,
        IEmit<UserCustomerAdminModifiedEvent>,
        IEmit<UserCustomerGroupModuleRightsModifiedEvent>,
        IEmit<UserCustomerModifiedEvent>,
        IEmit<UserCustomerModuleRightsModifiedEvent>,
        IEmit<UserDossierModuleRightsModifiedEvent>,
        IEmit<UserEmailModifiedEvent>,
        IEmit<UserTenantSetEvent>
    {
        private bool _created;

        private Dictionary<string, ModuleRightModel> _customerGroupModuleRights =
            new Dictionary<string, ModuleRightModel>();

        private string? _customerId;
        private Dictionary<string, ModuleRightModel> _customerModuleRights = new Dictionary<string, ModuleRightModel>();
        private bool _deleted;
        private Dictionary<string, ModuleRightModel> _dossierModuleRights = new Dictionary<string, ModuleRightModel>();
        private string _email = "";
        private string? _employeeId;
        private bool _isCustomerAdmin;
        private string _userName = "";
        private string? _tenant = "";

        public UserAggregate(UserId id) : base(id)
        {
        }

        public bool Deleted
        {
            get => _deleted;
            set
            {
                if (!value) return;
                Emit(new UserDeletedEvent());
            }
        }

        public string Email
        {
            get => _email;
            set => Emit(new UserEmailModifiedEvent(value));
        }

        public string UserName
        {
            get => _userName;
            set => Emit(new UserNameModifiedEvent(value));
        }

        public string? CustomerId
        {
            get => _customerId;
            set => Emit(new UserCustomerModifiedEvent(value));
        }

        public bool IsCustomerAdmin
        {
            get => _isCustomerAdmin;
            set => Emit(new UserCustomerAdminModifiedEvent(value));
        }

        public string? EmployeeId
        {
            get => _employeeId;
            set => Emit(new UserConnectedToEmployeeEvent(value));
        }

        public IReadOnlyDictionary<string, ModuleRightModel> CustomerModuleRights
        {
            get => _customerModuleRights;
            set => Emit(new UserCustomerModuleRightsModifiedEvent(value));
        }

        public IReadOnlyDictionary<string, ModuleRightModel> CustomerGroupModuleRights
        {
            get => _customerGroupModuleRights;
            set => Emit(new UserCustomerGroupModuleRightsModifiedEvent(value));
        }

        public IReadOnlyDictionary<string, ModuleRightModel> DossierModuleRights
        {
            get => _dossierModuleRights;
            set => Emit(new UserDossierModuleRightsModifiedEvent(value));
        }

        public string? Tenant
        {
            get => _tenant;
            set => Emit(new UserTenantSetEvent(value));
        }

        #region IExecutionResult

        public IExecutionResult CreateUser(string email,
            string userName,
            string customerId,
            bool isCustomerAdmin,
            string employeeId,
            string? commandTenant)
        {
            if (_created) return ExecutionResult.Failed("User already exist!");
            Emit(new UserCreatedEvent(email, userName, customerId, isCustomerAdmin, employeeId, commandTenant ?? ""));
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyUserName(string userName)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            Emit(new UserNameModifiedEvent(userName));
            return ExecutionResult.Success();
        }

        public IExecutionResult AssignCompanyAdmin(bool isCompanyAdmin)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            var _event = new UserCompanyAdminAssignedEvent(IsCustomerAdmin);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ConnectToEmployee(string employeeId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            var _event = new UserConnectedToEmployeeEvent(employeeId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult DeleteUser()
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            var _event = new UserDeletedEvent();
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyDossierModuleRight(ModuleRightModel moduleRight,
            string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!DossierModuleRights.Any(x => x.Key == dossierId))
                return ExecutionResult.Failed($"Could not find dossier {dossierId}");

            var _event = new ModuleRightDossierModifiedEvent(moduleRight, dossierId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyCustomerModuleRight(ModuleRightModel moduleRight,
            string customerId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!CustomerModuleRights.Any(x => x.Key == customerId))
                return ExecutionResult.Failed($"Could not find customer {customerId}");

            var _event = new ModuleRightCustomerModifiedEvent(moduleRight, customerId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult ModifyCustomerGroupModuleRight(ModuleRightModel moduleRight,
            string customerGroupId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!CustomerGroupModuleRights.Any(x => x.Key == customerGroupId))
                return ExecutionResult.Failed($"Cound not find customerGroup {customerGroupId}");

            var _event = new ModuleRightCustomerGroupModifiedEvent(moduleRight, customerGroupId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddDossierModuleRight(ModuleRightModel moduleRight,
            string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (DossierModuleRights.Any(x => x.Key == dossierId))
                return ExecutionResult.Failed($"Dossier {dossierId} already exists!");

            var _event = new ModuleRightDossierAddedEvent(moduleRight, dossierId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddCustomerModuleRight(ModuleRightModel moduleRight,
            string customerId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (CustomerModuleRights.Any(x => x.Key == customerId))
                return ExecutionResult.Failed($"Customer {customerId} already exixts!");

            var _event = new ModuleRightCustomerAddedEvent(moduleRight, customerId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult AddCustomerGroupModuleRight(ModuleRightModel moduleRight,
            string customerGroupId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (CustomerGroupModuleRights.Any(x => x.Key == customerGroupId))
                return ExecutionResult.Failed($"CustomerGroup {customerGroupId} already exixts!");

            var _event = new ModuleRightCustomerGroupAddedEvent(moduleRight, customerGroupId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveDossierModuleRight(string dossierId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!DossierModuleRights.Any(x => x.Key == dossierId))
                return ExecutionResult.Failed($"Could not find dossier {dossierId}");

            var _event = new ModuleRightDossierRemovedEvent(dossierId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveCustomerModuleRight(string customerId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!CustomerModuleRights.Any(x => x.Key == customerId))
                return ExecutionResult.Failed($"Could not find customer {customerId}");

            var _event = new ModuleRightCustomerRemovedEvent(customerId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        public IExecutionResult RemoveCustomerGroupModuleRight(string customerGroupId)
        {
            if (!_created || Deleted) return ExecutionResult.Failed($"User with id {Id.Value} does not exist");
            if (!CustomerGroupModuleRights.Any(x => x.Key == customerGroupId))
                return ExecutionResult.Failed($"Cound not find customerGroup {customerGroupId}");

            var _event = new ModuleRightCustomerGroupRemovedEvent(customerGroupId);
            Emit(_event);
            return ExecutionResult.Success();
        }

        #endregion IExecutionResult

        #region Apply

        public void Apply(UserCreatedEvent aggregateEvent)
        {
            _email = aggregateEvent.Email;
            _customerId = aggregateEvent.CustomerId;
            _isCustomerAdmin = aggregateEvent.IsCustomerAdmin;
            _userName = aggregateEvent.UserName;
            _created = true;
        }

        public void Apply(UserConnectedToEmployeeEvent aggregateEvent)
        {
            _employeeId = aggregateEvent.EmployeeId;
        }

        public void Apply(UserDeletedEvent aggregateEvent)
        {
            _deleted = true;
        }

        public void Apply(UserCompanyAdminAssignedEvent aggregateEvent)
        {
            _isCustomerAdmin = aggregateEvent.IsCompanyAdmin;
        }

        public void Apply(ModuleRightCustomerRemovedEvent aggregateEvent)
        {
            _customerModuleRights.Remove(aggregateEvent.CustomerId);
        }

        public void Apply(ModuleRightDossierAddedEvent aggregateEvent)
        {
            _dossierModuleRights.Add(aggregateEvent.DossierId, aggregateEvent.ModuleRight);
        }

        public void Apply(ModuleRightCustomerAddedEvent aggregateEvent)
        {
            _customerModuleRights.Add(aggregateEvent.CustomerId, aggregateEvent.ModuleRight);
        }

        public void Apply(ModuleRightCustomerGroupAddedEvent aggregateEvent)
        {
            _customerGroupModuleRights.Add(aggregateEvent.CustomerGroupId, aggregateEvent.ModuleRight);
        }

        public void Apply(ModuleRightDossierRemovedEvent aggregateEvent)
        {
            _dossierModuleRights.Remove(aggregateEvent.DossierId);
        }

        public void Apply(ModuleRightCustomerGroupRemovedEvent aggregateEvent)
        {
            _customerGroupModuleRights.Remove(aggregateEvent.CustomerGroupId);
        }

        public void Apply(ModuleRightDossierModifiedEvent aggregateEvent)
        {
            _dossierModuleRights[aggregateEvent.DossierId] = aggregateEvent.ModuleRight;
        }

        public void Apply(ModuleRightCustomerModifiedEvent aggregateEvent)
        {
            _customerModuleRights[aggregateEvent.CustomerId] = aggregateEvent.ModuleRight;
        }

        public void Apply(ModuleRightCustomerGroupModifiedEvent aggregateEvent)
        {
            _customerGroupModuleRights[aggregateEvent.CustomerGroupId] = aggregateEvent.ModuleRight;
        }

        public void Apply(UserNameModifiedEvent aggregateEvent)
        {
            _userName = aggregateEvent.UserName;
        }

        public void Apply(UserCustomerAdminModifiedEvent aggregateEvent)
        {
            this._isCustomerAdmin = aggregateEvent.IsCustomerAdmin;
        }

        public void Apply(UserCustomerGroupModuleRightsModifiedEvent aggregateEvent)
        {
            this._customerGroupModuleRights =
                aggregateEvent.ModuleRights.ToDictionary(pair => pair.Key, pair => pair.Value);
        }

        public void Apply(UserCustomerModifiedEvent aggregateEvent)
        {
            _customerId = aggregateEvent.CustomerId;
        }

        public void Apply(UserCustomerModuleRightsModifiedEvent aggregateEvent)
        {
            this._customerModuleRights = aggregateEvent.ModuleRights.ToDictionary(pair => pair.Key, pair => pair.Value);
        }

        public void Apply(UserDossierModuleRightsModifiedEvent aggregateEvent)
        {
            this._dossierModuleRights = aggregateEvent.ModuleRights.ToDictionary(pair => pair.Key, pair => pair.Value);
        }

        public void Apply(UserEmailModifiedEvent aggregateEvent)
        {
            _email = aggregateEvent.Email;
        }

        #endregion Apply

        public void Apply(UserTenantSetEvent aggregateEvent)
        {
            _tenant = aggregateEvent.Tenant;
        }
    }
}