﻿using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Branch;
using CRM.Domain.Branch.Models;
using CRM.Domain.Branch.ReadModels;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using WebPortal.API.Services.Geocoder;
using WebPortal.Application.Branches.Commands;
using WebPortal.Application.Branches.Queries;

namespace WebPortal.API.Controllers
{
    public class BranchController : ReprojectController<BranchReadModel>
    {
        private readonly IConfiguration _configuration;

        public BranchController(IConfiguration configuration)
        {
            _configuration = configuration;
        }

        [HttpGet]
        [Route("/api/branches")]
        [ProducesResponseType(typeof(IEnumerable<BranchReadModel>), 200)]
        public async Task<IActionResult> GetAllBranches([FromQuery] GetBranchesByCustomerIdQuery query,
            CancellationToken cancellationToken)
        {
            try
            {
                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
                if (result.Any())
                    return Ok(result);
                return NoContent();
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("")]
        [ProducesResponseType(typeof(IEnumerable<BranchReadModel>), 200)]
        public async Task<IActionResult> GetBranch([FromQuery] GetBranchByBranchIdQuery query,
            CancellationToken cancellationToken)
        {
            try
            {
                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);

                if (result != null)
                    return Ok(result);
                return NoContent();
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("hq")]
        [ProducesResponseType(typeof(BranchReadModel), 200)]
        public async Task<IActionResult> GetHqBranch([FromQuery] GetHqByCustomerIdQuery query,
            CancellationToken cancellationToken)
        {
            try
            {
                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);

                if (result == null) return NoContent();

                if (!string.IsNullOrEmpty(result.Address?.Latitude) && !string.IsNullOrEmpty(result.Address?.Longitude))
                    return Ok(result);

                await GetAndSaveCoordinations(result.Id, result.Address, cancellationToken);
                var q = new GetBranchByBranchIdQuery(result.Id, result.CustomerId);
                var newResult = await QueryProcessor.ProcessAsync(q, cancellationToken);
                return Ok(newResult);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        private async Task GetAndSaveCoordinations(string branchId,
            AddressModel model,
            CancellationToken cancellationToken)
        {
            var geocoder = new GeoCoder(_configuration.GetSection("GoogleApiKey").Get<string>());
            var address =
                geocoder.GetGeographicInfo(model.Street, model.City, model.State, model.PostalCode, model.Country);

            if (!address.status.ToLower().Equals("ok"))
                return;

            model.Latitude = address.results.FirstOrDefault()?.geometry.location.lat;
            model.Longitude = address.results.FirstOrDefault()?.geometry.location.lng;

            var command = new ModifyBranchAddressCommand(new BranchId(branchId), model);
            await CommandBus.PublishAsync(command, cancellationToken);
        }
    }
}