﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.DutyCategoryAndTemplate.Aggregates;
using CRM.Domain.DutyCategoryAndTemplate.Entities;
using CRM.Domain.DutyCategoryAndTemplate.ReadModels;
using CRM.Domain.Shared.ValueObjects;
using Microsoft.AspNetCore.Mvc;
using WebPortal.API.Model;
using WebPortal.Application.DutyCategories.Commands;
using WebPortal.Application.DutyCategories.Commands.Duties;
using WebPortal.Application.DutyCategories.Commands.Duties.TimeSlots;
using WebPortal.Application.DutyCategories.Commands.Duties.TimeSlots.DutyContacts;
using WebPortal.Application.DutyCategories.Commands.Duties.TimeSlots.DutyContacts.DutyContactPoints;
using WebPortal.Application.DutyCategories.Commands.Duties.TimeSlots.DutyContacts.Tags;
using WebPortal.Application.DutyCategories.Commands.DutyTemplate;
using WebPortal.Application.DutyCategories.Models;
using WebPortal.Application.DutyCategories.Queries;
using WebPortal.Application.Extentions;
using WebPortal.Application.GlobalDutyTemplates.Commands;
using WebPortal.Application.GlobalDutyTemplates.Models;
using WebPortal.Application.GlobalDutyTemplates.Queries;

namespace WebPortal.API.Controllers
{
    public class DutyCategoryController: ReprojectController<DutyCategoryReadModel>
    {
        #region Queries
        [HttpGet]
        [ProducesDefaultResponseType(typeof(List<DutyOverviewModel>))]
        [Route("/api/dutycategories/duties")]
        public async Task<IActionResult> GetDuties(
            [FromQuery] string dossierId,
            [FromQuery] DateTime start,
            [FromQuery] DateTime end, 
            [FromQuery] bool showActiveDuties,
            [FromQuery] bool showPlannedDuties,
            [FromQuery] bool showExpiredDuties,
            [FromQuery] bool showEmptyCategories, 
            CancellationToken token, 
            [FromQuery] string? filter = null,
            [FromQuery] string? employeeId = null)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetDutyCategoriesQuery(dossierId, start, end, showActiveDuties, showPlannedDuties,
                    showExpiredDuties, showEmptyCategories, filter, employeeId);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(PageResult<DutyCategoryOverviewModel>))]
        [Route("/api/dutycategories")]
        public async Task<IActionResult> GetDutyCategories(CancellationToken token, string dossierId,
            bool asc = true,
            string? sort = "Title",
            int page = 1,
            int pageSize = 100,
            string? filter = "")
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetDutyCategoryOverviewQuery(asc, sort, page, pageSize, dossierId, filter);
                var result = await QueryProcessor.ProcessAsync(q, token);
                Response.Headers.Add("X-pagination-total", result.Total.ToString());
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {

                throw;
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(DutyCategoryReadModel))]
        [Route("{id}")]
        public async Task<IActionResult> GetOneDutyCategory(
            string id, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetOneDutyCategoryQuery(id);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(IdAndTitleModel))]
        [Route("/api/dutycategories/globaldutytitles")]
        public async Task<IActionResult> GetGlobalDutyTitles(string filter, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetTitlesGlobalDutyQuery(filter);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        #endregion

        #region DutyCategory

        [HttpPost]
        public async Task<IActionResult> CreateDutyCategory([FromBody]CreateDutyCategoryModel model, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new CreateDutyCategoryCommand(DutyCategoryId.With(model.DutyCategoryId), model.DossierId);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess && !string.IsNullOrEmpty(model.GlobalDutyTemplateId))
                {
                    var q = new GetOneGlobalDutyTemplateQuery(model.GlobalDutyTemplateId);
                    var gdt = await QueryProcessor.ProcessAsync(q, token);

                    var commandI = new InheritFromGlobalDutyTemplateCommand(DutyCategoryId.With(model.DutyCategoryId), gdt, model.GlobalDutyTemplateId);
                    var resultC = await CommandBus.PublishAsync(commandI, token);
                    if (!resultC.IsSuccess)
                    {
                        return StatusCode(500, new { errorMessage = $"The Duty Category was created but failed to inherit from Global Duty Template '{model.GlobalDutyTemplateId}'" });
                    }

                }
                if(result.IsSuccess && !string.IsNullOrEmpty(model.DutyCategoryName))
                {
                    var commandN = new ModifyNameDutyCategoryCommand(DutyCategoryId.With(model.DutyCategoryId), model.DutyCategoryName);
                    var resultN = await CommandBus.PublishAsync(commandN, token);

                    if (!resultN.IsSuccess)
                    {
                        return StatusCode(500, new { errorMessage = $"The Duty Category was created, but failed to give it Title: '{model.DutyCategoryName}'." });
                    }
                }

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not create a Duty Category in dossier {model.DossierId}" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPost]
        [Route("{id}/copy")]
        public async Task<IActionResult> CopyDutyCategory(string id, [FromBody]OneStringModel newId, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetOneDutyCategoryQuery(id);
                var catToCopy = await QueryProcessor.ProcessAsync(q, token);

                if (catToCopy.Changes != null)
                    return StatusCode(403, "You cannot duplicate a Duty Category with unsaved changes!");


                var command = new CopyDutyCategoryCommand(DutyCategoryId.With(newId.ValOs), catToCopy);
                var resultC = await CommandBus.PublishAsync(command, token);

                if (resultC.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not copy Duty Category '{catToCopy.CategoryName}'." });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/globalduty")]
        public async Task<IActionResult> InheritFromGlobalDutyTemplate(string id, [FromBody]OneStringModel globalDutyTemplateId, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var q = new GetOneGlobalDutyTemplateQuery(globalDutyTemplateId.ValOs);
                var gdt = await QueryProcessor.ProcessAsync(q, token);

                var clearCommand = new ClearDutyCategoryCommand(DutyCategoryId.With(id));
                var clearResult = await CommandBus.PublishAsync(clearCommand, token);

                if (!clearResult.IsSuccess) return StatusCode(500, new { errorMessage = $"Could not clear the DutyCategory with Id: {id}. Copy of '{gdt.Title}' was aborted" });

                var command = new InheritFromGlobalDutyTemplateCommand(DutyCategoryId.With(id), gdt, globalDutyTemplateId.ValOs);
                var resultC = await CommandBus.PublishAsync(command, token);

                if (resultC.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not inherit from Global Duty Template '{gdt.Title}'." });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/globalduty")]
        public async Task<IActionResult> ClearDutyCategory(string id, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();
                var clearCommand = new ClearDutyCategoryCommand(DutyCategoryId.With(id));
                var clearResult = await CommandBus.PublishAsync(clearCommand, token);

                if (!clearResult.IsSuccess) return StatusCode(500, new { errorMessage = $"Could not clear the DutyCategory with Id: {id}." });

                return Ok();
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}")]
        public async Task<IActionResult> DeleteDutyCategory(string id,  CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var query = new GetGlobalDutyIdsQuery(id);
                var qRes = await QueryProcessor.ProcessAsync(query, token);

                if (qRes != null && !string.IsNullOrEmpty(qRes.SavedId))
                {
                    var decreaseCommand = new ModifyUsageGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(qRes.SavedId), false);
                    await CommandBus.PublishAsync(decreaseCommand, token);
                }

                var command = new DeleteDutyCategoryCommand(DutyCategoryId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/categoryname")]
        public async Task<IActionResult> ModifyDutyCategoryName(string id, [FromBody]OneStringModel categoryName, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyNameDutyCategoryCommand(DutyCategoryId.With(id), categoryName.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();


                return StatusCode(500, new { errorMessage = $"Could not set '{categoryName}' as  categoryName for Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/note")]
        public async Task<IActionResult> ModifyDutyCategoryDutyTemplateNote(string id, [FromBody] OneStringModel note, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyNoteDutyTemplateCommand(DutyCategoryId.With(id), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add note to Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/globaldutytemplateid")]
        public async Task<IActionResult> ModifyDutyCategoryGlobalDutyId(string id, [FromBody] OneStringModel globalDutyTemplateId, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyGlobalDutyTemplateDutyCategoryCommand(DutyCategoryId.With(id), globalDutyTemplateId.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{globalDutyTemplateId}' as  GlobalDutyTemplate for Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/usageIncrement")]
        public async Task<IActionResult> IncreaseUsageGlobalDutyTemplate(string id,
    CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyUsageGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), true);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set Usage  on Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/usageDecrement")]
        public async Task<IActionResult> DecreaseUsageGlobalDutyTemplate(string id,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var usageCommand = new ModifyUsageGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), false);
                var resultA = await CommandBus.PublishAsync(usageCommand, token);

                if (resultA.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set Usage  on Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }


        [HttpPost]
        [Route("{id}/save")]
        public async Task<IActionResult> SaveDutyCategory(string id,  CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var qDetails = new GetOneDutyCategoryQuery(id);
                var dcResults = await QueryProcessor.ProcessAsync(qDetails, token);

                if (dcResults.Changes == null) return Ok();

                if (string.IsNullOrWhiteSpace(dcResults.Changes.CategoryName))
                    return StatusCode(403, "A Duty Category always requires a title!");

                if (dcResults.Changes.TimeSlotTemplates == null || !dcResults.Changes.TimeSlotTemplates.Any())
                    return StatusCode(403, "You need at least one Timeslot!");

                var query = new GetGlobalDutyIdsQuery(id);
                var globalIds = await QueryProcessor.ProcessAsync(query, token);

                var editUsage = await EditGlobalDutyTemplateUsage(globalIds, token);

                if (editUsage != null)
                    return StatusCode(500, new { errorMessage = editUsage});

                var command = new SaveDutyCategoryCommand(DutyCategoryId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not Save Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        private async Task<string?> EditGlobalDutyTemplateUsage(GlobalDutyIdModel globalIds, CancellationToken token)
        {
            if (globalIds == null) return null;
            if (string.IsNullOrEmpty(globalIds.ChangesId) && string.IsNullOrEmpty(globalIds.SavedId)) return null;

            var c = globalIds.ChangesId ?? "";
            var s = globalIds.SavedId ?? "";

            if (c == s) return null;

            if (!string.IsNullOrEmpty(globalIds.ChangesId))
            {
                var changesCommand = new ModifyUsageGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(globalIds.ChangesId), true);
                var resultC = await CommandBus.PublishAsync(changesCommand, token);
                if (!resultC.IsSuccess)
                    return  $"Could not increase usage-counter for Global Duty Template '{globalIds.ChangesId}'";

            }

            if (!string.IsNullOrEmpty(globalIds.SavedId))
            {
                var savedCommand = new ModifyUsageGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(globalIds.SavedId), false);
                var resultS = await CommandBus.PublishAsync(savedCommand, token);

                if (!resultS.IsSuccess)
                    return $"Could not decrease usage-counter for Global Duty Template '{globalIds.SavedId}'";

            }
            return null;
        }

        [HttpPost]
        [Route("{id}/cancel")]
        public async Task<IActionResult> CancelDutyCategory(string id, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new CancelDutyCategoryCommand(DutyCategoryId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not Cancel Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion

        #region TimeSlots


        [HttpPost]
        [Route("{id}/timeslottemplates")]
        public async Task<IActionResult> AddTimeSlotInDutyTemplate(string id, [FromBody] OneStringModel timeSlotTemplateId, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new AddTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId.ValOs));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add TimeSlot Template to Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}")]
        public async Task<IActionResult> DeleteTimeSlotInDutyTemplate(string id, string timeSlotTemplateId, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new DeleteTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not Delete TimeSlot Template '{timeSlotTemplateId}' from Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/starttime")]
        public async Task<IActionResult> ModifyStartTimeTimeSlotDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneTimeSpanModel startTime, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyStartTimeTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), startTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add Starttime {startTime.TimeVal.ToString("g")} to TimeSlottemplate '{timeSlotTemplateId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/endtime")]
        public async Task<IActionResult> ModifyEndTimeTimeSlotDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneTimeSpanModel endTime, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyEndTimeTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), endTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add Endtime {endTime.TimeVal.ToString("g")} to TimeSlottemplate '{timeSlotTemplateId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/name")]
        public async Task<IActionResult> ModifyTimeSlotNameDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneStringModel name, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyNameTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), name.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{name.ValOs}' as Name in TimeSlottemplate '{timeSlotTemplateId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/note")]
        public async Task<IActionResult> ModifyNoteTimeSlotDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneStringModel note, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyNoteTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{note.ValOs}' as Note in TimeSlottemplate '{timeSlotTemplateId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/weekdays")]
        public async Task<IActionResult> ModifyWeekdaysTimeSlotDutyTemplate(string id, string timeSlotTemplateId, ICollection<WeekDay> weekDays, CancellationToken token)
        {
            try
            {
                if (!await AuthorizationService.IsPortalAdmin(UserContext.User, token)) return Unauthorized();

                var command = new ModifyWeekdaysTimeSlotTemplateCommand(DutyCategoryId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), weekDays);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set weekdays in TimeSlottemplate '{timeSlotTemplateId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion

        #region Duty
        [HttpPost]
        [Route("{id}/duties")]
        public async Task<IActionResult> AddDuty(string id,  [FromBody] OneStringModel dutyId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);
                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new AddDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId.ValOs));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add Duty to Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/duty/{dutyId}")]
        public async Task<IActionResult> DeleteDuty(string id, string dutyId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new DeleteDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete Duty '{dutyId}' from Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/start")]
        public async Task<IActionResult> ModifyStartDuty(string id,  string dutyId, [FromBody] OneDateTimeModel start, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new ModifyStartDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), start.ValDat);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{start.ValDat.ToString("g")}' as Start for Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/end")]
        public async Task<IActionResult> ModifyEndDuty(string id,  string dutyId, [FromBody] OneDateTimeModel end, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new ModifyEndDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), end.ValDat);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{end.ValDat.ToString("g")}' as End for Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/note")]
        public async Task<IActionResult> ModifyNoteDuty(string id,  string dutyId, [FromBody] OneStringModel note, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new ModifyNoteDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{note.ValOs}' as Note for Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }


        #endregion

        #region TimeSlots

        [HttpPost]
        [Route("{id}/duty/{dutyId}/timeslots")]
        public async Task<IActionResult> AddTimeSlotToDuty(string id,  string dutyId, [FromBody] OneStringModel timeSlotId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();


                var command = new AddTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId.ValOs));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add a new TimeSlot to Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}")]
        public async Task<IActionResult> DeleteTimeSlotFromDuty(string id,  string dutyId, string timeSlotId,  CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new DeleteTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete TimeSlot '{timeSlotId}' from Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/starttime")]
        public async Task<IActionResult> ModifyTimeSlotStartTimeInDuty(string id,  string dutyId, string timeSlotId, [FromBody] OneTimeSpanModel startTime, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyStartTimeTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), startTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set {startTime.TimeVal.ToString("g")} as StartTime in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/endtime")]
        public async Task<IActionResult> ModifyTimeSlotEndTimeInDuty(string id,  string dutyId, string timeSlotId, [FromBody] OneTimeSpanModel endTime, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyEndTimeTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), endTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set {endTime.TimeVal.ToString("g")} as EndTime in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/note")]
        public async Task<IActionResult> ModifyTimeSlotNoteInDuty(string id,  string dutyId, string timeSlotId, [FromBody] OneStringModel note, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyNoteTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{note.ValOs}' as Note in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/name")]
        public async Task<IActionResult> ModifyTimeSlotNameInDuty(string id,  string dutyId, string timeSlotId, [FromBody] OneStringModel name, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyNameTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), name.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{name.ValOs}' as Name in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }


        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/weekdays")]
        public async Task<IActionResult> ModifyTimeSlotWeekdaysInDuty(string id,  string dutyId, string timeSlotId, ICollection<WeekDay> weekDays, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyWeekdaysTimeSlotDutyCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), weekDays);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set weekdays in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion

        #region DutyContacts


        //[HttpPost]
        //[Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontacts")]
        //public async Task<IActionResult> AddDutyContact(string id,  string dutyId, string timeSlotId, AddDutyContactModel model, CancellationToken token)
        //{
        //    try
        //    {
        //        var q = new GetDutyDossierIdQuery(id);
        //        var dossierId = await QueryProcessor.ProcessAsync(q, token);

        //        var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
        //        if (!p.Duties.Read) return Unauthorized();

        //        var command = new AddDutyContactCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(model.DutyContactId), model.EmployeeId, model.OrderId);
        //        var result = await CommandBus.PublishAsync(command, token);

        //        if (result.IsSuccess) return Ok();

        //        return StatusCode(500, new { errorMessage = $"Could not add employee '{model.EmployeeId}' as a new contact to TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

        //    }
        //    catch (Exception e)
        //    {
        //        return StatusCode(500, new { errorMessage = e.Message });
        //    }
        //}

        [HttpDelete]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}")]
        public async Task<IActionResult> DeleteDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new DeleteDutyContactCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete Contact '{dutyContactId}' from TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/employeeid")]
        public async Task<IActionResult> ModifyEmployeeDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, [FromBody] OneStringModel employeeId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyEmployeeDutyContactCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), employeeId.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not modify employee '{employeeId.ValOs}' in  contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/orderid")]
        public async Task<IActionResult> ModifyOrderIdDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, [FromBody] OneIntModel orderId, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new ModifyOrderIdDutyContactCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), orderId.ValInt);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{orderId.ValInt.ToString()}' as OrderId in  contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion

        #region ContactPoints en Tags
        [HttpPost]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/dutycontactpoints")]
        public async Task<IActionResult> AddContactPointToDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, AddContactPointToDutyModel model, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new AddDutyContactPointCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), DutyContactPointId.With(model.DutyContactPointId), model.ContactDetailId, model.Description);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add a new Contact Point to contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/dutycontactpoint/{dutyContactPointId}")]
        public async Task<IActionResult> DeleteContactPointToDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, string dutyContactPointId,
 CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new DeleteDutyContactPointCommand(DutyCategoryId.With(id), DutyId.With(dutyId), TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), DutyContactPointId.With(dutyContactPointId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete ContactPoint '{dutyContactPointId}' from Contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPost]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/tags")]
        public async Task<IActionResult> AddTagToDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId,
            [FromBody] OneStringModel tag, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new AddDutyContactTagCommand(DutyCategoryId.With(id), DutyId.With(dutyId),
                    TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), tag.ValOs);                
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add Tag '{tag.ValOs}' to Contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}/duty/{dutyId}/timeslot/{timeSlotId}/dutycontact/{dutyContactId}/tag/{tag}")]
        public async Task<IActionResult> DeleteTagFromDutyContact(string id,  string dutyId, string timeSlotId, string dutyContactId, 
            string tag, CancellationToken token)
        {
            try
            {
                var q = new GetDutyDossierIdQuery(id);
                var dossierId = await QueryProcessor.ProcessAsync(q, token);

                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, token);
                if (!p.Duties.Read) return Unauthorized();

                var command = new DeleteDutyContactTagCommand(DutyCategoryId.With(id), DutyId.With(dutyId),
                    TimeSlotId.With(timeSlotId), DutyContactId.With(dutyContactId), tag);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete Tag '{tag}' from Contact '{dutyContactId}' in TimeSlot '{timeSlotId}' in Duty '{dutyId}' in Duty Category '{id}'" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion
    }
}
