﻿using System;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Duty;
using CRM.Domain.Duty.ReadModels;
using Microsoft.AspNetCore.Mvc;
using WebPortal.Application.Duties.Commands;
using WebPortal.Application.Duties.Models;
using WebPortal.Application.Duties.Queries;
using WebPortal.Application.Extentions;

namespace WebPortal.API.Controllers
{
    public class DutyController : ReprojectController<DutyReadModel>
    {
        [HttpGet]
        [ProducesDefaultResponseType(typeof(PageResult<GetOverviewDutyModel>))]
        [Route("/api/duties")]
        public async Task<IActionResult> GetAllDuties(string dossierId,
            CancellationToken token,
            bool? asc = true)
        {
            try
            {
                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId,
                    token);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetDutiesQuery(dossierId,
                    asc: asc ?? true, pageSize: 100);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(PageResult<GetOverviewDutyModel>))]
        [Route("/api/duties/active")]
        public async Task<IActionResult> GetAllActiveDuties(string dossierId,
            CancellationToken ctx,
            bool? asc = true)
        {
            try
            {
                var p = await AuthorizationService.GetDossierAccess(UserContext.User, dossierId,
                    ctx);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetActiveDutiesQuery(dossierId, asc: asc ?? true);
                var result = await QueryProcessor.ProcessAsync(q, ctx);
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("{dutyId}")]
        [ProducesDefaultResponseType(typeof(DutyReadModel))]
        public async Task<IActionResult> GetDuty(string dutyId,
            CancellationToken token)
        {
            try
            {
                var q = new GetDutyQuery(dutyId);
                var result = await QueryProcessor.ProcessAsync(q, token);
                if (result == null) return NotFound();
                var p = await AuthorizationService.GetDossierAccess(UserContext.User, result.DossierId,
                    token);
                if (!p.Duties.Read) return Unauthorized();
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpDelete]
        [Route("{dutyId}")]
        [ProducesDefaultResponseType(typeof(DutyReadModel))]
        public async Task<IActionResult> DeleteDuty(string dutyId,
            CancellationToken ctx)
        {
            var c = new RemoveDutyCommand(new DutyId(dutyId));

            _ = await CommandBus.PublishAsync(c, ctx) ?? throw new Exception();
            return Ok();
        }

        [HttpPost]
        public async Task<IActionResult> CreateDuty(CreateDutyModel model,
            CancellationToken token)
        {
            var d = await QueryProcessor.ProcessAsync(new GetDutyQuery(model.Id), token);
            if (d != null && (await AuthorizationService.GetDossierAccess(UserContext.User, d.DossierId, token)).Duties.Update)
            {
                var command = new UpdateDutyCommand(DutyId.With(model.Id), model);
                var r = await CommandBus.PublishAsync(command, token);
                return Ok(model);
            }

            if (!(await AuthorizationService.GetDossierAccess(UserContext.User, model.DossierId, token)).Duties.Add) return Unauthorized();

            var c = new AddDutyCommand(DutyId.With(model.Id), model.DossierId, model.Category, model.StartDate,
                model.EndDate, model.ActionDescription, model.StartAt, model.EndAt, model.DutyEmployees);

            try
            {
                _ = await CommandBus.PublishAsync(c, token) ?? throw new Exception("We got a null in the create duty");

                var query = new GetDutyQuery(model.Id);
                var result = await QueryProcessor.ProcessAsync(query, token);

                return CreatedAtAction(nameof(GetDuty), new
                {
                    dutyId = model.Id,
                    token
                }, result);
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpGet]
        [Route("category")]
        [ProducesDefaultResponseType(typeof(GetCategoriesModel))]
        public async Task<IActionResult> GetCategory(string dossierId,
            CancellationToken token)
        {
            try
            {
                var q = new GetDutyCategoriesQuery(dossierId);
                var result = await QueryProcessor.ProcessAsync(q, token);
                if (result == null) return NotFound();
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpPost]
        [Route("reproject2")]
        public async Task<IActionResult> reproject2()
        {
            try
            {
                await ReadModelPopulator.PopulateAsync<DutyReadModelV2>(CancellationToken.None);
                return Ok();
            }
            catch (Exception e)
            {
                throw e;
                return NoContent();
            }
        }
    }
}