﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Duty;
using CRM.Domain.Duty.Models;
using CRM.Domain.Employee;
using CRM.Domain.Employee.Models;
using CRM.Domain.Employee.ReadModels;
using Microsoft.AspNetCore.Mvc;
using WebPortal.Application.Duties.Commands;
using WebPortal.Application.Duties.Queries;
using WebPortal.Application.Employees.Commands;
using WebPortal.Application.Employees.Models;
using WebPortal.Application.Employees.Queries;
using WebPortal.Application.Extentions;

namespace WebPortal.API.Controllers
{
    public class EmployeeController : ReprojectController<EmployeeReadModel>
    {
        [HttpGet]
        [Route("{employeeId}")]
        [ProducesResponseType(typeof(EmployeeReadModel), 200)]
        public async Task<IActionResult> Get(string employeeId,
            CancellationToken cancellationToken)
        {
            try
            {
                var query = new GetEmployeeQuery(employeeId);

                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
                if (result == null) return NotFound();
                if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, result.CustomerId,
                    cancellationToken)).UserManagement.Read) return Unauthorized();
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("/api/employees")]
        [ProducesResponseType(typeof(PageResult<EmployeeReadModel>), 200)]
        public async Task<IActionResult> Find(
            CancellationToken cancellationToken,
            bool asc,
            string? sort,
            int? page,
            int? pageSize,
            string? email,
            string? fullName,
            string? phone,
            string? tag,
            string? customerId)
        {
            customerId ??= UserContext.User.CustomerId ?? "";
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, customerId, cancellationToken))
                 .UserManagement.Read) return Unauthorized();
            var query = new GetAllEmployeesQuery(asc, sort ?? "Id", page ?? 1, pageSize ?? 100, email ?? "",
                fullName ?? "", phone ?? "", tag ?? "", customerId);
            var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
            Response.Headers.Add("X-pagination-total", result.Total.ToString());
            return Ok(result.Results);
        }

        [HttpGet]
        [Route("onDuty")]
        public async Task<IActionResult> getOnDuty(CancellationToken cancellationToken,
            bool asc,
            string? sort,
            int? page,
            int? pageSize,
            string? email,
            string? fullName,
            string? phone,
            string? tag,
            string? customerId)
        {
            try
            {
                customerId ??= UserContext.User.CustomerId ?? "";
                if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, customerId, cancellationToken)
                    )
                    .UserManagement.Read) return Unauthorized();
                var query = new GetAllEmployeesQuery(asc, sort ?? "Id", page ?? 1, pageSize ?? 20, email ?? "",
                    fullName ?? "", phone ?? "", tag ?? "", customerId);
                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("numbers")]
        public async Task<IActionResult> GetDashboardData([FromQuery] GetEmployeeNumbersQuery query,
            CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, query.CustomerId, cancellationToken))
                .UserManagement.Read) return Unauthorized();
            var result = await QueryProcessor.ProcessAsync(query, cancellationToken);

            return Ok(result);
        }

        [HttpPost]
        public async Task<IActionResult> CreateEmployee(CreateEmployeeModel model,
            CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, model.CustomerId, cancellationToken))
                .UserManagement.Add) return Unauthorized();
            var command = new AddEmployeeCommand(new EmployeeId(model.Id), model.CustomerId, model.FirstName,
                model.LastName, model.CcEmail, model.EmployeeComment, model.SendMailWithTicket, false,
                model.ContactDetails, model.Tags, model.DefaultTags, model.CanBeOnDuty, model.EmployeeAddress);

            try
            {
                var result = await CommandBus.PublishAsync(command, cancellationToken);
                return CreatedAtAction(nameof(Get), new
                {
                    employeeId = model.Id
                }, model);
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessaga = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}")]
        public async Task<IActionResult> UpdateEmployee(ModifyEmployeeModel model,
            CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, model.CustomerId, cancellationToken))
                .UserManagement.Update) return Unauthorized();

            var query = new GetEmployeeQuery(model.Id);
            var employee = await QueryProcessor.ProcessAsync(query, cancellationToken);

            if (employee.ContactDetails.Count > model.ContactDetails.Count)
            {
                var deletingContactpoints = new List<EmployeeContactDetailModel>();

                foreach (var cd in employee.ContactDetails.Where(x => !model.ContactDetails.Contains(x)))
                {
                    deletingContactpoints.Add(cd);
                }

                var containingDutiesQ = new GetDutyContainingEmployeeQuery(model.Id);
                var containingDuties = await QueryProcessor.ProcessAsync(containingDutiesQ, cancellationToken);

                foreach (var duty in containingDuties)
                {
                    var updatedEmployees = duty.DutyEmployees.Select(de => new DutyEmployeeModel
                    {
                        EmployeeId = de.EmployeeId,
                        Id = de.Id,
                        MarkedForRemoval = false,
                        OrderId = de.OrderId,
                        Tags = de.Tags,
                        ContactPoints = de.ContactPoints.Where(x => !deletingContactpoints.Contains(x.ContactDetailEmployee)).Select(cp =>
                                new ContactPointModel(cp.ContactPointId, cp.ContactDetailEmployee.Id, cp.Description))
                            .ToList()
                    }).ToList();

                    var commandDuties = new ModifyDutyEmployeesCommand(new DutyId(duty.Id), updatedEmployees);
                    await CommandBus.PublishAsync(commandDuties, cancellationToken);
                }
            }

            var command = new ModifyEmployeeCommand(new EmployeeId(model.Id), model.CustomerId, model.FirstName,
                model.LastName, model.CcEmail, model.EmployeeComment, model.SendMailWithTicket, false,
                model.ContactDetails, model.Tags, model.DefaultTags, model.CanBeOnDuty, model.EmployeeAddress);
            try
            {
                await CommandBus.PublishAsync(command, cancellationToken);
                return Ok();
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessaga = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}")]
        public async Task<IActionResult> DeleteEmployee(string id,
            CancellationToken cancellationToken)
        {
            var employee = await QueryProcessor.ProcessAsync(new GetEmployeeQuery(id), cancellationToken);

            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, employee.CustomerId, cancellationToken))
                .UserManagement.Delete) return Unauthorized();

            try
            {
                var containingDutiesQ = new GetDutyContainingEmployeeQuery(id);
                var containingDuties = await QueryProcessor.ProcessAsync(containingDutiesQ, cancellationToken);

                foreach (var duty in containingDuties)
                {
                    var updatedEmployees = duty.DutyEmployees.Select(de => new DutyEmployeeModel
                    {
                        EmployeeId = de.EmployeeId,
                        Id = de.Id,
                        MarkedForRemoval = de.EmployeeId.Equals(id),
                        OrderId = de.OrderId,
                        Tags = de.Tags,
                        ContactPoints = de.ContactPoints.Select(cp => new ContactPointModel(cp.ContactPointId, cp.ContactDetailEmployee.Id, cp.Description)).ToList()
                    }).ToList();

                    var commandDuties = new ModifyDutyEmployeesCommand(new DutyId(duty.Id), updatedEmployees);
                    await CommandBus.PublishAsync(commandDuties, cancellationToken);
                }

                var command = new RemoveEmployeeCommand(new EmployeeId(id));
                var result = await CommandBus.PublishAsync(command, cancellationToken);
                if (result.IsSuccess)
                {
                    return Accepted();
                }
                else
                {
                    return StatusCode(500, result);
                }
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessag = e.Message });
            }
        }

        [HttpGet]
        [Route("tags")]
        public async Task<IActionResult> GetAllTags([FromQuery] string customerId, CancellationToken cancellationToken)
        {
            var q = new GetAllTagsQuery(customerId);
            var r = await QueryProcessor.ProcessAsync(q, cancellationToken);

            if (r.Count > 0)
            {
                return Ok(r);
            }
            else
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("export")]
        public async Task<IActionResult> ExportAllEmployees([FromQuery] string customerId, CancellationToken cancellationToken)
        {
            var q = new ExportAllEmployeesQuery(customerId);
            var r = await QueryProcessor.ProcessAsync(q, cancellationToken);

            if (r != null)
            {
                var content = r.ToArray();
                var contentType = "APPLICATION/csv";
                var fileName = "employee_export";
                return File(content, contentType, fileName);
            }
            else
            {
                return NoContent();
            }
        }
    }
}