﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.DutyCategoryAndTemplate.Aggregates;
using CRM.Domain.DutyCategoryAndTemplate.Entities;
using CRM.Domain.DutyCategoryAndTemplate.ReadModels;
using CRM.Domain.Shared.ValueObjects;
using Microsoft.AspNetCore.Mvc;
using WebPortal.API.Model;
using WebPortal.Application.DutyCategories.Commands;
using WebPortal.Application.DutyCategories.Queries;
using WebPortal.Application.Extentions;
using WebPortal.Application.GlobalDutyTemplates.Commands;
using WebPortal.Application.GlobalDutyTemplates.Commands.TimeSlots;
using WebPortal.Application.GlobalDutyTemplates.Models;
using WebPortal.Application.GlobalDutyTemplates.Queries;

namespace WebPortal.API.Controllers
{
    public class GlobalDutyTemplateController : ReprojectController<GlobalDutyTemplateReadModel>
    {
        #region Queries
        [HttpGet]
        [ProducesDefaultResponseType(typeof(PageResult<GlobalDutyTemplateOverviewModel>))]
        [Route("/api/globaldutytemplates")]
        public async Task<IActionResult> GetGlobalDuties(CancellationToken token,
            bool asc = true,
            string? sort= "Title",
            int page= 1,
            int pageSize= 100,
            string? filter = "")
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetGlobalDutyTemplatesQuery(asc, sort, page, pageSize, filter);
                var result = await QueryProcessor.ProcessAsync(q, token);
                Response.Headers.Add("X-pagination-total", result.Total.ToString());
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(GlobalDutyTemplateReadModel))]
        [Route("{globalDutyTemplateId}")]
        public async Task<IActionResult> GetOneGlobalDuty(string globalDutyTemplateId,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetOneGlobalDutyTemplateQuery(globalDutyTemplateId);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);

            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(List<IdAndTitleModel>))]
        [Route("/api/globaldutytemplates/idsandtitles")]
        public async Task<IActionResult> GetTitlesOfGlobalDuties(string filter, CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetTitlesGlobalDutyQuery(filter);
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [ProducesDefaultResponseType(typeof(List<IdAndTitleModel>))]
        [Route("/api/globaldutytemplates/unsaved")]
        public async Task<IActionResult> GetUnsavedGlobalDuties(CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Read) return Unauthorized();
                var q = new GetUnsavedGlobalDutyTemplatesQuery();
                var result = await QueryProcessor.ProcessAsync(q, token);
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }
        #endregion

        #region GlobalDutyTemplate

        [HttpPost]
        public async Task<IActionResult> CreateGlobalDutyTemplate([FromBody]  OneStringModel Id, CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Add) return Unauthorized();

                var command = new CreateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(Id.ValOs));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = "Could not create a Global Duty Template" });

            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}")]
        public async Task<IActionResult> DeleteGlobalDutyTemplate(string id,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Delete) return Unauthorized();

                var q = new GetDutyCategoryIdsByGlobalDutyTemplateIdQuery(id);
                var rq = await QueryProcessor.ProcessAsync(q, token);

                if (rq != null && rq.Count != 0)
                {
                    foreach (var ducaId in rq)
                    {
                        var dcCommand = new RemoveGlobalDutyIdCommand(DutyCategoryId.With(ducaId));
                        var dcResult = await CommandBus.PublishAsync(dcCommand, token);
                        if (!dcResult.IsSuccess)
                            return StatusCode(500, new { errorMessage = $"Could not remove Global Duty Template '{id}' from Duty Category '{ducaId}'." });
                    }
                }

                var command = new DeleteGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete a Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/title")]
        public async Task<IActionResult> ModifyTitleGlobalDutyTemplate(string id,
            [FromBody] OneStringModel title, CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyTitleGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id),title.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set Title '{title}' on Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/note")]
        public async Task<IActionResult> ModifyNoteGlobalDutyTemplate(string id,
            [FromBody] OneStringModel note,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyNoteGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500,
                    new { errorMessage = $"Could not set Note '{note}' on Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }


        [HttpPost]
        [Route("{id}/copy")]
        public async Task<IActionResult> CopyGlobalDutyTemplate(string id, [FromBody] OneStringModel newId,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var query = new GetOneGlobalDutyTemplateQuery(id);
                var gdt = await QueryProcessor.ProcessAsync(query, token);

                if (gdt.Changes != null)
                    return StatusCode(403, "You cannot duplicate a Global Duty Template with unsaved changes!");

                var newTemplateId = newId.ValOs;
                var osm = newId;
                await CreateGlobalDutyTemplate(osm, token);



                var title = $"Copy of {gdt.Title}";
                osm.ValOs = title;
                await ModifyTitleGlobalDutyTemplate(newTemplateId, osm, token);

                if (!string.IsNullOrWhiteSpace(gdt.Note))
                {
                    osm.ValOs = gdt.Note;
                    await ModifyNoteGlobalDutyTemplate(newTemplateId, osm, token);
                }

                foreach (var tst in gdt.TimeSlotTemplates)
                {

                    var guid = Guid.NewGuid();
                    var tstId = $"timeslottemplate-{guid}";
                    osm.ValOs = tstId;
                    await AddTimeSlotToGlobalDutyTemplate(newTemplateId, osm, token);

                    if (!string.IsNullOrWhiteSpace(tst.TimeSlotName))
                    {
                        osm.ValOs = tst.TimeSlotName;
                        await ModifyTimeSlotNameGlobalDutyTemplate(newTemplateId, tstId, osm, token);
                    }

                    if (!string.IsNullOrWhiteSpace(tst.Note))
                    {
                        osm.ValOs = tst.Note;
                        await ModifyNoteTimeSlotGlobalDutyTemplate(newTemplateId, tstId, osm, token);
                    }

                    var time = new OneTimeSpanModel() {TimeVal = tst.StartTime};
                    await ModifyStartTimeTimeSlotGlobalDutyTemplate(newTemplateId, tstId, time, token);


                    time.TimeVal = tst.EndTime;
                    await ModifyEndTimeTimeSlotGlobalDutyTemplate(newTemplateId, tstId, time, token);

                    await ModifyWeekdaysTimeSlotGlobalDutyTemplate(newTemplateId, tstId, tst.WeekDays, token);

                }

                var command = new SaveGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(newTemplateId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(403, new { errorMessage = $"Could not save Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                throw;
            }
        }

        [HttpPost]
        [Route("{id}/save")]
        public async Task<IActionResult> SaveGlobalDutyTemplate(string id,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var query = new GetOneGlobalDutyTemplateQuery(id);
                var gdt = await QueryProcessor.ProcessAsync(query, token);

                if (gdt.Changes == null) return Ok();

                if (string.IsNullOrWhiteSpace(gdt.Changes.Title))
                    return StatusCode(403, "A Global Duty Template always requires a title!");

                if (gdt.Changes.TimeSlotTemplates == null || !gdt.Changes.TimeSlotTemplates.Any())
                    return StatusCode(403, "You need at least one Timeslot!");

                var command = new SaveGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(403, new { errorMessage = $"Could not save Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                throw;
            }

        }

        [HttpPost]
        [Route("{id}/cancel")]
        public async Task<IActionResult> CancelGlobalDutyTemplate(string id,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new CancelGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not cancel Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }

        }

        #endregion

        #region TimeSlotTemplates
        [HttpPost]
        [Route("{id}/timeslottemplates")]
        public async Task<IActionResult> AddTimeSlotToGlobalDutyTemplate(string id, [FromBody] OneStringModel timeSlotTemplateId,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new AddTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id),
                    TimeSlotTemplateId.With( timeSlotTemplateId.ValOs));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not add TimeSlot Template to Global Duty Template with Id '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }

        }

        [HttpDelete]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}")]
        public async Task<IActionResult> DeleteTimeSlotFromGlobalDutyTemplate(string id, string timeSlotTemplateId,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new DeleteTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId));
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not delete TimeSlot Template '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }

        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/starttime")]
        public async Task<IActionResult> ModifyStartTimeTimeSlotGlobalDutyTemplate(string id, string timeSlotTemplateId, OneTimeSpanModel startTime,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();


                var command = new ModifyStartTimeTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), startTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{startTime.TimeVal.ToString("g")}' as StartTime for TimeSlotTemplate '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }

        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/endtime")]
        public async Task<IActionResult> ModifyEndTimeTimeSlotGlobalDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneTimeSpanModel endTime,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyEndTimeTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), endTime.TimeVal);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{endTime.TimeVal.ToString("g")}' as EndTime for TimeSlotTemplate '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/name")]
        public async Task<IActionResult> ModifyTimeSlotNameGlobalDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneStringModel name,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyNameTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), name.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{name}' as Name for TimeSlotTemplate '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }

        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/note")]
        public async Task<IActionResult> ModifyNoteTimeSlotGlobalDutyTemplate(string id, string timeSlotTemplateId, [FromBody] OneStringModel note,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyNoteTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), note.ValOs);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set '{note}' as Note for TimeSlotTemplate '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}/timeslottemplate/{timeSlotTemplateId}/weekdays")]
        public async Task<IActionResult> ModifyWeekdaysTimeSlotGlobalDutyTemplate(string id, string timeSlotTemplateId, [FromBody] ICollection<WeekDay> weekdays,
            CancellationToken token)
        {
            try
            {
                var p = await AuthorizationService.GetCustomerAccess(UserContext.User, "*", token);
                if (!p.Duties.Update) return Unauthorized();

                var command = new ModifyWeekdaysTimeSlotTemplateGlobalDutyTemplateCommand(GlobalDutyTemplateId.With(id), TimeSlotTemplateId.With(timeSlotTemplateId), weekdays);
                var result = await CommandBus.PublishAsync(command, token);

                if (result.IsSuccess) return Ok();

                return StatusCode(500, new { errorMessage = $"Could not set weekdays in TimeSlotTemplate '{timeSlotTemplateId}' from Global Duty Template '{id}'" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        #endregion

    }
}
