﻿using System;
using System.Linq;
using System.Security.Claims;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Ticket;
using CRM.Domain.Ticket.ReadModels;
using Microsoft.AspNetCore.Mvc;
using WebPortal.API.Helpers;
using WebPortal.Application.Employees.Queries;
using WebPortal.Application.Tickets.Commands;
using WebPortal.Application.Tickets.Models;
using WebPortal.Application.Tickets.Queries;

namespace WebPortal.API.Controllers
{
    public class TicketController : ReprojectController<TicketReadModel>
    {
        [HttpGet]
        [Route("/api/tickets")]
        public async Task<IActionResult> GetTickets(CancellationToken cancellationToken,
            string dossierId,
            string? userId,
            bool asc,
            DateTime? startDateTime,
            DateTime? endDateTime,
            bool? unread,
            int? page,
            int? pageSize,
            string? sort,
            string? tagIds,
            string? filterText)
        {
            try
            {
                if (!(await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, cancellationToken))
                    .Tickets.Read) return Unauthorized();
                if (!string.IsNullOrEmpty(userId))
                {
                    if (User.Identity is ClaimsIdentity claimsIdentity)
                        if (!claimsIdentity.Claims.Any(x => x.Type == ClaimTypes.Role && x.Value == "Admin"))
                            return Unauthorized();
                }
                else
                {
                    if (User.Identity is ClaimsIdentity claimsIdentity)
                        userId = claimsIdentity.Claims.FirstOrDefault(x => x.Type == "user_id")?.Value;
                }

                var start = startDateTime ?? DateTime.UtcNow.AddDays(-365);
                var end = endDateTime ?? DateTime.UtcNow;

                var query = new GetOverviewTicketsQuery(dossierId, userId ?? "", start, end, unread ?? false, page ?? 1,
                    asc, pageSize ?? 20, string.IsNullOrEmpty(sort) ? "CreatedInCRM" : sort, tagIds ?? "",
                    filterText ?? "");
                var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
                Response.Headers.Add("X-pagination-total", result.Total.ToString());
                return Ok(result.Results);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("{id}")]
        public async Task<IActionResult> GetTicket(string id, CancellationToken cancellationToken)
        {
            try
            {
                var q = new GetTicketQuery(id, UserContext.User.Id);
                var result = await QueryProcessor.ProcessAsync(q, cancellationToken);

                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("interactionId")]
        public async Task<IActionResult> GetTicketWithInteractionId([FromQuery] string interactionId, CancellationToken cancellationToken)
        {
            try
            {
                var q = new GetTicketWithInteractionKeyQuery(interactionId);
                var result = await QueryProcessor.ProcessAsync(q, cancellationToken);

                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("numbers")]
        public async Task<IActionResult> GetTicketNumbers(CancellationToken cancellationToken,
            string dossierId,
            string? userId,
            DateTime? startDateTime,
            DateTime? endDateTime)
        {
            try
            {
                if (!(await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, cancellationToken))
                    .Tickets.Read) return Unauthorized();
                if (!string.IsNullOrEmpty(userId))
                {
                    if (User.Identity is ClaimsIdentity claimsIdentity)
                        if (!claimsIdentity.Claims.Any(x => x.Type == ClaimTypes.Role && x.Value == "Admin"))
                            return Unauthorized();
                }
                else
                {
                    if (User.Identity is ClaimsIdentity claimsIdentity)
                        userId = claimsIdentity.Claims.FirstOrDefault(x => x.Type == "user_id")?.Value;
                }

                var start = startDateTime ?? DateTime.UtcNow.AddMonths(-1);
                var end = endDateTime ?? DateTime.UtcNow;

                var q = new GetTicketStatsQuery(dossierId, start, end, userId);
                var result = await QueryProcessor.ProcessAsync(q, cancellationToken);

                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpPost]
        [Route("read")]
        public async Task<IActionResult> AddReadStatus(TicketStatusReadModel model, CancellationToken cancellationToken)
        {
            var ticket = await QueryProcessor.ProcessAsync(
                new GetTicketQuery(model.TicketId, UserContext.User.Id), cancellationToken);
            if (!(await AuthorizationService.GetDossierAccess(UserContext.User, ticket.DossierId, cancellationToken))
                .Tickets.Read) return Unauthorized();
            var command = new SetTicketAsReadOrUnreadCommand(new TicketId(model.TicketId), model.UserId, model.Read);
            var result = await CommandBus.PublishAsync(command, cancellationToken);

            if (result.IsSuccess)
                return await GetTicket(model.TicketId, cancellationToken);
            else
                return StatusCode(500, result);
        }

        [HttpGet]
        [Route("/api/tickets/export")]
        public async Task<IActionResult> ExportTickets(CancellationToken cancellationToken,
            string dossierId,
            DateTime? startDateTime,
            DateTime? endDateTime,
            bool? unread,
            string? fileType,
            string? userId)
        {
            try
            {
                if (!(await AuthorizationService.GetDossierAccess(UserContext.User, dossierId, cancellationToken))
                    .Tickets.Read) return Unauthorized();

                //if (!string.IsNullOrEmpty(userId))
                //{
                //    if (User.Identity is ClaimsIdentity claimsIdentity)
                //        if (!claimsIdentity.Claims.Any(x => x.Type == ClaimTypes.Role && x.Value == "Admin"))
                //            return Unauthorized();
                //}
                //else
                //{
                if (User.Identity is ClaimsIdentity claimsIdentity)
                    userId = claimsIdentity.Claims.FirstOrDefault(x => x.Type == "user_id")?.Value;
                //}

                var start = startDateTime ?? DateTime.UtcNow.AddDays(-365);
                var end = endDateTime ?? DateTime.UtcNow;

                if (string.IsNullOrEmpty(fileType))
                {
                    var query = new GetTicketStatsQuery(dossierId, start, end, "");
                    var result = await QueryProcessor.ProcessAsync(query, cancellationToken);
                    return Ok(result.TotalTickets);
                }

                var q = new GetTicketsForExportQuery(dossierId, start, end, unread ?? false, userId ?? "", fileType);
                var r = await QueryProcessor.ProcessAsync(q, cancellationToken);

                if (r != null)
                {
                    var content = r.ToArray();
                    var contentType = $"APPLICATION/{fileType}";
                    var fileName = $"ticket_export_{DateTime.Now.Day + DateTime.Now.Month + DateTime.Now.Year + "" }";
                    return File(content, contentType, fileName);
                }
                else
                {
                    return NoContent();
                }
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpPost]
        public async Task<IActionResult> CreateTicket(CreateTicketModel model, CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetDossierAccess(UserContext.User, model.DossierId, cancellationToken))
                .Tickets.Read) return Unauthorized();
            var command = new AddTicketCommand(new TicketId(model.Id), model.DossierId, model.TicketText,
                string.IsNullOrEmpty(model.OriginalCcmMessageId)? 0: int.Parse(model.OriginalCcmMessageId), model.CreatedInCrm ?? DateTime.Now, model.LastModifiedInCrm ?? DateTime.Now, UserContext.User.UserName, UserContext.User.UserName);

            try
            {
                await CommandBus.PublishAsync(command, cancellationToken);
                return CreatedAtAction(nameof(GetTicket), new
                {
                    id = model.Id
                }, model);
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessaga = e.Message });
            }
        }

        [HttpPut]
        [Route("{id}")]
        public async Task<IActionResult> UpdateTicket(ModifyTicketModel model,
            CancellationToken cancellationToken)
        {
            var c = new ModifyTicketCommand(new TicketId(model.Id), model.TicketText, model.LastModifiedInCrm ?? DateTime.Now, UserContext.User.UserName);
            try
            {
                await CommandBus.PublishAsync(c, cancellationToken);
                return CreatedAtAction(nameof(GetTicket), new
                {
                    id = model.Id
                }, model);
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessaga = e.Message });
            }
        }


        [HttpDelete]
        [Route("{id}")]
        public async Task<IActionResult> DeleteTicket(string id,
            CancellationToken cancellationToken)
        {
            var c = new RemoveTicketCommand(new TicketId(id));
            _ = await CommandBus.PublishAsync(c, cancellationToken) ?? throw new Exception();
            return Ok();
        }
    }
}