﻿using System;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.User;
using CRM.Domain.User.ReadModels;
using FirebaseAdmin.Auth;
using Microsoft.AspNetCore.Mvc;
using WebPortal.Application.Extentions;
using WebPortal.Application.Users.Commands;
using WebPortal.Application.Users.Models;
using WebPortal.Application.Users.Queries;

namespace WebPortal.API.Controllers
{
    public class UserController : ReprojectController<UserReadModel>
    {
        private readonly FirebaseAuth _auth;

        public UserController(FirebaseAuth auth)
        {
            _auth = auth;
        }

        [HttpGet]
        [Route("{id}/username")]
        public async Task<IActionResult> GetName(string id,
            CancellationToken cancellationToken)
        {
            var q = new GetUsernameQuery(id);
            var result = await QueryProcessor.ProcessAsync(q, cancellationToken);
            return Ok(result);
        }

        [HttpGet]
        [Route("{id}")]
        public async Task<IActionResult> GetUserById(string id,
            CancellationToken token)
        {
            try
            {
                if (id == UserContext.User.Id) return Ok(UserContext.User);
                if (id.Equals("undefined")) return BadRequest("id is undefined");
                var q = new GetUserQuery(id);
                var result = await QueryProcessor.ProcessAsync(q, token);
                if (result == null) return NoContent();
                if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, result.CustomerId ?? "*",
                        token))
                    .UserManagement.Read && UserContext.User.Id != id) return Unauthorized();
                return Ok(result);
            }
            catch (TaskCanceledException e)
            {
                return NoContent();
            }
        }

        [HttpGet]
        [Route("users")]
        [ProducesResponseType(typeof(PageResult<GetUserListModel>), 200)]
        public async Task<IActionResult> GetGlobalUsers(
            bool asc,
            int? page,
            int? pageSize,
            string? sort,
            string? filter, CancellationToken cancellationToken)
        {
            if (!string.IsNullOrEmpty(UserContext.User.CustomerId))
                return await GetCustomerUsers(cancellationToken, UserContext.User.CustomerId, sort ?? null, filter ?? "", page ?? 1, pageSize ?? 20, asc);
            var query = new GetGlobalUsersQuery(asc, sort ?? "", page ?? 1, pageSize ?? 20, filter ?? "");
            var resp = await QueryProcessor.ProcessAsync(query, cancellationToken);
            Response.Headers.Add("X-pagination-total", resp.Total.ToString());
            return Ok(resp.Results);
        }

        [HttpGet]
        [Route("customer-users")]
        [ProducesResponseType(typeof(PageResult<GetUserListModel>), 200)]
        public async Task<IActionResult> GetCustomerUsers(CancellationToken cancellationToken,
            string? customerId,
            string? sort,
            string? filter,
            int page = 1,
            int pageSize = 20,
            bool asc = true)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, customerId, cancellationToken))
                .UserManagement.Read) return Unauthorized();
            var query = new GetUsersByCustomerIdQuery(asc, sort ?? "", page, pageSize, customerId ?? "", filter ?? "");
            var resp = await QueryProcessor.ProcessAsync(query, cancellationToken);

            Response.Headers.Add("X-pagination-total", resp.Total.ToString());
            return Ok(resp.Results);
        }

        [HttpGet]
        [Route("numbers")]
        public async Task<IActionResult> GetUserMgmtDashboardData([FromQuery] GetUserMgmtDashboardDataQuery query,
            CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, query.CustomerId, cancellationToken))
                .UserManagement.Read) return Unauthorized();
            var resp = await QueryProcessor.ProcessAsync(query, cancellationToken);
            return Ok(resp);
        }

        [HttpPut]
        [Route("{id}")]
        public async Task<IActionResult> updateUser(UpdateUserModel model,
            string id,
            CancellationToken cancellationToken)
        {
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, model.CustomerId, cancellationToken))
                .UserManagement.Update) return Unauthorized();
            model.Id = id;
            await CommandBus.PublishAsync(new ModifyUserCommand(model), cancellationToken);
            return Ok();
        }

        [HttpPut]
        [Route("credentials/{id}")]
        public async Task<IActionResult> UpdateUserEmailAndUsername(UpdateEmailUsernameModel model,
            string id,
            CancellationToken cancellationToken)
        {
            if (!UserContext.User.Id.Equals(id)) return Unauthorized();

            await CommandBus.PublishAsync(new ModifyEmailUsernameCommand(new UserId(id), model.Email, model.Username), cancellationToken);
            return Ok();
        }

        [HttpPost]
        public async Task<IActionResult> CreateUser(CreateUserModel model,
            CancellationToken token)
        {
            
            if (UserContext.User.CustomerId != null)
                model.CustomerId ??= UserContext.User.CustomerId;
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, model.CustomerId, token))
                .UserManagement.Add) return Unauthorized();
            var c = new CreateUserCommand(new UserId(model.Id), model.Email, model.UserName, model.CustomerId,
                model.IsCustomerAdmin ?? false, model.Tenant);

            try
            {
                var result = await CommandBus.PublishAsync(c, token);
                return CreatedAtAction(nameof(GetUserById), new
                {
                    id = model.Id,
                    token
                }, model);
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpDelete]
        [Route("{id}")]
        public async Task<IActionResult> DeleteUser(string id,
            CancellationToken token)
        {
            var user = await QueryProcessor.ProcessAsync(new GetUserQuery(id), token);
            if (!(await AuthorizationService.GetCustomerAccess(UserContext.User, user.CustomerId, token))
                .UserManagement.Delete) return Unauthorized();
            try
            {
                var c = new DeleteUserCommand(new UserId(id));
                var result = await CommandBus.PublishAsync(c, token);
                if (result.IsSuccess)
                    return Ok();
                return StatusCode(500, new { errorMessage = "Delete user failed" });
            }
            catch (Exception e)
            {
                return StatusCode(500, new { errorMessage = e.Message });
            }
        }

        [HttpGet]
        [Route("check")]
        public async Task<IActionResult> CheckIfExist(string email,
            CancellationToken token) =>  Ok(await QueryProcessor.ProcessAsync(new GetUserByEmailQuery(email), token));
    }
}