﻿using System;
using AspNetCore.Firebase.Authentication.Extensions;
using CRM.Domain.Branch;
using CRM.Domain.Branch.ReadModels;
using CRM.Domain.CallScript;
using CRM.Domain.CallScript.ReadModels;
using CRM.Domain.Customer;
using CRM.Domain.Customer.ReadModels;
using CRM.Domain.Dossier;
using CRM.Domain.Dossier.ReadModels;
using CRM.Domain.DutyCategoryAndTemplate.Aggregates;
using CRM.Domain.DutyCategoryAndTemplate.ReadModels;
using CRM.Domain.Duty;
using CRM.Domain.Duty.ReadModels;
using CRM.Domain.Employee;
using CRM.Domain.Employee.ReadModels;
using CRM.Domain.InteractionStatistic;
using CRM.Domain.InteractionStatistic.ReadModels;
using CRM.Domain.Ticket;
using CRM.Domain.Ticket.ReadModels;
using CRM.Domain.User;
using CRM.Domain.User.ReadModels;
using EventFlow.AspNetCore.Extensions;
using EventFlow.DependencyInjection.Extensions;
using EventFlow.Extensions;
using EventFlow.MongoDB.Extensions;
using FirebaseAdmin;
using FirebaseAdmin.Auth;
using Google.Apis.Auth.OAuth2;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.OpenApi.Models;
using WebPortal.API.Helpers;
using WebPortal.Application.Authorization;
using WebPortal.Application.Branches.Commands;
using Prometheus;
using Prometheus.HttpMetrics;
using CRM.Domain.NeoDuty;
using CRM.Domain.NeoDuty.ReadModels;

namespace WebPortal.API;

    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        public void ConfigureServices(IServiceCollection services)
        {
            services.AddMvc().AddNewtonsoftJson();
            services.AddSingleton<AuthorizationService>();
            services.AddHttpContextAccessor();
            services.AddFirebaseAuthentication(Configuration["FirebaseAuthentication:Issuer"],
                Configuration["FirebaseAuthentication:Audience"]);

            services.AddAuthorization(options => { options.AddPolicy("User", policy => policy.RequireRole("User")); });
            services.AddSwaggerGen(c =>
            {
                c.SwaggerDoc("v1", new OpenApiInfo
                {
                    Version = "v1",
                    Title = "Web Portal API",
                    Description = "",
                    Contact = new OpenApiContact
                    {
                        Name = "Triple A IT", Email = "info@triple-a-it.be", Url = new Uri("https://www.triple-a-it.be")
                    }
                });

                c.AddSecurityDefinition("Bearer",
                    new OpenApiSecurityScheme
                    {
                        In = ParameterLocation.Header,
                        Description = "Please enter into field the word 'Bearer' following by space and JWT",
                        Name = "Authorization",
                        Type = SecuritySchemeType.ApiKey
                    });
            });
            services.AddTransient<ResponseHelper>();
            services.AddCors(o => o.AddPolicy("AllowAll", builder =>
            {
                builder.WithOrigins(Configuration.GetSection("AppUrl").Get<string[]>())
                    .AllowAnyMethod()
                    .AllowCredentials()
                    .AllowAnyHeader();
            }));

            services.AddEventFlow(o =>
                o
                    .ConfigureMongoDb(Configuration["Db:ConnectionString"], Configuration["Db:DbName"])
                    .AddDefaults(typeof(UserAggregate).Assembly)
                    .AddDefaults(typeof(CustomerAggregate).Assembly)
                    .AddDefaults(typeof(BranchAggregate).Assembly)
                    .AddDefaults(typeof(CallScriptAggregate).Assembly)
                    .AddDefaults(typeof(DossierAggregate).Assembly)
                    .AddDefaults(typeof(DutyAggregate).Assembly)
                    .AddDefaults(typeof(NeoDutyAggregate).Assembly)
                    .AddDefaults(typeof(EmployeeAggregate).Assembly)
                    .AddDefaults(typeof(InteractionStatisticAggregate).Assembly)
                    .AddDefaults(typeof(TicketAggregate).Assembly)
                    .AddDefaults(typeof(DutyCategoryAggregate).Assembly)
                    .AddDefaults(typeof(GlobalDutyTemplateAggregate).Assembly)
                    .AddDefaults(typeof(CreateBranchCommand).Assembly)
                    .UseMongoDbEventStore()
                    .UseMongoDbSnapshotStore()
                    .UseMongoDbReadModel<CallScriptReadModel>()
                    .UseMongoDbReadModel<CustomerReadModel>()
                    .UseMongoDbReadModel<BranchReadModel>()
                    .UseMongoDbReadModel<DossierReadModel>()
                    .UseMongoDbReadModel<EmployeeReadModel>()
                    .UseMongoDbReadModel<InteractionStatisticReadModel>()
                    .UseMongoDbReadModel<TicketReadModel>()
                    .UseMongoDbReadModel<UserReadModel>()
                    .UseMongoDbReadModel<DutyCategoryReadModel>()
                    .UseMongoDbReadModel<NeoDutyReadModel>()
                    .UseMongoDbReadModel<GlobalDutyTemplateReadModel>()
                    .UseMongoDbReadModel<DutyReadModelV2>()
                    .AddMetadataProvider<MetaDataProvider>()
                    //.AddAspNetCore(o => o.AddUserClaimsMetadata()).Configure(c => c.PopulateReadModelEventPageSize = 10000)
                    .AddAspNetCore().Configure(c => c.PopulateReadModelEventPageSize = 1000000)
            );

            services.AddSingleton<IConfiguration>(Configuration);
            // Firebase
            services.AddSingleton(x => FirebaseApp.Create(new AppOptions
            {
                Credential = GoogleCredential.GetApplicationDefault()
            }));

            services.AddSingleton(x =>
            {
                var firebase = x.GetService<FirebaseApp>();
                return FirebaseAuth.GetAuth(firebase);
            });

            services.AddScoped<IUserContext, UserContext>();

            services.AddHealthChecks().ForwardToPrometheus();
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app,
            IWebHostEnvironment env)
        {
            app.UseMiddleware<ResponseTimeMiddleware>();
            if (!env.IsProduction())
            {
                app.UseCors("AllowAll");
                app.UseDeveloperExceptionPage();
                app.UseSwagger();
                app.UseSwaggerUI(c => { c.SwaggerEndpoint("/swagger/v1/swagger.json", "Web Portal API"); });
            }
            else
            {
                app.UseCors("Production");
            }

            app.UseRouting();
            app.UseHttpMetrics(options =>
            {
            });
            app.UseStaticFiles();
            app.UseAuthentication();
            app.UseAuthorization();

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapRazorPages();
                endpoints.MapControllers();
                endpoints.MapMetrics();
            });

            if (env.IsDevelopment()) 
                app.UseSpa(spa => spa.UseProxyToSpaDevelopmentServer("http://localhost:4200"));
        }
    }
