﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.DutyCategoryAndTemplate.ReadModels;
using CRM.Domain.Employee.ReadModels;
using CRM.Domain.NeoDuty.Models;
using CRM.Domain.NeoDuty.ReadModels;
using CRM.Domain.Shared.ValueObjects;
using EventFlow.MongoDB.ReadStores;
using EventFlow.Queries;
using WebPortal.Application.NeoDuties.Models.Dashboard;
using WebPortal.Application.NeoDuties.Models.Detail;

namespace WebPortal.Application.NeoDuties.Queries
{
    public class GetNeoDutyDashBoardQuery: IQuery<List<DutyCategoryDbModel>>
    {
        public GetNeoDutyDashBoardQuery(string dossierId)
        {
            DossierId = dossierId;
        }

        public string DossierId { get; set; }
    }

    public class GetNeoDutyDashBoardQueryHandler: IQueryHandler<GetNeoDutyDashBoardQuery, List<DutyCategoryDbModel>>
    {
        private readonly IMongoDbReadModelStore<NeoDutyReadModel> _ndStore;
        private readonly IMongoDbReadModelStore<EmployeeReadModel> _empStore;
        private readonly IMongoDbReadModelStore<DutyCategoryReadModel> _dcStore;

        public GetNeoDutyDashBoardQueryHandler(IMongoDbReadModelStore<NeoDutyReadModel> ndStore, IMongoDbReadModelStore<EmployeeReadModel> empStore, IMongoDbReadModelStore<DutyCategoryReadModel> dcStore)
        {
            _ndStore = ndStore;
            _empStore = empStore;
            _dcStore = dcStore;
        }

        public Task<List<DutyCategoryDbModel>> ExecuteQueryAsync(GetNeoDutyDashBoardQuery query, CancellationToken cancellationToken)
        {
            return Task.FromResult(
                GetResults(query)
                );
        }

        public List<DutyCategoryDbModel> GetResults(GetNeoDutyDashBoardQuery query)
        {
            var rtn = new List<DutyCategoryDbModel>();

            var allDutyCats = _dcStore.AsQueryable()
                .Where(x => x.DossierId == query.DossierId)
                .ToList();

            if (!allDutyCats.Any()) return rtn;

            var allCatIds = allDutyCats.Select(x => x.Id).ToList();

            var duties = _ndStore.AsQueryable()
                .Where(x => !x.Deleted 
                    && allCatIds.Contains(x.DutyCategoryId) 
                    && x.Start < DateTime.UtcNow 
                    && x.End > DateTime.UtcNow)
                .ToList();

            if (!duties.Any()) return rtn;

            var dutyCatIds = duties.Select(x => x.DutyCategoryId).Distinct().ToList();

            var catNameAndId = allDutyCats
                .Where(x => !x.Deleted && dutyCatIds.Contains(x.Id))
                .Select(x => new 
                {
                    x.CategoryName,
                   x.Id
                })
                .OrderBy(x => x.CategoryName)
                .ToList();

            var n = DateTime.Now;

            var rightNow = new TimeSpan(n.Hour, n.Minute, 0);

            var tempTimeslots = new List<TimeSlotTemp>();

            foreach (var d in duties)
            {
                foreach (var t in d.TimeSlots)
                {
                    if (t.DutyContacts == null || t.DutyContacts.Count == 0) continue;

                    if (!(t.StartTime <= rightNow && t.EndTime >= rightNow)) continue;

                    var toDay = new WeekDay(((int) n.DayOfWeek).ToString());

                    var isToDay = t.WeekDays.Contains(toDay);

                    if (!isToDay) continue;

                    var ts = new TimeSlotTemp()
                    {
                        DutyCatId = d.DutyCategoryId,
                        ContactDetails = t.DutyContacts.Select(x => new ContactDetailModel(x)).ToList(),
                        Note = SetNote(d.Note, t.Note)
                    };

                    tempTimeslots.Add(ts);

                }
            }

            if (!tempTimeslots.Any()) return rtn;

            var employeeIds = tempTimeslots
                .SelectMany(x => x.ContactDetails
                .Select(y => y.EmployeeId))
                .Distinct().ToList();

            var allEmployees = _empStore.AsQueryable()
                .Where(x => employeeIds.Contains(x.Id))
                .ToList()
                .Select(x => new
                {
                    x.Id,
                    x.FullName,
                    x.ContactDetails
                })
                .ToList();

            foreach (var t in tempTimeslots)
            {
                t.ContactDbList = new List<ContactDbModel>();

                foreach (var c in t.ContactDetails)
                {
                    var contact = new ContactDbModel()
                    {
                        OrderId = c.OrderId,
                        Tags = c.Tags,
                        ContactPoints = new List<ContactPointDbModel>()
                    };


                    var emp = allEmployees.SingleOrDefault(x => x.Id == c.EmployeeId);
                    if (emp != null)
                        contact.FullName = emp.FullName;
                    else
                    {
                        contact.FullName = "Employee is missing!";
                        t.ContactDbList.Add(contact);
                        continue;
                    }
                    foreach (var p in c.ContactPoints)
                    {
                        var cpDb = new ContactPointDbModel();
                        var con = emp.ContactDetails.SingleOrDefault(x => x.Id == p.ContactDetailId);

                        if (con != null)
                        {
                            cpDb.ContactPointTitle = con.ContactDetailName;
                            cpDb.ContactPointValue = con.ContactDetailValue;
                        }
                        else
                        {
                            cpDb.ContactPointTitle = "Contactpoint is missing!";
                        }
                        contact.ContactPoints.Add(cpDb);
                    }
                    t.ContactDbList.Add(contact);
                    t.ContactDbList = t.ContactDbList.OrderBy(x => x.OrderId).ToList();
                }
            }


            foreach (var cnId in catNameAndId)
            {
                var ts = tempTimeslots.FirstOrDefault(x => x.DutyCatId == cnId.Id);
                if (ts == null) continue;
                var dcm = new DutyCategoryDbModel()
                {
                    CategoryName = cnId.CategoryName,
                    Note = ts.Note,
                    Contacts = ts.ContactDbList
                };
                rtn.Add(dcm);
            }

            return rtn; 
        }

        private string SetNote (string dutyNote, string tsNote )
        {
            if (string.IsNullOrEmpty(dutyNote) && string.IsNullOrEmpty(tsNote)) return "";

            if (string.IsNullOrEmpty(dutyNote)) return tsNote;

            if (string.IsNullOrEmpty(tsNote)) return dutyNote;

            return $"{dutyNote}\n\n{tsNote}";
        }
    }

    public class TimeSlotTemp
    {
        public string DutyCatId { get; set; }
        public string Note { get; set; }
        public List<ContactDetailModel> ContactDetails { get; set; }
        public List<ContactDbModel> ContactDbList { get; set; }
    }
}
