﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Dossier.ReadModels;
using CRM.Domain.DutyCategoryAndTemplate.ReadModels;
using CRM.Domain.Employee.ReadModels;
using CRM.Domain.NeoDuty.Enumerations;
using CRM.Domain.NeoDuty.ReadModels;
using EventFlow.MongoDB.ReadStores;
using EventFlow.Queries;
using WebPortal.Application.Extentions;
using WebPortal.Application.NeoDuties.Models;

namespace WebPortal.Application.NeoDuties.Queries
{
    public class GetOverviewNeoDutiesQuery : PageQuery<NeoDutyOverviewModel>
    {
        public string DossierId { get; set; }
        public string DutyCategoryId { get; set; }
        public DateTime StartDate { get; set; }
        public DateTime EndDate { get; set; }
        public bool ShowActive { get; set; }
        public bool ShowPlanned { get; set; }
        public bool ShowExpired { get; set; }
        public bool ShowEmpty { get; set; }
        public string Filter { get; set; }

        public GetOverviewNeoDutiesQuery(string dossierId, string dutyCategoryId, DateTime startDate, DateTime endDate, bool showActive, bool showPlanned, bool showExpired, bool showEmpty, string filter, bool asc, string sort, int page, int pageSize) : base(asc, sort, page, pageSize)
        {
            DossierId = dossierId;
            DutyCategoryId = dutyCategoryId;
            StartDate = startDate;
            EndDate = endDate;
            ShowActive = showActive;
            ShowPlanned = showPlanned;
            ShowExpired = showExpired;
            ShowEmpty = showEmpty;
            Filter = filter;
        }
    }

    public class GetOverviewNeoDutiesQueryHandler : IQueryHandler<GetOverviewNeoDutiesQuery, PageResult<NeoDutyOverviewModel>>
    {
        private readonly IMongoDbReadModelStore<DutyCategoryReadModel> _dcStore;
        private readonly IMongoDbReadModelStore<NeoDutyReadModel> _ndStore;
        private readonly IMongoDbReadModelStore<EmployeeReadModel> _empStore;
        private readonly IMongoDbReadModelStore<DossierReadModel> _dosStore;

        public GetOverviewNeoDutiesQueryHandler(IMongoDbReadModelStore<DutyCategoryReadModel> dcStore, IMongoDbReadModelStore<NeoDutyReadModel> ndStore, IMongoDbReadModelStore<EmployeeReadModel> empStore, IMongoDbReadModelStore<DossierReadModel> dosStore)
        {
            _dcStore = dcStore;
            _ndStore = ndStore;
            _empStore = empStore;
            _dosStore = dosStore;
        }

        public Task<PageResult<NeoDutyOverviewModel>> ExecuteQueryAsync(GetOverviewNeoDutiesQuery query, CancellationToken cancellationToken)
        {
            return Task.FromResult(
                GetOverview(query)
                );
        }

        private PageResult<NeoDutyOverviewModel> GetOverview(GetOverviewNeoDutiesQuery query)
        {

            List<string> employeeIds = new List<string>();

            if (!string.IsNullOrEmpty(query.Filter))
            {
                var custId = _dosStore.AsQueryable().SingleOrDefault(x => x.Id == query.DossierId).CustomerId;
                employeeIds = _empStore.AsQueryable()
                    .Where(x => x.Deleted 
                    && x.CanBeOnDuty 
                    && x.CustomerId == custId)
                    .ToList()
                   .Where (x => x.SearchName.ToUpper().Contains(query.Filter.ToUpper()))
                    .Select(x => x.Id)
                    .ToList();
            }

            var neoDuties = _ndStore.AsQueryable()
                .Where(x => !x.Deleted && x.DutyCategoryId == query.DutyCategoryId)
                .Where(x => (x.Start >= query.StartDate && x.Start <= query.EndDate) 
                || (x.End >= query.StartDate && x.End <= query.EndDate))
                .ToList()
                .Where(x => employeeIds.Count == 0 || x.EmployeeIds.Any(y => employeeIds.Contains(y)))
                .Where(x => (x.Status == NeoDutyStatus.Upcoming && query.ShowPlanned)
                || (x.Status == NeoDutyStatus.Active && query.ShowActive)
                || (x.Status == NeoDutyStatus.Expired && query.ShowExpired))
                .ToList();

            if (!neoDuties.Any() && !query.ShowEmpty) return new PageResult<NeoDutyOverviewModel>();

            if (!neoDuties.Any() && query.ShowEmpty)
            {
                //haal alles op uit DutyCat
                var dutyCat = _dcStore.AsQueryable()
                    .SingleOrDefault(x => !x.Deleted && x.Id == query.DutyCategoryId && x.TimeSlotTemplates.Any());
                if (dutyCat == null)
                    return new PageResult<NeoDutyOverviewModel>();

                var duca = new NeoDutyOverviewModel()
                {
                    DutyCategoryId = dutyCat.Id,
                    End = DateTime.Now,
                    Start = DateTime.Now,
                    Id = "No Id",
                    HasChanges = false,
                    IsDuty = false,
                    IsFirstEnd = false,
                    IsFirstStart = false,
                    IsLastEnd = false,
                    IsLastStart = false,
                    Note = dutyCat.Note,
                    TimeSlots = dutyCat.TimeSlotTemplates.Select(t => new TimeSlotOverviewModel(t)).ToList()
                };

                var templateList = new List<NeoDutyOverviewModel>();
                templateList.Add(duca);
                return templateList.AsQueryable()
                    .OrderBy(query.Sort, query.Asc)
                    .GetPage(1, query.PageSize);
            }

            var rtnList = neoDuties.Select(x => new NeoDutyOverviewModel()
            {
                Id = x.Id,
                DutyCategoryId = x.DutyCategoryId,
                HasChanges = x.Changes != null,
                IsDuty = true,
                Note = x.Note,
                Start = x.Start,
                End = x.End,
                TimeSlots = x.TimeSlots.Select(t => new TimeSlotOverviewModel(t)).ToList(),
                Changes = x.Changes == null
                ? null
                : new NeoDutyOverviewChangesModel(x.Changes.Note, x.Changes.Start, x.Changes.End, x.Changes.TimeSlots.Select(t => new TimeSlotOverviewModel(t)).ToList())
            }).ToList();




            //Voeg namen van employees toe.
            var allEmployeeIds = rtnList
                .SelectMany(x => x.TimeSlots)
                .SelectMany(e => e.Contacts)
                .Select(f => f.EmployeeId)
                .Distinct().ToList();
            //var allEmployees = new List<Emp>

            var changesAllEmployeeIds = rtnList.Where(x => x.Changes != null)
                                .SelectMany(x => x.Changes.TimeSlots)
                .SelectMany(e => e.Contacts)
                .Select(f => f.EmployeeId)
                .Distinct().ToList();

            allEmployeeIds.AddRange(changesAllEmployeeIds);

            allEmployeeIds = allEmployeeIds.Distinct().ToList();

            var allEmployees = _empStore.AsQueryable()
                .Where(x => !x.Deleted && allEmployeeIds.Contains(x.Id))
                .ToList()
                .Select(x => new
                {
                    x.Id,
                    x.FullName
                })
                .ToList();

            rtnList.ForEach(d => d.TimeSlots
                .ForEach(t => t.Contacts
                    .ForEach(c => c.FullName = allEmployees
                        .Single(x => x.Id == c.EmployeeId).FullName)));

            rtnList.Where(x => x.Changes != null).ToList().ForEach(c => c.Changes.TimeSlots
            .ForEach (t => t.Contacts
            .ForEach(c => c.FullName = allEmployees
            .Single(x => x.Id == c.EmployeeId).FullName)));

            rtnList.OrderBy(x => x.Start).First().IsFirstStart = true;
            rtnList.OrderByDescending(x => x.Start).First().IsLastStart = true;
            rtnList.OrderBy(x => x.End).First().IsFirstEnd = true;
            rtnList.OrderByDescending(x => x.End).First().IsLastEnd = true;

            return rtnList.AsQueryable()
                .OrderBy(query.Sort, query.Asc)
                .GetPage(query.Page, query.PageSize);
        }
    }
}
