﻿using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using CRM.Domain.Customer.ReadModels;
using CRM.Domain.Dossier.ReadModels;
using CRM.Domain.Duty.ReadModels;
using CRM.Domain.InteractionStatistic.ReadModels;
using CRM.Domain.Shared.Enums;
using CRM.Domain.Ticket.ReadModels;
using EventFlow.MongoDB.ReadStores;
using EventFlow.Queries;
using WebPortal.CCMConnector.Application.Dossiers.Models;

namespace WebPortal.CCMConnector.Application.Dossiers.Queries
{
    public class GetFaultyDossiersQueryHandler: IQueryHandler<GetFaultyDossiersQuery, List<GetFaultyDossierModel>>
    {
        public GetFaultyDossiersQueryHandler(IMongoDbReadModelStore<DossierReadModel> store,
            IMongoDbReadModelStore<DutyReadModel> dutyStore,
            IMongoDbReadModelStore<InteractionStatisticReadModel> statStore,
            IMongoDbReadModelStore<TicketReadModel> ticketStore,
            IMongoDbReadModelStore<CustomerReadModel> customerStore)
        {
            _store = store;
            _dutyStore = dutyStore;
            _statStore = statStore;
            _ticketStore = ticketStore;
            _customerStore = customerStore;
        }

        private readonly IMongoDbReadModelStore<DossierReadModel> _store;
        private readonly IMongoDbReadModelStore<DutyReadModel> _dutyStore;
        private readonly IMongoDbReadModelStore<InteractionStatisticReadModel> _statStore;
        private readonly IMongoDbReadModelStore<TicketReadModel> _ticketStore;
        private readonly IMongoDbReadModelStore<CustomerReadModel> _customerStore;

        public Task<List<GetFaultyDossierModel>> ExecuteQueryAsync(GetFaultyDossiersQuery query,
            CancellationToken cancellationToken)
        {
            return Task.Run(() =>
            {
                var multipleCCMIds = _store.AsQueryable()
                    .Where(x => !x.Deleted)
                    .GroupBy(x => x.CCMId)
                    .Select(x => new
                    {
                        CCMId = x.Key.Value,
                        Counter = x.Count()

                    })
                    .Where(x => x.Counter > 1)
                    .Select(x => x.CCMId)
                    .ToList();

                var rtn = new List<GetFaultyDossierModel>();

                rtn = _store.AsQueryable()
                    .Where(x => !x.Deleted
                                && multipleCCMIds.Contains(x.CCMId.Value))
                    .Select(x => new GetFaultyDossierModel()
                    {
                        CCMId = x.CCMId,
                        DossierName = x.DossierName,
                        Id = x.Id,
                        CustomerId = x.CustomerId,
                        ConnectedToDossier = new List<GetAmountOfDataAssociatedWithDossier>()
                    })
                    .OrderBy(x => x.CCMId)
                    .ToList();

                var tempList = new List<GetFaultyDossierModel>();
                tempList.AddRange(rtn);

                foreach (var d in tempList)
                {
                    var toModify = rtn.Single(x => x.Id == d.Id);
                    toModify.CustomerName = _customerStore.AsQueryable()
                        .SingleOrDefault(x => x.Id == toModify.CustomerId)
                        ?.CustomerName;
                    var counter = _ticketStore.AsQueryable()
                        .Count(x => !x.Deleted
                                    && x.DossierId == d.Id);
                    AddAmountToDossier(counter, toModify, AggregateNameEnum.Ticket);

                    counter = _statStore.AsQueryable()
                        .Count(x => !x.Deleted
                                    && x.DossierId == d.Id);
                    AddAmountToDossier(counter, toModify, AggregateNameEnum.Statistic);

                    var distinctCustomerId = rtn
                        .Select(x => x.CustomerId)
                        .Distinct()
                        .ToList();

                    var nullCust = distinctCustomerId.Count(x => string.IsNullOrWhiteSpace(x));

                    if (distinctCustomerId.Count > 1 && nullCust == 0 ) continue;
                    
                    //Alle employees verbonden aan een duty zitten in dezelfde customer of (indien cusId == null) er zijn geen employees verbonden aan de duty.

                    counter = _dutyStore.AsQueryable()
                        .Count(x => !x.Deleted
                                    && x.DossierId == d.Id);
                    AddAmountToDossier(counter, toModify, AggregateNameEnum.Duty);
                }
                return rtn;
            }, cancellationToken);

        }

        private static void AddAmountToDossier(int counter,
            GetFaultyDossierModel toModify, AggregateNameEnum description)
        {
            if (counter > 0)
            {
                toModify.ConnectedToDossier.Add(new GetAmountOfDataAssociatedWithDossier()
                {
                    Amount = counter,
                    Description = description
                });
            }
        }
    }
}
