﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using CRM.Domain.Branch;
using CRM.Domain.Branch.Models;
using CRM.Domain.Customer;
using CRM.Domain.Dossier;
using EventFlow;
using EventFlow.Aggregates.ExecutionResults;
using WebPortal.CCMConnector.Application.Branches.Commands;
using WebPortal.CCMConnector.Application.Customers.Commands;
using WebPortal.CCMConnector.Application.Dossiers.Commands;
using WebPortal.CCMConnector.Data.Models;

namespace WebPortal.CCMConnector.Data.ConnectorLogic.Customers.CommandsWP
{
    public class CreateCustomersAndBranchesCommand
    {
        private ICommandBus _commandBus;

        private static readonly NLog.Logger Logger = NLog.LogManager.GetCurrentClassLogger();
        public CreateCustomersAndBranchesCommand(ICommandBus commandBus)
        {
            this._commandBus = commandBus;
        }

        /// <summary>
        /// Worden er wpSubscriptions meegegeven, dan gaat het om een inschrijving via CCM waar er Customers voor moeten gegenereerd worden.
        /// </summary>
        /// <param name="tempCompanyItems"></param>
        /// <param name="wpSubscriptions"></param>
        public void CreateCustomersAndBranches(List<TempCompanyItem> tempCompanyItems, List<WPSubscriptionOverview> wpSubscriptions = null)
        {
            var tempCompanyItemsOrdered = new List<TempCompanyItem>( tempCompanyItems.OrderBy(x => x.ClientId).ToList());
            int clientId = 0;
            CustomerId custId = null;
            Guid subscriptionGuid = Guid.Empty;
            foreach (var tci in tempCompanyItemsOrdered)
            {
                bool goOnWithDossiers = tci.ClientId == clientId;
                if (!goOnWithDossiers)
                {
                    clientId = tci.ClientId;
                    subscriptionGuid = Guid.NewGuid();
                    custId = CustomerId.New;
                    int maxUsers = 0;
                    WPSubscriptionOverview thisSubscr = null;
                    if (wpSubscriptions != null)
                    {
                         thisSubscr = wpSubscriptions
                            .FirstOrDefault(x => x.ClientIdCCM == tci.ClientId);
                        if (thisSubscr != null)
                        {
                            maxUsers = thisSubscr.UserAmount;
                        }
                    }
                    goOnWithDossiers = CreateCustomer(tci, custId, maxUsers).IsSuccess;
                    if (goOnWithDossiers)
                    {
                        var branchId = BranchId.New;

                        CreateBranch(tci, custId, branchId);
                        AttachMainBranchToCustomer(custId, branchId.Value);
                        tci.IsSubscription = true;

                        var moduleStrings =
                            GetModuleStringsQuery.GetModuleStrings(thisSubscr);
                        if (moduleStrings != null && moduleStrings.Count != 0)
                        {
                            ModifySubscriptions(custId, moduleStrings);
                        }
                    }
                }
                if (goOnWithDossiers)
                {
                    var dossierId = DossierId.New;
                    CreateDossier(tci, custId, dossierId);
                    var tciToUpdate = tempCompanyItems.Single(x => x.Equals(tci));
                    tciToUpdate.DossierId = dossierId.Value;
                    if (wpSubscriptions != null)
                    {
                        wpSubscriptions.Single(x => x.ClientIdCCM == tci.ClientId).CustomerIdWP = custId.Value;
                    }
                    else if (tci.TempSubscription == null)
                    {
                        AutoGenerateSubscriptionsAndCustomerModules(tci, custId, subscriptionGuid);
                    } 
                }
                   
            }
        }



        private void AutoGenerateSubscriptionsAndCustomerModules(TempCompanyItem tci, CustomerId custId, Guid subscriptionGuid)
        {
            if (tci.IsSubscription)
            {
                tci.TempSubscription = new TempWPSubscription
                {
                    ClientIdCCM = tci.ClientId,
                    CreatedBy = "CCMConnector",
                    CreatedOn = DateTime.UtcNow,
                    CustomerIdWP = custId.Value,
                    StartDate = DateTime.UtcNow,
                    EndDate = null,
                    Id = subscriptionGuid,
                    IsActiphoneCust = tci.Tenant == "Actiphone",
                    LastModifiedBy = "CCMConnector",
                    LastModifiedOn = DateTime.UtcNow,
                    MonthlySubscriptionPrice = 0,
                    UserAmount = 0,
                };
            }
            if (tci.CompanyBranches != null && tci.CompanyBranches.Count > 0)
            {
                foreach (var branch in tci.CompanyBranches)
                {
                    branch.TempCustomerModule = new TempWPCustomerModule
                    {
                        WPSubscriptionId = subscriptionGuid,
                        CCMId = branch.CustomerId,
                        ModuleId = 6, //auto Treeview
                        StartDate = DateTime.UtcNow,
                        CCMLevel = "SUB",
                        CreatedBy = "CCMConnector",
                        LastModifiedBy = "CCMConnector",
                        CreatedOn = DateTime.UtcNow,
                        LastModifiedOn = DateTime.UtcNow,
                        UnitPrice = 0

                    };
                }
            }
            else
            {
                tci.TempCustomerModule = new TempWPCustomerModule
                {
                    WPSubscriptionId = subscriptionGuid,
                    CCMId = tci.CustomerId,
                    ModuleId = 6, //auto Treeview
                    StartDate = DateTime.UtcNow,
                    CCMLevel = "MAIN",
                    CreatedBy = "CCMConnector",
                    LastModifiedBy = "CCMConnector",
                    CreatedOn = DateTime.UtcNow,
                    LastModifiedOn = DateTime.UtcNow,
                    UnitPrice = 0
                };
            }
        }

        public IExecutionResult CreateCustomer(TempCompanyItem tci, CustomerId id, int maxUsers)
        {
            try
            {
                var command = new CreateCustomerCommand(id, tci.CustomerName, tci.Tenant, tci.ClientId,
                    maxUsers);
                var result = _commandBus.PublishAsync(command, new CancellationToken()).ConfigureAwait(false).GetAwaiter().GetResult();
                return result;
            }
            catch (Exception e)
            {
                Logger.Error(e,$"Couldn't Create Customer {tci.CustomerName} with CCM client-id: {tci.ClientId}");
                throw;
            }
        }

        public IExecutionResult CreateDossier(TempCompanyItem tci,
            CustomerId customerId, DossierId dossierId)
        {
            try
            {
                var command = new CreateDossierCommand(dossierId, tci.DossierName, customerId.Value, tci.CustomerId);
                var result = _commandBus.PublishAsync(command, new CancellationToken()).ConfigureAwait(false).GetAwaiter().GetResult();
                return result;
            }
            catch (Exception e)
            {
                Logger.Error(e,$"Couldn't Create Dossier '{tci.DossierName}' for customer '{customerId}' with CCM-Id: {tci.CustomerId}");
                throw;
            }

        }

        private IExecutionResult CreateBranch(TempCompanyItem tci, CustomerId id, BranchId branchId)
        {
            try
            {
                AddressModel address = new AddressModel
                {
                    City = tci.City,
                    Country = tci.Country,
                    Number = tci.HouseNumber,
                    PostalBoxNr = tci.BoxNumber,
                    PostalCode = tci.PostalCode,
                    Street = tci.Street
                };

                var contactDetails = AddBranchContactDetailsCommand.MakeContactDetailList(tci);

                var command = new CreateBranchCommand(branchId, tci.BranchName, address, id.Value, contactDetails);
                var result = _commandBus.PublishAsync(command, new CancellationToken()).ConfigureAwait(false).GetAwaiter().GetResult();
                return result;
            }
            catch (Exception e)
            {
                Logger.Error(e,$"Couldn't create branch '{tci.BranchName}' for customer '{id}' with CCM-Id: {tci.CustomerId}");
                throw;
            }

        }

        private IExecutionResult AttachMainBranchToCustomer(CustomerId custId, string branchId)
        {
            try
            {
                var command = new ModifyCustomerMainCustomerBranchCommand(custId, branchId);
                var result = _commandBus.PublishAsync(command, new CancellationToken()).ConfigureAwait(false).GetAwaiter().GetResult();
                return result;
            }
            catch (Exception e)
            {
                Logger.Error(e, $"Couldn't attach branch '{branchId}' to customer '{custId}'");
                throw;
            }

        }

        private IExecutionResult ModifySubscriptions(CustomerId custId,
            List<string> modules)
        {
            try
            {
                var command = new ModifyCustomerSubscriptionsCommand(custId, modules);
                var result = _commandBus.PublishAsync(command, new CancellationToken()).ConfigureAwait(false).GetAwaiter().GetResult();
                return result;
            }
            catch (Exception e)
            {
                Logger.Error(e, $"Couldn't modify subscriptions for customer '{custId}'");
                throw;
            }
        }

    }
}
