﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using EventFlow;
using EventFlow.Configuration;
using EventFlow.Queries;
using WebPortal.CCMConnector.Application.Dossiers.Models;
using WebPortal.CCMConnector.Application.Dossiers.Queries;
using WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers.Commands;
using WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers.QueriesCCM;
using WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers.QueriesWP;
using WebPortal.CCMConnector.DocExporter;

namespace WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers
{
    public class DossierLogic
    {
        private readonly ICommandBus _commandBus;
        private readonly IQueryProcessor _queryProcessor;

        private static readonly NLog.Logger Logger = NLog.LogManager.GetCurrentClassLogger();
        public int DossiersToClean;

        private List<GetFaultyDossierModel> _dossiers;

        public DossierLogic(IRootResolver serviceResolver)
        {
            _commandBus = serviceResolver.Resolve<ICommandBus>();
            _queryProcessor = serviceResolver.Resolve<IQueryProcessor>();
        }

        internal async Task<string> GetFaultyDossiers()
        {
            var q = new GetFaultyDossiersQuery();
            _dossiers = await _queryProcessor
               .ProcessAsync(q, new CancellationToken());

            //ga nu na of er dossiers zijn die je terugvind in de CCM-database.

            CheckForDossiersQuery.CheckDossierIdsInCcmFiches(_dossiers);

            ExportToCSV();

            DossiersToClean = _dossiers.Where(x => !x.DossierIsRegisteredInCCM).Count();
            var pathAndFileName = Environment.CurrentDirectory + "/CSV/" + "FaultyDossiers.csv";
            return $"We have found {_dossiers.Count} faulty dossiers.\n {_dossiers.Where(x => x.DossierIsRegisteredInCCM).Count()} dossiers are connected to CCM-fiches.\n {_dossiers.Where(x => x.ConnectedToDossier.Count > 0).Count()} dossiers are connected to statistics, tickets and/or duties. "
                + "\n" + $"{_dossiers.Where(x => !x.DossierIsRegisteredInCCM).Count()} dossiers will be removed from the WebPortal. Any duties, tickets and/or statistics in these dossiers, will receive another dossierId"
                + $"\n\nDetails can be found here: {pathAndFileName}";
        }

        public string RepairDossiers()
        {
            string rtn = "";
            MoveDataAttachedToDossierIds(out var doNotRemove);

            var notRemoveIds = doNotRemove.Select(x => x.Id).ToList();
            if (notRemoveIds.Count > 0)
            {
                rtn +=
                    "The following dossiers are still corrupted but have duties, tickets or statistics attached to them and were therefore not removed:";
                doNotRemove.ForEach(x => rtn += "\n" + $"{x.DossierName} ({x.Id})");
            }

            int counter = 0;
            foreach (var d in _dossiers
                .Where(x => !x.DossierIsRegisteredInCCM 
                            && !notRemoveIds.Contains(x.Id)))
            {
               new DeleteWPDossierCommand(_commandBus, _queryProcessor).DeleteDossierWP(d.Id);
               counter++;
            }

            rtn += "\n\n" + $"{counter} dossiers are removed. {notRemoveIds.Count} dossiers are still corrupted.";
            return rtn;
        }

        private void MoveDataAttachedToDossierIds(out List<GetFaultyDossierModel> doNotRemove)
        {
            //Eerst duties, tickets en stats verplaatsen indien nodig.
            doNotRemove = new List<GetFaultyDossierModel>();

            var toMove = _dossiers
                .Where(x => !x.DossierIsRegisteredInCCM
                            && x.ConnectedToDossier.Count > 0)
                .ToList();

            if (!toMove.Any()) return;

            var idGetter = new GetIdsForMultipleTypesByDossierIdQuery(_queryProcessor);


            foreach (var tm in toMove)
            {
                var dossierToMoveTo = _dossiers.SingleOrDefault(x => x.DossierIsRegisteredInCCM && x.CCMId == tm.CCMId);

                if (dossierToMoveTo == null)
                {
                    doNotRemove.Add(tm);
                    continue;
                }

                var ids = idGetter.GetIdsBySharedId(tm.Id)
                    .ConfigureAwait(false)
                    .GetAwaiter().GetResult();
                new ReplaceDossierIdCommand(_commandBus)
                    .ReplaceTheDossierIdsInDutiesTicketsAndStats(ids, tm.Id, dossierToMoveTo.Id)
                    .ConfigureAwait(false).GetAwaiter().GetResult();
            }

        }

        private void ExportToCSV()
        {
            if (_dossiers == null || _dossiers.Count == 0) return;

            string[,] arrayStrings = new string[7, _dossiers.Count + 1];
            for (int i = 0; i < _dossiers.Count; i++)
            {
                int ccmid = _dossiers[i].CCMId ?? 0;

                string connectedto = "";

                if (_dossiers[i].ConnectedToDossier.Count > 0)
                {
                    foreach (var cd in _dossiers[i].ConnectedToDossier)
                    {
                        connectedto += $"{cd.Amount} {cd.Description}; ";
                    }
                }

                if (i == 0)
                {
                    arrayStrings[0, 0] = "CustomerId";
                    arrayStrings[1, 0] = "CustomerName";
                    arrayStrings[2, 0] = "DossierId";
                    arrayStrings[3, 0] = "DossierName";
                    arrayStrings[4, 0] = "CCMId";
                    arrayStrings[5, 0] = "Registered in CCM";
                    arrayStrings[6, 0] = "Referenced in";
                }

                arrayStrings[0, i + 1] = string.IsNullOrWhiteSpace(_dossiers[i].CustomerId)
                    ? "-" : _dossiers[i].CustomerId;
                arrayStrings[1, i + 1] = string.IsNullOrWhiteSpace(_dossiers[i].CustomerName)
                    ? "-" : _dossiers[i].CustomerName;
                arrayStrings[2, i + 1] = _dossiers[i].Id;
                arrayStrings[3, i + 1] = _dossiers[i].DossierName;
                arrayStrings[4, i + 1] = ccmid.ToString();
                arrayStrings[5, i + 1] = _dossiers[i].DossierIsRegisteredInCCM ? "Yes" : "No";
                arrayStrings[6, i + 1] = connectedto;

            }

            string path = Environment.CurrentDirectory + "/CSV/";
            string fileName = "FaultyDossiers.csv";
            TextExporter.OverWriteCSV(arrayStrings, path, fileName);
        }
    }
}
