﻿using System;
using System.Collections.Generic;
using System.Linq;
using AutoMapper;
using CRM.Domain.Dossier;
using CRM.Domain.Shared.Enums;
using EventFlow;
using EventFlow.Configuration;
using EventFlow.Queries;
using WebPortal.CCMConnector.Application.Customers.Models;
using WebPortal.CCMConnector.Data.ConnectorLogic.Customers.CommandsCCM;
using WebPortal.CCMConnector.Data.ConnectorLogic.Customers.QueriesCCM;
using WebPortal.CCMConnector.Data.ConnectorLogic.Customers.QueriesWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers.Commands;
using WebPortal.CCMConnector.Data.ConnectorLogic.Dossiers.QueriesWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.Duties.CommandsWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.General;
using WebPortal.CCMConnector.Data.ConnectorLogic.General.QueriesCCM;
using WebPortal.CCMConnector.Data.Models;

namespace WebPortal.CCMConnector.Data.ConnectorLogic.Customers.CommandsWP
{
    public class ModificationLogic : SharedLogic
    {
        private ICommandBus _commandBus;
        private readonly IQueryProcessor _queryProcessor;
        private static readonly NLog.Logger Logger = NLog.LogManager.GetCurrentClassLogger();

        public ModificationLogic(IRootResolver serviceResolver)
        {
            _commandBus = serviceResolver.Resolve<ICommandBus>();
            _queryProcessor = serviceResolver.Resolve<IQueryProcessor>();
        }

        internal void TestDeleteDossier(string dossierId)
        {
            new DeleteWPDossierCommand(_commandBus, _queryProcessor).DeleteDossierWP(dossierId);

        }

        internal int ModifyCustomersDossiersAndBranches()
        {
            var modifications = new GetModificationsInCCMForWPQuery().GetModificationsCCM();
            if (modifications == null || modifications.Count == 0)
            {
                return 0;
            }
            Logger.Info($"{modifications.Count} modifications to process.");

            ModifyMultipleMaxUsers(modifications);


            var customerIds = modifications.Where(x => x.ModificationPinPoint != "MaxUsers")
                .Select(x => x.CustomerId)
                .ToList();
            //haal voor de bovenstaande subscriptions de customers en dossiers uit de WP.
            var customers = new GetCustomers(_queryProcessor).GetCustomersAndDossiers(customerIds);

            ModifyCustomersAndDossiers(modifications, customers, _mapper.GetMapper());

            return modifications.Count;
        }


        /// <summary>
        /// Mogelijkheden: MaxUsers, Customer|Name, Customer|ModifySearchName, Customer|ModifyAddress, Customer|ModifyContactDetails, Customer|AddSubfiche, Customer|RemoveSubfiche
        /// </summary>
        /// <param name="modifications"></param>
        /// <param name="customers"></param>
        internal void ModifyMultipleMaxUsers(List<WPModification> modifications)
        {
            if (!modifications.Any(x => x.ModificationPinPoint == "MaxUsers"))
            {
                return;
            }

            //Bekijk per modificatie welke info uit CCM moet getrokken worden
            foreach (var modif in modifications.Where(x => x.ModificationPinPoint == "MaxUsers"))
            {
                //Het aantal users werd aangepast in CCM. Er moet geen tempCompanyItem worden opgehaald.
                new ModifyMaxUsersCommand(_commandBus).ModifyMaxUsers(modif.CustomerId, modif.MaxUsers);
                DebriefUploadItem.SetItemAsUploadedInWP(modif.Id);
                Logger.Info($"Customer '{modif.CustomerId}' has now maximum {modif.MaxUsers} users.");
            }
        }

        internal void ModifyCustomersAndDossiers(List<WPModification> modifications, List<GetCustomerWithDossiersModel> customers, IMapper mapper)
        {
            if (!modifications.Any(x => x.ModificationPinPoint != "MaxUsers"))
            {
                return;
            }

            foreach (var modif in modifications
                .Where(x => x.ModificationPinPoint == "Dossier|Delete"))
            {
                try
                {
                    DeleteDossiers(customers, modif);
                }
                catch (Exception e)
                {
                    Logger.Error(e, $"Couldn't execute 'Delete Dossier' modification with Id: '{modif.Id}'.");                    
                    throw;
                }
            }

            foreach (var modif in modifications
                .Where(x => x.ModificationPinPoint == "Customer|Delete"))
            {
                try
                {
                    DeleteCustomer(modif);
                }
                catch (Exception e)
                {
                    Logger.Error(e, $"Couldn't execute 'Delete Customer' modification with Id: '{modif.Id}'.");
                    throw;
                }
            }

            ExecuteNonDeleteModifications(modifications, customers, mapper);
        }

        private void ExecuteNonDeleteModifications(List<WPModification> modifications,
            List<GetCustomerWithDossiersModel> customers,
            IMapper mapper)
        {
            var clientIds = modifications.Where(x
                    => x.CCMLevel == "CLIENT" && x.ModificationPinPoint != "MaxUsers")
                .Select(x => x.CCMId).Distinct().ToList();

            var mainCustIds = modifications.Where(x => x.CCMLevel == "MAIN").Select(x => x.CCMId).Distinct().ToList();
            var getCCMNames = new GetCCMNameModificationsQuery();
            var newCustomerNames = getCCMNames.GetClientNamesForWPCustomers(clientIds);
            var newDossierNames = getCCMNames.GetCustomerNamesForWPDossiers(mainCustIds);

            var dossiersToCheck = modifications
                .Where(x => x.CCMLevel == "MAIN")
                .Select(x => x.CCMId)
                .ToList();



            //var tempCompanyItems = new GetTempCompanyItemsWithsubfichesQuery().GetTempCompanyItemsWithSubfiches(customerIds, new Dictionary<int, int>());
            var success = false;
            foreach (var modif in modifications
                .Where(x => x.ModificationPinPoint != "MaxUsers"
                            && x.ModificationPinPoint != "Dossier|Delete"
                            && x.ModificationPinPoint != "Customer|Delete"))
            {
                try
                {
                    var existingDossiers = new CheckIfDossierExistsQuery(_queryProcessor).CheckIfDossierExists(dossiersToCheck);

                    success = ExecuteModification(customers, mapper, modif, newCustomerNames, newDossierNames,
                        existingDossiers);
                }
                catch (Exception e)
                {
                    Logger.Error(e, $"Couldn't execute modification with Id '{modif.Id}'!");
                    throw;
                }

                if (success) DebriefUploadItem.SetItemAsUploadedInWP(modif.Id);
                else
                {
                    Logger.Error(
                        $"CCM generated modification {modif.Id} with an unreadable or unexecutable modificationPinpoint: '{modif.ModificationPinPoint}'. The connector could not compute.");
                    DebriefUploadItem.SetItemAsUploadedInWP(modif.Id);
                }
            }
        }

        private void DeleteCustomer(WPModification modif)
        {
            //Verwijder eerst alle dossiers gelinkt aan de customer
            var allTheIdsWeNeed = new GetIdsForMultipleTypesByCustomerIdQuery(_queryProcessor)
                .GetIdsBySharedId(modif.CustomerId)
                .ConfigureAwait(false)
                .GetAwaiter()
                .GetResult();

            var clientAndCustIds = GetCustomerIdsQuery
                .GetMainFichesByClientId(modif.CCMId);

            RemoveAllDossiersFromClient(allTheIdsWeNeed, clientAndCustIds, modif);

            var deleteCustomerAndConnectedToCustomerCommand = new DeleteCustomerAndConnectedToCustomerCommand(_commandBus);
            deleteCustomerAndConnectedToCustomerCommand.DeleteCustomerInBranchesEmployeesAndUsers(
                    allTheIdsWeNeed, modif.CustomerId)
                .ConfigureAwait(false)
                .GetAwaiter()
                .GetResult();
            deleteCustomerAndConnectedToCustomerCommand.DeleteFireBaseApp();
            DeleteSubscriptionInCCM.DeleteSubscriptionByCustomerIdWp(modif.CustomerId);
            DebriefUploadItem.SetItemAsUploadedInWP(modif.Id);
        }

        private void RemoveAllDossiersFromClient(Dictionary<AggregateNameEnum, List<string>> allTheIdsWeNeed,
            List<IdsFromCCMandWP> clientAndCustIds, WPModification modif)
        {
            var dossierIds = allTheIdsWeNeed
                .SingleOrDefault(x => x.Key == AggregateNameEnum.Dossier)
                .Value;

            if (dossierIds == null) return;
            

            foreach (var dosId in dossierIds)
            {
                new DeleteWPDossierCommand(_commandBus, _queryProcessor).DeleteDossierWP(dosId);
                var custIdCcm = clientAndCustIds.SingleOrDefault(x =>
                        x.DossierId == dosId)?
                    .MainIdCCM;
                if (custIdCcm.HasValue)
                    DebriefUploadItem.RemoveDossierIdFromCustomerFiche(custIdCcm.Value);
                else
                {
                    Logger.Error($"Couldn't find '{dosId}' in CCM Database. This dossier is already removed from the webportal and was bound to be deleted. The modificationId is {modif.Id}");
                }
            }
        }

        private void DeleteDossiers(List<GetCustomerWithDossiersModel> customers,
            WPModification modif)
        {
            string dossierId = "";
            try
            {
                dossierId = customers.Single(x =>
                    x.Id == modif.CustomerId).Dossiers.Single(x => x.CCMId == modif.CCMId).Id;
            }
            catch (Exception e)
            {
                Logger.Error(e, $"Couldn't find the dossier with CCMId '{modif.CCMId}' in modification '{modif.Id}'!");
                throw;
            }

            new DeleteWPDossierCommand(_commandBus, _queryProcessor).DeleteDossierWP(dossierId);
            DebriefUploadItem.SetItemAsUploadedInWP(modif.Id);
            DebriefUploadItem.RemoveDossierIdFromCustomerFiche(modif.CCMId);
        }

        private bool ExecuteModification(List<GetCustomerWithDossiersModel> customers,
            IMapper mapper,
            WPModification modif,
            List<CCMIdAndString> newCustomerNames,
            List<CCMIdAndString> newDossierNames,
            Dictionary<string, int> existingDossiers)
        {
            string[] whatToGet = modif.ModificationPinPoint.Split('|');
            if (whatToGet[0] == "Customer")
            {
                switch (whatToGet[1])
                {
                    #region Name and Subscription
                    case "Name":
                        var newName = newCustomerNames.First(x => x.CCMId == modif.CCMId).Value;
                        new ModifyWpCustomerNameCommand(_commandBus).ModifyCustomerName(modif.CustomerId, newName);
                        return true;
                    case "Subscription":
                        //SubscriptionModules werden gewijzigd
                        var subscr = GetAllKeysQuery.GetOneSubscription(mapper, modif.SubscriptionId);
                        var moduleStrings = GetModuleStringsQuery
                            .GetModuleStrings(subscr);
                        if (moduleStrings != null && moduleStrings.Count != 0)
                        {
                            new ModifySubscriptionCommand(_commandBus).ModifySubscriptions(modif.CustomerId, moduleStrings);
                        }
                        return true;
                    #endregion
                    case "UpdateBranch":
                        new ModifyBranchHqAddressAndContactDetailsCommand(_commandBus, _queryProcessor).ModifyBranchHq(modif.CCMId);
                        return true;
                        #region Replace
                    case "Replace":
                        //we hebben twee soorten van replace
                    //Op Client-niveau (alles wat zich onder een client-id bevindt, gaat naar een ander client-id)
                        if (whatToGet.Length < 4) return false;
                        var customerIdSource = whatToGet[2];
                        var customerIdDestination = whatToGet[3];

                        //Client komt enkel voor bij een merge van Clients.
                        if (modif.CCMLevel == "CLIENT")
                        {
                            var theDict = new GetIdsForMultipleTypesByCustomerIdQuery(_queryProcessor)
                                .GetIdsBySharedId(customerIdSource)
                                .ConfigureAwait(false)
                                .GetAwaiter()
                                .GetResult();
                            new ReplaceCustomerIdCommand(_commandBus)
                                .ReplaceTheCustomerIdInDossiersBranchesAndEmployees(theDict,
                                    customerIdDestination,
                                    customerIdSource).ConfigureAwait(false).GetAwaiter().GetResult();
                            ReplaceClientIdsInCustomerFichesCommand
                                .ReplaceClientIdsInMainFiches(customerIdSource,
                                customerIdDestination);
                            new DeleteOneCustomerCommand(_commandBus).DeleteCustomer(customerIdSource);
                            DeleteSubscriptionInCCM.DeleteSubscriptionByCustomerIdWp(customerIdSource);
                            return true;
                        }

                        //Onderstaande lijkt me momenteel de enige relevante. Bovenstaande komt erop neer dat je client A volledig merged met client B.
                        //Haal de relevante sourceCust uit 'customers'. Tel het aantal dossiers. Indien count == 1: zet de employees over naar de destinationCust en verwijder de branch. (Evt kan je meteen de hele customer verwijderen).
                        else if (modif.CCMLevel == "MAIN")
                        {
                            if (whatToGet.Length != 5) return false;

                            var dossierIdToMoveToOtherCustomerId = whatToGet[4];
                            GetCustomerWithDossiersModel custSource;
                            try
                            {
                                custSource = new GetCustomers(_queryProcessor).GetCustomersAndDossiers(new List<string>() { customerIdSource }).Single();
                            }
                            catch (Exception e)
                            {
                                Logger.Error(e, $"Couldn't find the SourceCustomer with Id: '{customerIdSource}'");
                                throw;
                            }

                            //var cust = customers.Single(x => x.Id == customerIdSource);
                            var theDict = new Dictionary<AggregateNameEnum, List<string>>();

                            if (custSource.Dossiers.Count == 1)
                            {
                                theDict = new GetIdsForMultipleTypesByCustomerIdQuery(_queryProcessor)
                                    .GetIdsBySharedId(customerIdSource)
                                    .ConfigureAwait(false)
                                    .GetAwaiter()
                                    .GetResult();
                            }
                            else
                            {
                                var dossierIdList = new List<string>() { dossierIdToMoveToOtherCustomerId };
                                theDict.Add(AggregateNameEnum.Dossier, dossierIdList);
                            }

                            new ReplaceCustomerIdCommand(_commandBus)
                                .ReplaceTheCustomerIdInDossiersBranchesAndEmployees(theDict, customerIdDestination,
                                    customerIdSource).ConfigureAwait(false).GetAwaiter().GetResult();

                            if (custSource.Dossiers.Count == 1)
                            {
                                new DeleteOneCustomerCommand(_commandBus).DeleteCustomer(customerIdSource);
                                DeleteSubscriptionInCCM.DeleteSubscriptionByCustomerIdWp(customerIdSource);
                                return true;
                            }

                            new RemoveAllEmployeesFromDutiesByDossierIdCommand(_commandBus, _queryProcessor)
                                .RemoveAllEmployeesFromDutiesByDossierId(dossierIdToMoveToOtherCustomerId)
                                .ConfigureAwait(false).GetAwaiter().GetResult();
                            return true;
                        }

                        return false;
                    #endregion


                }
            }

            else if (whatToGet[0] == "Dossier")
            {
                string dossierId = null;
                if (whatToGet[1] != "Create")
                {
                    dossierId = customers.Single(x =>
                        x.Id == modif.CustomerId).Dossiers.Single(x => x.CCMId == modif.CCMId).Id;
                }

                //CompanyBranchModel branche = null;
                //Address address = null;
                switch (whatToGet[1])
                {
                    case "Name":
                        var dossierName = newDossierNames.Single(x => x.CCMId == modif.CCMId).Value;
                        new ModifyWpDossierNameCommand(_commandBus).ModifyDossierName(dossierId, dossierName);
                        return true;
                    case "Create":
                        var itExists = existingDossiers
                            .FirstOrDefault(x => x.Value == modif.CCMId);
                        if (itExists.Value > 0)
                        {
                            Logger.Error($"The dossier with CCMId '{modif.CCMId}' in modification '{modif.Id}' already existis as dossier '{itExists.Key}'!");

                            return false;
                        }

                        var dossierIdNew = DossierId.New;
                        new CreateWpDossierCommand(_commandBus).CreateDossier(modif.CustomerId, whatToGet[2], modif.CCMId, dossierIdNew);
                        UpdateCustomersWithDossierIdsCommand
                            .UpdateOneCustomerWithDossierId(modif.CCMId, dossierIdNew.Value);
                        return true;
                    default: return false;
                }

            }

            return false;
        }

    }
}
