﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using EventFlow;
using EventFlow.Configuration;
using EventFlow.Queries;
using WebPortal.CCMConnector.Application.Tickets.Models;
using WebPortal.CCMConnector.Application.Tickets.Queries;
using WebPortal.CCMConnector.Data.ConnectorLogic.General.QueriesCCM;
using WebPortal.CCMConnector.Data.ConnectorLogic.InteractionStats.CommandsWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.InteractionStats.QueriesWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.Tickets.CommandsWP;
using WebPortal.CCMConnector.Data.ConnectorLogic.Tickets.QueriesCCM;
using WebPortal.CCMConnector.Data.ConnectorLogic.Tickets.QueriesWP;
using WebPortal.CCMConnector.Data.Models;

namespace WebPortal.CCMConnector.Data.ConnectorLogic.Tickets
{
    public class TicketLogic
    {
        private readonly ICommandBus _commandBus;
        private readonly IQueryProcessor _queryProcessor;
        private static readonly NLog.Logger Logger = NLog.LogManager.GetCurrentClassLogger();
        private List<GetFaultyTicketModel> _faultyTickets;


        public TicketLogic(IRootResolver serviceResolver)
        {
            _commandBus = serviceResolver.Resolve<ICommandBus>();
            _queryProcessor = serviceResolver.Resolve<IQueryProcessor>();
        }

        public int TicketsToClean { get; set; }

        internal async Task<string> GetFaultyTickets()
        {
            var q = new GetFaultyTicketsQuery();
            _faultyTickets = await _queryProcessor
                .ProcessAsync(q, new CancellationToken());

            TicketsToClean = _faultyTickets
                .Select(x => x.CcmMessageId)
                .Distinct()
                .Count();

            return $"We have found {TicketsToClean} original CcmMessageIds that have one or more duplicate tickets.";

        }

        public string RepairTickets()
        {
            var allCcmMessageIds = _faultyTickets
                    .Select(x => x.CcmMessageId)
                    .Distinct()
                    .ToList()
                ;
            var allTicketIdsToDelete = new List<string>();
            foreach (var cmId in allCcmMessageIds)
            {
                var ticketsOneMessageId = _faultyTickets
                    .Where(x => x.CcmMessageId == cmId)
                    .OrderByDescending(x => x.LastModified)
                    .ToList();
                var ticketToKeep = ticketsOneMessageId.First();
                ticketsOneMessageId.Remove(ticketToKeep);

                allTicketIdsToDelete.AddRange(ticketsOneMessageId.Select(x => x.TicketId));
            }

            if (allTicketIdsToDelete.Count > 0)
            {
                new DeleteTicketsCommand(_commandBus).DeleteTickets(allTicketIdsToDelete);
            }

            return $"{allTicketIdsToDelete.Count} tickets are removed. No duplicate tickets remain.";
        }

        internal int UploadTickets(List<WPModules> ticketKeys, DateTime previousRun, out List<IdsFromCCMandWP> allCustIds)
        {
            //Drie categorieën:
            //SUB: tickets ophalen voor subficheId
            //MAIN: tickets ophalen voor mainFicheId OF alle subficheIds
            //CLIENT: Tickets ophalen voor alle main én subficheIds behorende tot client.

            var clientIds = ticketKeys
                .Where(x => x.CCMLevel == "CLIENT")
                .Select(x => x.CCMId).ToList();

            var mainCustIds = ticketKeys
                .Where(x => x.CCMLevel == "MAIN")
                .Select(x => x.CCMId).ToList();

            var subCustIds = ticketKeys
                .Where(x => x.CCMLevel == "SUB")
                .Select(x => x.CCMId).ToList();

            allCustIds = GetCustomerIdsQuery
                .GetCustomerIds(true, clientIds, mainCustIds, subCustIds);


            //new GetDossierIdsQuery(_queryProcessor).AddDossierIdsToMainAndSubfiches(allCustIds);


            var messages = GetTicketsAfterPreviousRun
                .GetNewAndModifiedMessages(allCustIds, previousRun);

            if (messages.Count == 0)
                return 0;

            Logger.Info($"{messages.Count} tickets to upload.");


            var tempMessageNames = messages.SelectMany(x => x.StructuredMessages.Select(y =>
                new TempMessageName()
                {
                    CustomerId = x.CustomerId,
                    MessageName = y.MessageName,
                    Version = x.Version
                })).ToList();

            GetTicketsAfterPreviousRun.GetMessageNames(tempMessageNames);

            //Nu verder filteren op messages
            var filteredMessages = FilterMessages(messages, tempMessageNames);

            var dynamicParams = GetTicketsAfterPreviousRun.GetDynamicParams(filteredMessages);

            var allTickets = new List<TempTicketItem>();

            foreach (var mes in filteredMessages)
            {
                var dynamicList = dynamicParams.Where(x => 
                    x.MessageId == mes.MessageId).ToList();
                var ticket = new TempTicketItem(mes, dynamicList);
                allTickets.Add(ticket);
            }

            if (allTickets.Count == 0)
            {
                return 0;
            }
            var ticketsToModify = new GetExistingTicketIdsQuery(_queryProcessor)
                .GetUploadedTickets(allTickets.Select(x =>
                    x.OriginalCCMMessageId).ToList());

            new UploadTicketsCommand(_commandBus).UploadTickets(allTickets, ticketsToModify);

            UpdateStatistics(allTickets);

            return allTickets.Count;



        }

        private void UpdateStatistics(List<TempTicketItem> allTickets)
        {
            var ticketsThatCouldHaveALinkToStatistic = allTickets
                .Where(x => x.TicketId != null
                            && x.InteractionId != null)
                .Select(x => new TicketMessageAndStatId()
                {
                    MessageIdCcm = x.OriginalCCMMessageId,
                    TicketId = x.TicketId
                })
                .ToList();

            if (ticketsThatCouldHaveALinkToStatistic.Count == 0) return;

            var messageIdCcmList = ticketsThatCouldHaveALinkToStatistic
                .Select(x => x.MessageIdCcm)
                .ToList();

            var messageIdsToAdd = new GetStatsByMessageIdCcmQuery(_queryProcessor)
                .GetStatsByMessageIdCcm(messageIdCcmList);

            if (messageIdsToAdd.Count == 0) return;

            var ticketMessageAndStatIdList =
                ticketsThatCouldHaveALinkToStatistic.Where(x 
                        => messageIdsToAdd.Values.Contains(x.MessageIdCcm))
                    .ToList();

            ticketMessageAndStatIdList.ForEach(x 
                => x.StatisticId = messageIdsToAdd.Single(y 
                    => y.Value == x.MessageIdCcm).Key);

            new ModifyInteractionStatAddTicketIdCommand(_commandBus).AddTicketIdsToStatistics(ticketMessageAndStatIdList).GetAwaiter().GetResult();

        }

        private static List<TempMessageAndVersion> FilterMessages(List<TempMessageAndVersion> messages,
            List<TempMessageName> tempMessageNames)
        {
            string sectionToPass = "No MessageNames";

            try
            {
                if (tempMessageNames == null || tempMessageNames.Count == 0)
                {
                    return messages
                        .Where(x => x.IsRead && x.ControllerOK).ToList();
                }

                sectionToPass = "readAndControlled";
                var readAndControlled = messages
                    .Where(x => x.IsRead && x.ControllerOK).ToList();

                sectionToPass = "unreadAndUnControlled";
                var unreadAndUnControlled = messages
                    .Where(x => !x.IsRead && !x.ControllerOK 
                                          && x.StructuredMessages.Count > 0).ToList();

                sectionToPass = "justUnRead";
                var justUnRead = messages
                    .Where(x => !x.IsRead && x.ControllerOK 
                                          && x.StructuredMessages.Count > 0).ToList();
                sectionToPass = "justUnControlled";
                var justUnControlled = messages
                    .Where(x => x.IsRead && !x.ControllerOK 
                                         && x.StructuredMessages.Count > 0).ToList();

                sectionToPass = "filterUnReadAndUncontrolled";
                var filterUnReadAndUncontrolled = unreadAndUnControlled
                    .Where(x => x.IsRead == tempMessageNames
                                    .First(y => y.CustomerId == x.CustomerId 
                                                && y.MessageName == x.StructuredMessages
                                                    .First().MessageName && y.Version == x.Version).IsRead
                                && x.ControllerOK == tempMessageNames
                                    .First(y => y.CustomerId == x.CustomerId 
                                                && y.MessageName == x.StructuredMessages
                                                    .First().MessageName && y.Version == x.Version).ControllerOk
                    )
                    .ToList();

                sectionToPass = "filterUnread";
                var filterUnread = justUnRead
                    .Where(x => x.IsRead == tempMessageNames
                                    .First(y => y.CustomerId == x.CustomerId
                                                && y.MessageName == x.StructuredMessages
                                                    .First().MessageName && y.Version == x.Version).IsRead)
                    .ToList();

                sectionToPass = "filterUnControlled";
                var filterUnControlled = justUnControlled
                    .Where(x => x.ControllerOK == tempMessageNames
                                    .First(y => y.CustomerId == x.CustomerId
                                                && y.MessageName == x.StructuredMessages
                                                    .First().MessageName && y.Version == x.Version).ControllerOk)
                    .ToList();

                sectionToPass = "AddRange";
                readAndControlled.AddRange(filterUnReadAndUncontrolled);
                readAndControlled.AddRange(filterUnControlled);
                readAndControlled.AddRange(filterUnread);
                return readAndControlled;
            }
            catch (Exception e)
            {
                Logger.Error(e, $"Filtering tickets went wrong at '{sectionToPass}'.");
                throw;
            }


            #region all in one query (Comment)
            //var filteredMessages =
            //    messages
            //        .Where(x => (
            //                        //Geen structuredMessageNames: IsRead en ControllerOk moeten True zijn.
            //                        (x.StructuredMessages == null || x.StructuredMessages.Count == 0)
            //                        && x.IsRead == true && x.ControllerOK == true)
            //                    ||
            //                    (x.IsRead && x.ControllerOK)
            //                    ||
            //                    (
            //                        (!x.IsRead && !x.ControllerOK)

            //                        //check op beide
            //                        && (x.IsRead == tempMessageNames
            //                                .First(y => y.CustomerId == x.CustomerId
            //                                             && y.MessageName == x.StructuredMessages
            //                                                 .First().MessageName && y.Version == x.Version).IsRead)
            //                        && (x.ControllerOK == tempMessageNames
            //                                .First(y => y.CustomerId == x.CustomerId
            //                                             && y.MessageName == x.StructuredMessages
            //                                                 .First().MessageName && y.Version == x.Version).ControllerOk)
            //                    )

            //                    //Check op één van de twee.
            //                    ||
            //                    (
            //                        (!x.IsRead)
            //                        && (x.IsRead == tempMessageNames
            //                                .First(y => y.CustomerId == x.CustomerId
            //                                             && y.MessageName == x.StructuredMessages
            //                                                 .First().MessageName && y.Version == x.Version).IsRead)
            //                    )
            //                    ||
            //                    (
            //                        (!x.ControllerOK)
            //                        &&
            //                        (x.ControllerOK == tempMessageNames
            //                             .First(y => y.CustomerId == x.CustomerId
            //                                          && y.MessageName == x.StructuredMessages
            //                                              .First().MessageName && y.Version == x.Version).ControllerOk)
            //                    )
            //        ).ToList();


            //return filteredMessages;
#endregion
        }
    }
}
